/*
**    Copyright 2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "AdcStatusTableModel.h"

#include <QColor>

#include "AdcStatus.h"
#include "base/DeviceIndexEnabled.h"

namespace {
enum Columns {
    COLUMN_NUMBER
};
}

struct AdcStatusTableEntry
{
    AdcStatus status;
};

class AdcStatusTable : public QMap<DeviceIndex, AdcStatusTableEntry>
{
public:
    int rowOf(const DeviceIndex &deviceIndex) const;
    DeviceIndex indexOf(int row) const;
};

int AdcStatusTable::rowOf(const DeviceIndex &deviceIndex) const
{
    return std::distance(begin(), find(deviceIndex));
}

DeviceIndex AdcStatusTable::indexOf(int row) const
{
    int r = 0;
    for (auto it = begin(); it != end(); ++it)
        if (r++ == row)
            return it.key();
    return {};
}

AdcStatusTableModel::AdcStatusTableModel(QObject *parent)
    : DeviceIndexTableModel(parent)
    , devices(new AdcStatusTable)
{
}

AdcStatusTableModel::~AdcStatusTableModel()
{
    delete devices;
}

int AdcStatusTableModel::rowCount(const QModelIndex &parent) const
{
    return devices->size();
}

int AdcStatusTableModel::columnCount(const QModelIndex &parent) const
{
    return COLUMN_NUMBER;
}

DeviceIndex AdcStatusTableModel::deviceIndexOf(int row) const
{
    return devices->indexOf(row);
}

int AdcStatusTableModel::rowOf(const DeviceIndex &sourceIndex) const
{
    return devices->rowOf(sourceIndex);
}

QVariant AdcStatusTableModel::data(const QModelIndex &index, int role) const
{
    auto column = index.column();
    auto deviceIndex = deviceIndexOf(index.row());
    if (deviceIndex.isNull())
        return QVariant();

    // const auto device = (*devices)[deviceIndex];
    // const auto &status = device.status;

    switch (role) {
    case Qt::DisplayRole: return data_DisplayRole(deviceIndex, column);
    }
    return QVariant();
}

QVariant AdcStatusTableModel::headerData(int section, Qt::Orientation orientation, int role) const
{
    if (orientation == Qt::Horizontal && role == Qt::DisplayRole) {
        switch (static_cast<Columns>(section)) {
        case COLUMN_NUMBER: return {};
        }
    }
    return DeviceIndexTableModel::headerData(section, orientation, role);
}

void AdcStatusTableModel::setDeviceList(const DeviceIndexEnabledMap &map)
{
    beginResetModel();
    const auto &newSet = map.getEnabled();
    const auto &oldSet = makeDeviceIndexSet(devices->keys());
    const auto toAdd = newSet-oldSet;
    const auto toRemove = oldSet-newSet;
    for (const auto &i : toRemove)
        devices->remove(i);
    for (const auto &i : toAdd)
        devices->insert(i, AdcStatusTableEntry());
    endResetModel();
}

void AdcStatusTableModel::updateAdcStatus(const AdcStatus &status)
{
    const auto &devIndex = status.getDeviceIndex();
    auto it = devices->find(devIndex);
    if(it == devices->end())
        return;
    it->status = status;
    auto row = devices->rowOf(devIndex);
    emit dataChanged(index(row, 0), index(row, COLUMN_NUMBER-1),
                     {Qt::DisplayRole, Qt::BackgroundRole});
}

QVariant AdcStatusTableModel::data_DisplayRole(const DeviceIndex &index, int column) const
{
    const auto device = (*devices)[index];
    const auto &status = device.status;

    switch (column) {
    }
    return {};
}

QMap<int, bool> AdcStatusTableModel::defaultColumnState() const
{
    return {};
}
