#include "BmnTrcCore.h"

#include <QElapsedTimer>
#include <QThread>

#include "BmnTrcSignalRouter.h"

#include "base/DeviceIndexEnabled.h"
#include "base/RunDocument.h"
#include "fsm/MStreamDeviceMgrFsm.h"
#include "mldiscover/DeviceDispatcher.h"
#include "lib-common/ReadoutAppConfigDb.h"
#include "remote-control-server/RemoteControlServer.h"

BmnTrcCore::BmnTrcCore(QObject *parent)
    : ReadoutAppCore(parent)
    , deviceMgrFsm(new MStreamDeviceMgrFsm(this))
    , configDb(new ReadoutAppConfigDb(this))
{
    qRegisterMetaType<DeviceIndex>();
    qRegisterMetaType<LldpInfoCache>();
    qRegisterMetaType<RunDocument>();
    qRegisterMetaType<BmnTrcStatus>();

    auto signalRouter = &BmnTrcSignalRouter::instance();
    signalRouter->setDeviceDispatcher(deviceDispatcher);
    signalRouter->setRemoteControlServer(remoteControlServer);
    signalRouter->setConfigDb(configDb);
    signalRouter->setCore(this);
    signalRouter->setDeviceMgrFsm(deviceMgrFsm);
    signalRouter->connectAll();

    initTrcProgDescr();
}

const BaseFsm *BmnTrcCore::getFsm() const
{
    return deviceMgrFsm;
}

void BmnTrcCore::setConfig(const ConfigKey &config_key, const BaseDeviceAppConfig &newConfig)
{
    Q_UNUSED(config_key)
    if(config == newConfig)
        return;
    const auto &oldSet = config.getDeviceIndexEnabledMap().getEnabled();
    const auto &newSet = newConfig.getDeviceIndexEnabledMap().getEnabled();
    config = newConfig;

    const auto &toRemove = oldSet-newSet;
    if(!toRemove.isEmpty()) {
        for (const auto &i : toRemove)
            removeDevice(i);
    }

    const auto &toCreate = newSet-oldSet;
    for (const auto &i : toCreate) {
        if (! devices.contains(i))
            addDevice(i);
    }
    updateTrcProgDescrOptions();
    emit allowRemoteControl(config.allowRemoteControl);
}

void BmnTrcCore::performLocalStart()
{
    deviceMgrFsm->setRequestState(FsmState::Configure);
}

void BmnTrcCore::performLocalStop()
{
    deviceMgrFsm->setRequestState(FsmState::Stopping);
}

void BmnTrcCore::remoteStartRequest(const RcRunId &newRunId, const QVariant &startEvNum)
{
    if(deviceMgrFsm->getState() != FsmState::Ready) {
        const auto &msg = QString("Application isn't in '%1' state (curState='%2')")
            .arg(fsmStateToString(FsmState::Ready))
            .arg(fsmStateToString(deviceMgrFsm->getState()));
        qWarning() << msg;
        emit sendRunWarning(newRunId, msg);
        return;
    }

    rcRunId = newRunId;
    runStartEvNum = startEvNum;
    deviceMgrFsm->setRequestState(FsmState::Configure);
}

void BmnTrcCore::remoteStopRequest(const RcRunId &newRunId, const QVariant &stopEvNum)
{
    Q_UNUSED(stopEvNum)
    if(rcRunId.runIndex != newRunId.runIndex) {
        qWarning()<< QString("Got stop request with wrong run index: runIndex (%1); "
                             "remote control runIndex (%2)")
                     .arg(newRunId.runIndex)
                     .arg(newRunId.runIndex);
    }

    if(deviceMgrFsm->getState() != FsmState::Run) {
        const auto &msg = QString("Application isn't in '%1' state (curState='%2')")
                              .arg(fsmStateToString(FsmState::Run))
                              .arg(fsmStateToString(deviceMgrFsm->getState()));
        qWarning() << msg;
        emit sendRunWarning(newRunId, msg);
    }

    deviceMgrFsm->setRequestState(FsmState::Stopping);
}

void BmnTrcCore::onRemoteStatusUpdated(const RemoteControlServerStatus &status)
{
    bool changed = progDescr.updateProgramInterface(
                ProgramInterfaceRemoteControl, 0,
                status.tcpPort, status.listening, !status.portInUse,
                status.peerHost, status.peerPort);
    if(changed)
        emit updateProgramDescription(progDescr);

    if(status.portInUse) {
        QString ready4run_str = "Ready for run";
        for(DeviceIndex &index : devices.keys()){
            if(!devices[index]->getTrcDevStatus().connected){
                ready4run_str = "Not ready for run";
            }
        }
    } else if(lastRcStatus.portInUse) {
        qWarning() << "Remote control disconnected.";

        if(deviceMgrFsm->getState() != FsmState::Ready) {
            qCritical() << "Stopping current run";
            deviceMgrFsm->setRequestState(deviceMgrFsm->getState() == FsmState::Run
                         ? FsmState::Stopping : FsmState::Error);
        }
    }
    lastRcStatus = status;
}

void BmnTrcCore::onFsmStateChange(const FsmState &state, const FsmState &oldState)
{
    ReadoutAppCore::onFsmStateChange(state, oldState);

    switch (state) {
    case FsmState::Error :
        if(oldState != FsmState::Init) {
            runStartEvNum.clear();
            config.setRunState(false);
            emit configChanged(config);
        }
        break;
    case FsmState::Ready :
    {
        switch (oldState) {
        case FsmState::Calibrate :
        case FsmState::FindAdcDelay :
        case FsmState::Stopping:
            qInfo() << QString("%1 completed successfully")
                       .arg(fsmStateToString(oldState));
            break;
        default:{}
        }
        break;
    }
    case FsmState::Configure :
    {
        if(++mstreamSeqNum < 100) mstreamSeqNum = 100;
        deviceMgrFsm->setMStreamSeqNum(mstreamSeqNum);
        emit mstreamResetAll(mstreamSeqNum); // todo: do we need this resetting?
        deviceMgrFsm->setRequestState(FsmState::Starting);
        break;
    }
    case FsmState::Starting :
    {
        emit preRunReset(runStartEvNum);
        config.setRunState(true);
        emit configChanged(config);
        deviceMgrFsm->setRequestState(FsmState::Run);
        break;
    }
    case FsmState::Run :
    {
        qInfo() << QString("Start completed successfully");
        sendRunDocument();
        break;
    }
    case FsmState::Stopping :
    {
        runStartEvNum.clear();
        config.setRunState(false);
        emit configChanged(config);
        deviceMgrFsm->setRequestState(FsmState::Ready);
        break;
    }
    default:;
    }
}

void BmnTrcCore::addDevice(const DeviceIndex &index)
{
    try {
        devices.insert(index, QSharedPointer<TrcDeviceGlobal>(new TrcDeviceGlobal(index)));
    } catch (std::exception &e){
        qCritical() << QString("[%1] Error on connect to device: %2")
                       .arg(index.getDeviceIdName()).arg(e.what());
    }
}

void BmnTrcCore::removeDevice(const DeviceIndex &index)
{
    devices.remove(index);
}

void BmnTrcCore::sendRunDocument() const
{
    auto cfgObj = QVariant::fromValue(config).value<QJsonObject>();
    RunDocument rd(rcRunId, configDb->getClientIndex(), cfgObj);
    emit sendRunDocumentToRc(rd);
}

void BmnTrcCore::deviceStatusUpdated(const BmnTrcStatus &status)
{
    dev_status[status.getDeviceIndex()] = status;
    bool device_disconnected = false;
    for(DeviceIndex &index : dev_status.keys()){
        if(!dev_status[index].connected)
            device_disconnected = true;
    }
    if(isRun() && device_disconnected){
        sendRunError(rcRunId, status.getDeviceIdent() + " disconnected");
    }
}

void BmnTrcCore::initTrcProgDescr()
{
    auto cs = configDb->getConfigSelector();
    progDescr.setClientIndex(cs.getClientIndex());

    //todo: transfer it from libs.pro
#ifdef GIT_DESCR
    progDesc.ver_hash = GIT_DESCR;
#endif
#ifdef GIT_LAST_COMMIT
    progDesc.ver_date = GIT_LAST_COMMIT;
#endif

    emit updateProgramDescription(progDescr);
}

void BmnTrcCore::updateTrcProgDescrOptions()
{
    QStringList devList;
    for(const auto &dg : devices)
        devList << dg->getDeviceIndex().toString();
    progDescr.options["DevList"] = devList.join(";");
    progDescr.options["run"] = QVariant(isRun()).toString();

    emit updateProgramDescription(progDescr);
}
