#include "EthFlashDev.h"

#include <QDir>
#include <QFile>
#include <QFileInfo>
#include <QHostAddress>

#include "device-modules/FlashProg/FlashProgModule.h"
#include "device-modules/FlashProg/FlashProgStatus.h"
#include "flash-prog/FlashDevTask.h"
#include "flash-prog/FpgaBitFile.h"
#include "mregdev/qtmregdevice.h"

#include "DfpFsmStates.h"

EthFlashDev::EthFlashDev(const DeviceIndex &index, QObject *parent)
    : ModularMregDevice(index, status, parent)
    , status(index)
    , fsm(index, this)
    , taskTimer(new QTimer(this))
{
    setLiveMagicCheck(false);

    connect(this, &EthFlashDev::printFsmInput,
            &fsm, &BaseFsm::printFsmInput);

    connect(&fsm, &BaseFsm::stateChanged,
            this, &EthFlashDev::onFsmStateChange);
    connect(&fsm, &BaseFsm::stateChanged,
            [this](const FsmState &cur, const FsmState &old) {
        emit fsmStateChanged(getDeviceIndex(), cur, old);
    });
    connect(taskTimer, &QTimer::timeout, this, &EthFlashDev::runFlashProgTasks);
    taskTimer->setSingleShot(true);
}

QSharedPointer<FlashProgModule> EthFlashDev::flashProgModule()
{
    return getFirstModule<FlashProgModule>();
}

bool EthFlashDev::devFileUpdated(const DeviceIndex &i, const QString &path)
{
    if (getDeviceIndex() != i)
        return false;
    filePath = path;
    return readFile();
}

void EthFlashDev::autoDetectFile(const DeviceIndex &i, const QString &path)
{
    if (!i.isNull() && i != getDeviceIndex())
        return;

    const auto &pn = getProductSignature();
    if (pn.isEmpty())
        return;
    QFileInfo fi = QFileInfo(path, pn);
    QFileInfoList list = QDir(fi.filePath())
                             .entryInfoList({"*.bit"}, QDir::NoDotAndDotDot | QDir::Files, QDir::Time);
    if (list.isEmpty())
        return;
    auto newFilePath = list.first().filePath();

    emit devFileChanged(getDeviceIndex(), newFilePath);
}

void EthFlashDev::startDevWrite(FlashImageType imageType)
{
    FlashImage image {
        .fileName = filePath,
        .data = fpgaBitFile.getBitstream(),
        .type = imageType
    };
    bool ok = flashProgModule() && flashProgModule()->submitTask(image, FlashOp::Write);
    if(!ok)
        fsm.setRequestState(DfpFsmStateError);
}

void EthFlashDev::startDevVerify(FlashImageType imageType)
{
    if (!readFile()) {
        fsm.setRequestState(DfpFsmStateError);
        return;
    }
    FlashImage image {
        .fileName = filePath,
        .data = fpgaBitFile.getBitstream(),
        .type = imageType
    };
    bool ok = flashProgModule() && flashProgModule()->submitTask(image, FlashOp::Verify);
    if(!ok)
        fsm.setRequestState(DfpFsmStateError);
}

void EthFlashDev::startDevErase(FlashImageType imageType)
{
    FlashImage image {
        .fileName = {},
        .data = {},
        .type = imageType
    };
    bool ok = flashProgModule() && flashProgModule()->submitTask(image, FlashOp::Erase);
    fsm.setRequestState(ok ? DfpFsmStateDone : DfpFsmStateError);
}

void EthFlashDev::startTask(DFP_TASK task, FlashImageType imageType)
{
    if (fsm.getState() != DfpFsmStateMine && fsm.getState() != DfpFsmStateAlien)
        return;
    taskImageType = imageType;

    switch (task) {
    case DFP_TASK_ERASE:
        fsm.setRequestState(DfpFsmStateErasing);
        return;
    case DFP_TASK_VERIFY:
        fsm.setRequestState(DfpFsmStateReading);
        return;
    case DFP_TASK_WRITE:
        fsm.setRequestState(DfpFsmStateWriting);
        return;
    }
}

void EthFlashDev::allowFsmTransit()
{
    fsm.setAllowTransit(true);
}

void EthFlashDev::setSingleCheckMode(bool v)
{
    if(auto m = flashProgModule())
        m->setSingleCheckMode(v);
}

void EthFlashDev::on_deviceConnected()
{
    ModularMregDevice::on_deviceConnected();

    if (!flashProgModule())
        return;
    connect(flashProgModule().data(), &FlashProgModule::taskFinished,
            this, &EthFlashDev::onTaskFinished, Qt::UniqueConnection);

    if (!flashProgModule()->isConfigured())
        qCritical() << "configureFlash failed";
    readFile();
    fsm.setDeviceOnline(true);

    emit devConnected(getDeviceIndex(), getFirmwareVersionStr(), getDeviceFwVersion());

    taskTimer->start();
}

void EthFlashDev::on_deviceDisconnected()
{
    ModularMregDevice::on_deviceDisconnected();
    fsm.setDeviceOnline(false);
    taskTimer->stop();
}

void EthFlashDev::onFsmStateChange(const FsmState &state)
{
    switch (state) {
    case DfpFsmStateErasing:
        return startDevErase(taskImageType);
    case DfpFsmStateReading:
        return startDevVerify(taskImageType);
    case DfpFsmStateWriting:
        return startDevWrite(taskImageType);
    default:;
    }
}

void EthFlashDev::runFlashProgTasks()
{
    int delayMs = 100;
    if (auto mod = flashProgModule()) {
        auto task = mod->getTasks();
        bool taskIsPending = task.has_value() && task->isRunning();
        if (taskIsPending)
            delayMs = 0;
        mod->runTasks();
        if(taskIsPending) {
            auto task = mod->getTasks();
            if(task) {
                Q_ASSERT(task->isRunning());
            } else {
                task = mod->getFinishedTasks();
                Q_ASSERT(task);
                if(task->isAborted()) {
                    fsm.setRequestState(DfpFsmStateError);
                } else if(task->isComplete()){
                    fsm.setRequestState(DfpFsmStateDone);
                } else {
                    Q_UNREACHABLE();
                }
            }
        }
    }
    taskTimer->start(delayMs);
}

void EthFlashDev::onTaskFinished(const FlashDevTask &task)
{
    if (task.op == FlashOp::Write && task.isComplete())
        emit imageWritten(getDeviceIndex(), task.getImageType(), task.getImageFileName());
}

bool EthFlashDev::readFile()
{
    auto ret = readFileInt();
    fsm.setFileStatus(ret);
    return ret;
}

DFP_FILE EthFlashDev::readFileInt()
{
    if (filePath.isEmpty())
        return DFP_FILE_INVALID;
    qDebug() << __func__ << filePath;


    if (!fpgaBitFile.readFile(filePath)) {
        return DFP_FILE_INVALID;
    }

    auto pn = getProductSignature();
    // QFileInfo fileInfo(filePath);
    // auto type = !pn.isEmpty() && fileInfo.fileName().contains(pn) ? DFP_FILE_MINE : DFP_FILE_ALIEN;
    auto type = (fpgaBitFile.getProject() == pn) ? DFP_FILE_MINE : DFP_FILE_ALIEN;

    return type;
}

QString EthFlashDev::getProductSignature() const
{
    return getSdb().getProductSignature();
}

bool EthFlashDev::readStatus()
{
    bool ok = ModularMregDevice::readStatus();

    if (flashProgModule()) {
        status.flashProgStatus = flashProgModule()->readStatus();
    } else {
        status.flashProgStatus.reset();
    }
    emit statusUpdated(status);
    return ok;
}
