#include <QStatusBar>
#include <qglobal.h>
#include <QLabel>

#include "HrbAnMainWindow.h"
#include "ui_HrbAnMainWindow.h"

#include "FileReader.h"
#include "FileWriter.h"
#include "HrbPlot.h"
#include "HistogramItem.h"
#include "HrbAnalyzer.h"
#include "ReorganizationDialog.h"
#include "TcpClient.h"
#include "pnp-server/PnpDiscovery.h"
#include "add-pnp-dialog/AddPnpDialog.h"
#include "base/DaqModule.h"
#include "base/PrintInfo.h"
#include "base/threadQuit.h"

#include <qwt_plot.h>
#include <qwt_plot_grid.h>
#include <qwt_plot_marker.h>
#include <qwt_interval_data.h>
#include <qwt_valuelist.h>

#include <QCommandLineParser>
#include <QElapsedTimer>
#include <QFileDialog>
#include <QHostAddress>
#include "QSemaphore"
#include <QSettings>
#include <QShortcut>
#include <QThread>
#include <QInputDialog>

#ifdef QT_DEBUG
#   define ACCEPT_MAIN_DATA_FLOW
#endif
#define ACCEPT_MAIN_DATA_FLOW

namespace {
const int DATA_QUEUE_SIZE = 50;

const DaqModuleSet SOURCE_TYPES = {
    DaqModule(DaqModule::Type::EventBuilder),
    DaqModule(DaqModule::Type::MStream)
};

ProgramInterface getFreeProgramInterface(const ProgramDescription &pd)
{
    ProgramInterface i = pd.getProgramInterface(ProgramInterfaceMonitorData);
    if(!i.isEmpty() && i.isFree && i.enabled)
        return i;

    if(pd.getDaqModule() == DaqModule(DaqModule::Type::MStream)){
        i = pd.getProgramInterface(ProgramInterfaceDataFlow);
        if(!i.isEmpty() && i.isFree && i.enabled)
            return i;
    }

#ifdef ACCEPT_MAIN_DATA_FLOW
    i = pd.getProgramInterface(ProgramInterfaceDataFlow);
    if(!i.isEmpty() && i.isFree && i.enabled)
        return i;
#endif

    return ProgramInterface();
}

bool isFreeProg(const ProgramDescription &pd)
{
    return !getFreeProgramInterface(pd).isEmpty();
}
}

HrbAnMainWindow::HrbAnMainWindow(QWidget *parent):
    QMainWindow(parent),
    ui(new Ui::HrbAnMainWindow),
    queueFreeSizeSem(new QSemaphore(DATA_QUEUE_SIZE)),
    hrbAnalyzer(new HrbAnalyzer(queueFreeSizeSem)),
    hrbAnalyzerThread(new QThread(this)),
    reader(new FileReader(queueFreeSizeSem)),
    outputFileName("counters.dat"),
    tcpClient(new TcpClient(queueFreeSizeSem)),
    tcpClientThread(new QThread(this)),
    sourceLable(new QLabel(this)),
    eventNumLable(new QLabel(this)),
    spillNumLable(new QLabel(this)),
    curDeviceId(0)
{
    qRegisterMetaType<QVector< QVector<quint64> > >();
    ui->setupUi(this);
    QActionGroup* group = new QActionGroup( this );
    ui->actionChannels->setActionGroup(group);
    ui->actionSamples->setActionGroup(group);
    ui->comboBoxHistGroupping->insertItem(0, "Channels", HrbPlot::TYPE_CH);
    ui->comboBoxHistGroupping->insertItem(0, "Samples", HrbPlot::TYPE_SAMPLE);
    ui->comboBoxHistGroupping->insertItem(0, "Info", HrbPlot::TYPE_INFO);
#ifndef QT_DEBUG
    ui->widgetAdd->hide();
#endif

    PnpDiscovery::subscribeCacheUpdaed(this);

    setWindowTitle(QCoreApplication::applicationName());

    autoClearTimer.setInterval(20000);
    connect(&autoClearTimer, &QTimer::timeout, ui->pushButtonClear, &QPushButton::click);
    if(ui->checkBoxAutoClear->checkState() == Qt::Checked){
        emit autoClearUpdated(true);
//        autoClearTimer.start();
    }

    tcpClient->moveToThread(tcpClientThread);
    tcpClientThread->start();
    connect(this, &HrbAnMainWindow::tcpAddressChanged,
            tcpClient, &TcpClient::setAddress);
    connect(this, &HrbAnMainWindow::closing, tcpClient, &TcpClient::deleteLater);
    connect(tcpClient, &TcpClient::destroyed, tcpClientThread, &QThread::quit);

    hrbAnalyzer->moveToThread(hrbAnalyzerThread);
    hrbAnalyzerThread->start();
    connect(tcpClient, &TcpClient::dataAccuired,
            hrbAnalyzer, &HrbAnalyzer::dataAcived);
    connect(reader, &FileReader::dataAccuired,
            hrbAnalyzer, &HrbAnalyzer::dataAcived);
    connect(hrbAnalyzer, &HrbAnalyzer::countersUpdated,
            this, &HrbAnMainWindow::hrbCountersUpdated);
    connect(this, &HrbAnMainWindow::clear, hrbAnalyzer, &HrbAnalyzer::clear);
    connect(this, &HrbAnMainWindow::closing, hrbAnalyzer, &HrbAnalyzer::deleteLater);
    connect(hrbAnalyzer, &HrbAnalyzer::destroyed, hrbAnalyzerThread, &QThread::quit);
    connect(hrbAnalyzer, &HrbAnalyzer::endOfSpill, this, &HrbAnMainWindow::endOfSpill);
    connect(hrbAnalyzer, &HrbAnalyzer::eventNumUpdated, this, &HrbAnMainWindow::eventNumUpdated);
    connect(this, &HrbAnMainWindow::autoClearUpdated, hrbAnalyzer, &HrbAnalyzer::autoClearUpdated);

    new QShortcut(QKeySequence(Qt::CTRL + Qt::Key_O), this, SLOT(openFile()));
    connect(this, &HrbAnMainWindow::readerFileNameUpdated,
            reader, &FileReader::setFileName);
//    emit readerFileNameUpdated("/home/ivan/doubleHrb.dat");

    QSettings settings;
    settings.beginGroup("HrbAnMainWindow");
    restoreGeometry(settings.value("geometry").toByteArray());
    restoreState(settings.value("windowState").toByteArray());
    settings.endGroup();

    settings.beginGroup("Places");
    bool ok;
    foreach (QString key, settings.childKeys()) {
        QStringList keyParts = key.split('_');
        if(keyParts.size()!=2)
            continue;
        quint64 serial = keyParts[0].toULongLong(&ok,16);
        int type = 0;
        if(keyParts[1] == "row")
            type = 1;
        if(keyParts[1] == "col")
            type = 2;
        if(!ok || type==0 )
            continue;
        int val = settings.value(key).toInt(&ok);
        if(!ok)
            continue;
        if(type==1)
            places[serial].first = val;
        else
            places[serial].second = val;
    }
    settings.endGroup();

    settings.beginGroup("SelectedProgram");
    const auto &daqModule = DaqModule::fromString(settings.value("type").toString().trimmed());
    const auto &programIndex = ProgramIndex::fromString(settings.value("index").toString().trimmed());
    selectedProgram.setClientIndex({daqModule, programIndex});
    selectedInterface.type = settings.value("interface").toString().trimmed();
    settings.endGroup();

    settings.beginGroup("Main");
    ui->checkBoxAutoClear->setChecked(settings.value("autoClear", false).toBool());
    settings.endGroup();

    qInfo().noquote() << QString("[MW] initial source is: %1; interface: %2")
                             .arg(selectedProgram.getName()).arg(selectedInterface.type);
    statusBar()->addPermanentWidget(sourceLable);
    statusBar()->addPermanentWidget(eventNumLable);
    statusBar()->addPermanentWidget(spillNumLable);
    updateSourceLabel();
}


HrbAnMainWindow::~HrbAnMainWindow()
{
    emit closing();
    QElapsedTimer timer;
    timer.start();
    if(!threadQuiting(tcpClientThread, 2000, &timer)){
        qWarning()<<"Failed to stop tcpClientThread";
    }
    if(!threadQuiting(hrbAnalyzerThread, 2000, &timer)){
        qWarning()<<"Failed to stop hrbAnalyzerThread";
    }
    delete queueFreeSizeSem;
    delete ui;
}

void HrbAnMainWindow::init(QStringList args)
{
    switch (args.size()) {
    case 1:
    {
        // Read settings from .config
        QSettings settings;
        settings.beginGroup("Main");
        if(settings.contains("fileName")){
            emit readerFileNameUpdated(settings.value("fileName").toString());
        }
        if(settings.contains("host") && settings.contains("port")){
            QString hostStr(settings.value("host").toString());
            QString portStr(settings.value("port").toString());
            bool portOk;
            uint port = portStr.toUInt(&portOk);
            QHostAddress host(hostStr);
            if(!host.isNull() && portOk){
                emit tcpAddressChanged(host, port);
            }
        }
        settings.endGroup();
        break;
    }
    case 2:
    {
        // argument is path to file
        emit readerFileNameUpdated(args[1]);
        break;
    }
    case 3:
    {
        // arguments are host+port pair
        QString hostStr(args[1]);
        QString portStr(args[2]);
        bool portOk;
        uint port = portStr.toUInt(&portOk);
        QHostAddress host(hostStr);
        if(!host.isNull() && portOk){
            emit tcpAddressChanged(host, port);
        }
        break;
    }
    default:
        break;
    }
//    testAdd(10);
//    testAdd(8);
//    on_actionReorganize_plots_triggered();
//    reader->setFileName(argv[1]); //todo
}

void HrbAnMainWindow::pnpCacheUpdated(const PnpInfoCache &cache)
{
    pnpCache = cache;
    for(auto pd : pnpCache) {
        if(pd.sameProgram(selectedProgram)) {
            updateTcpInterface(pd);
        }
    }
    ui->actionConnect_to_data_source->setEnabled(hasFreeProgs());
}

void HrbAnMainWindow::openFile()
{
    QString fileName = QFileDialog::getOpenFileName(this, tr("Open File"),
                                                 QDir::homePath(),
                                                 tr("*.dat;;All (*)"));

    QFile f(fileName);
    if(!f.exists())
        return;

    QSettings settings;
    settings.beginGroup("Main");
    settings.setValue("fileName", fileName);
    settings.endGroup();
    emit readerFileNameUpdated(fileName);
}

void HrbAnMainWindow::hrbCountersUpdated(quint32 serial, QVector<QVector<quint64> > data, quint32 trigPassed)
{
    if(!plots.contains(serial)){
        addPlot(serial);
    }

    HrbPlot *plot = plots[serial];
    plot->updateCounters(data, trigPassed, histLimit[serial]);
}

void HrbAnMainWindow::on_pushButtonClear_clicked()
{
//    QString idStr = ui->comboBoxShowTrg->currentText().remove('-');
//    bool ok;
//    quint32 id = idStr.toUInt(&ok, 16);
//    if(ok) {
//        emit clear(id);
//    } else {
        emit clear();
//    }
}


void HrbAnMainWindow::on_checkBoxAutoClear_toggled( bool state )
{
    if(state) {
//        autoClearTimer.start();
    } else {
        autoClearTimer.stop();
    }
    saveConfig();
    emit autoClearUpdated(state);
}

void HrbAnMainWindow::on_comboBoxShowTrg_currentTextChanged(const QString &arg1)
{
    QString idStr = ui->comboBoxShowTrg->currentText().remove('-');
    bool ok;
    quint32 id = idStr.toUInt(&ok, 16);
    if(ok) {
        curDeviceId = id;
        ui->spinBoxHistLimit->setValue(histLimit[curDeviceId]);
    } else
        curDeviceId = 0;
    ui->spinBoxHistLimit->setEnabled(ok);
    foreach (quint32 serial, plots.keys()) {
        plots[serial]->setVisible(!ok || id==serial);
    }
}

void HrbAnMainWindow::on_comboBoxHistGroupping_currentTextChanged(const QString &)
{
    bool ok;
    int type = ui->comboBoxHistGroupping->currentData(Qt::UserRole).toInt(&ok);
    Q_ASSERT(ok);
    emit groppingUpdated(type);
}

void HrbAnMainWindow::on_spinBoxHistLimit_valueChanged(int limit)
{
    if(curDeviceId)
        histLimit[curDeviceId] = limit;
}

void HrbAnMainWindow::on_pushButtonAdd_clicked()
{
    quint64 id = ui->spinBoxAdd->value();
    testAdd(id);
}

void HrbAnMainWindow::on_actionConnect_to_data_source_triggered()
{
    if(hasFreeProgs()){
        ui->actionConnect_to_data_source->setEnabled(false);
        return;
    }

    AddPnpDialog dialog(pnpCache, false, this);
    dialog.set_function_filter(std::bind(&isFreeProg, std::placeholders::_1));
    PnpDiscovery::subscribeCacheUpdaed(&dialog);

    if(dialog.exec() != QDialog::Accepted)
        return;

    const auto &pd_list = dialog.getProgDescr();
    Q_ASSERT(!pd_list.isEmpty());
    if(pd_list.isEmpty()) return;
    const auto &pd = pd_list.first();

    if(!pd.sameProgram(selectedProgram) || selectedInterface.isEmpty()) {
        const auto &i = getFreeProgramInterface(pd);
        if(!i.isEmpty()) {
            qInfo().noquote() << QString("[MW] change source from %1 to %2 (%3)")
                  .arg(selectedProgram.getName()).arg(pd.getName()).arg(i.type);
            selectedProgram = pd;
            selectedInterface.type = i.type;
            selectedInterface.id = i.id;
        }
        updateTcpInterface(pd);
    }
}

void HrbAnMainWindow::on_actionReorganize_plots_triggered()
{
    ReorganizationDialog dialog(places, this);
    if(dialog.exec() == QDialog::Accepted){
        QMap<quint32, QPair<int, int> > newPlaces = dialog.getPlaces();
        foreach (quint32 ser, newPlaces.keys()) {
            Q_ASSERT(places.contains(ser));
            foreach (quint32 ser2, newPlaces.keys()) {
                Q_ASSERT(ser==ser2 || (newPlaces[ser] != newPlaces[ser2]));
            }
        }
        Q_ASSERT(newPlaces.size() == places.size());

        //remove All
        foreach (HrbPlot *plot, plots)
            ui->plotsLayout->removeWidget(plot);
        places = newPlaces;
        foreach (quint32 serial, places.keys()) {
            if(!plots.contains(serial)) continue;
            ui->plotsLayout->addWidget(plots[serial], places[serial].first, places[serial].second);
            qInfo()<<"[Set init index]\t"<<places[serial].first<<":"<<places[serial].second<<"="<<hex<<serial;
        }
        saveConfig();
    }
}

void HrbAnMainWindow::on_actionRelease_missing_plots_triggered()
{
    foreach (quint32 serial, places.keys()) {
        if(!plots.contains(serial)){
            places.remove(serial);
        }
    }
    saveConfig();
}

void HrbAnMainWindow::on_actionChannels_toggled(bool isByChannels)
{
    if(isByChannels)
        emit groppingUpdated(HrbPlot::TYPE_CH);
}

void HrbAnMainWindow::endOfSpill(int spillNumber)
{
    spillNumLable->setText(QString("Spill:%1").arg(spillNumber));
//    if(autoClearTimer.isActive())
//        autoClearTimer.start();
}

void HrbAnMainWindow::eventNumUpdated(quint32 eventNumber)
{
    eventNumLable->setText(QString("evNum:%1").arg(eventNumber));
}

void HrbAnMainWindow::addPlot(quint32 serial)
{
    if(!plots.contains(serial)){
        int row, col;
        if(places.contains(serial)){
            // take reserved place
            row = places[serial].first;
            col = places[serial].second;
        } else {
            int maxRow=0, maxCol=0;
            foreach (quint32 ser, places.keys()) {
                maxRow = qMax(maxRow, places[ser].first);
                maxCol = qMax(maxCol, places[ser].second);
            }

            bool hasFreeCell=false;
            if((maxRow+1)*(maxCol+1) > places.size()){
                // find any free cell in maxRow X maxColumn grid
                for(row=0; row <= maxRow; ++row) {
                    for(col=0; col <= maxCol; ++col) {
                        hasFreeCell=true;
                        foreach (quint32 ser, places.keys()) {
                            if(places[ser].first == row && places[ser].second == col){
                                hasFreeCell=false;
                                break;
                            }
                        }
                        if(hasFreeCell) break;
                    }
                    if(hasFreeCell) break;
                }
            }
            if(!hasFreeCell){
                // find any free place in maxRow X maxColumn grid
                if(1.*(maxCol+1)/(maxRow+1) > 1.5){
                    row = maxRow+1;
                    col = 0;
                } else {
                    row = 0;
                    col = maxCol+1;
                }
            }
            places[serial] = QPair<int, int>(row, col);
        }

        HrbPlot *plot = new HrbPlot(serial, this);
        ui->plotsLayout->addWidget(plot, row, col);
        qDebug()<<"[Add plot]\t"<<places[serial].first<<":"<<places[serial].second<<"="<<hex<<serial;

        plots[serial] = plot;
        connect(this, &HrbAnMainWindow::groppingUpdated,
                plot, QOverload<int>::of(&HrbPlot::gropingUpdated));
        updateClearTargets();
    }
}

void HrbAnMainWindow::updateClearTargets()
{
    ui->comboBoxShowTrg->clear();
    ui->comboBoxShowTrg->addItem("All");
    foreach (quint32 serial, plots.keys()) {
        QString serialStr = QString("%1-%2")
                .arg(serial >> 16,4,16,QChar('0'))
                .arg(serial & 0xFFFFu,4,16,QChar('0'));

        ui->comboBoxShowTrg->addItem(serialStr);
    }
}

void HrbAnMainWindow::closeEvent(QCloseEvent *event)
{
    saveConfig();
    QMainWindow::closeEvent(event);
}

void HrbAnMainWindow::setStatus(const QString &status)
{
    statusBar()->showMessage(status);
}

void HrbAnMainWindow::updateHistogram()
{
//    int sum=0;
////    QwtPlot *plot = ui->plot;
//    QwtPlot *plot = new QwtPlot();

//    HistogramItem *histogram;
//    foreach(QwtPlotItem *item, plot->itemList()){
//        histogram = dynamic_cast<HistogramItem*>(item);
//        if(histogram) break;
//    }

//    if(!histogram){
//        histogram = new HistogramItem();
//        histogram->setColor(Qt::darkCyan);
//        histogram->attach(plot);
//    }

//    int numValues = 0;
//    bool flagOk = false;
//    quint32 devId = curDevId.toUInt(&flagOk, 16);
//    if(!flagOk){
//        devId=0;
//    } else {
////        numValues = reader->getCounters()[devId].size(); //todo
//    }

//    QwtArray<QwtDoubleInterval> intervals(numValues);
//    QwtArray<double> values(numValues);

//    int pos = 0;
//    for (int i = 0; i < numValues; ++i)
//    {
//        intervals[i] = QwtDoubleInterval(pos, pos + 1);
////        values[i] = reader->getCounters()[devId][i];//todo
////        sum += reader->getCounters()[devId][i];//todo
//        ++pos;
//    }

//    histogram->setData(QwtIntervalData(intervals, values));

//    qDebug() << "Print " << curDevId;
//    plot->replot();
//    //todo
////    ui->labelSum->setText(QString("Integral: %1; Passed Trig: %2; Rate: %3")
////                          .arg(sum)
////                          .arg(reader->getPassTrig())
    ////                          .arg(reader->getPassTrig()?(double)sum/reader->getPassTrig():0, 2,'g', 2));
}

void HrbAnMainWindow::testAdd(quint64 id)
{
    return;
    int samNum = qrand()%10+10;
    int chNum = 96;
    QVector< QVector <quint64> > data;
    data.resize(samNum);
    for(int s=0; s<samNum; ++s){
        data[s].resize(chNum);
        for(int ch=0; ch<chNum; ++ch){
            data[s][ch] = qrand()%1000;
        }
    }

//    data.resize(3);
//    data[0].push_back(3);
//    data[0].push_back(12);
//    data[0].push_back(13);
//    data[0].push_back(5);

//    data[1].push_back(3);
//    data[1].push_back(12);
//    data[1].push_back(16);
//    data[1].push_back(5);

//    data[2].push_back(3);
//    data[2].push_back(25);
//    data[2].push_back(20);
//    data[2].push_back(5);

    hrbCountersUpdated(id, data, 0);
}

void HrbAnMainWindow::saveConfig()
{
    QSettings settings;
    settings.beginGroup("HrbAnMainWindow");
    settings.setValue("geometry", saveGeometry());
    settings.setValue("windowState", saveState());
    settings.endGroup();

    settings.remove("Places");
    settings.beginGroup("Places");
    foreach (quint64 serial, places.keys()) {
        settings.setValue(QString("0x%1_row").arg(serial, 8, 16, QChar('0')), places[serial].first);
        settings.setValue(QString("0x%1_col").arg(serial, 8, 16, QChar('0')), places[serial].second);
    }
    settings.endGroup();

    settings.remove("SelectedProgram");
    settings.beginGroup("SelectedProgram");
    settings.setValue(QString("type"), selectedProgram.getDaqModule().toString());
    settings.setValue(QString("index"), selectedProgram.getProgramIndex().toString());
    settings.setValue(QString("interface"), selectedInterface.type);
    settings.endGroup();

    settings.beginGroup("Main");
    settings.setValue("autoClear", ui->checkBoxAutoClear->checkState()==Qt::Checked);
    settings.endGroup();
}

bool HrbAnMainWindow::hasFreeProgs() const
{
    for(auto pd : pnpCache) {
        if(isFreeProg(pd))
            return true;
    }
    return false;
}

void HrbAnMainWindow::updateTcpInterface(const ProgramDescription &descr)
{
    if(descr.sameProgram(selectedProgram)) {
        bool ok;
        const ProgramInterface &i = descr.getProgramInterface(selectedInterface.type, &ok);
        if(!ok)
            return;
        if(i.host != selectedInterface.host
                || i.port != selectedInterface.port){
            qInfo().noquote() << QString("[MW] change interface from %1:%2 to %3:%4")
                  .arg(selectedInterface.host.toString()).arg(selectedInterface.port)
                  .arg(i.host.toString()).arg(i.port);

            selectedInterface = i;
            emit tcpAddressChanged(i.host, i.port);
            updateSourceLabel();
        }
    }
}

void HrbAnMainWindow::updateSourceLabel()
{
    QString msg = QString("Source: %1")
            .arg(selectedProgram.getName());
    if(!selectedInterface.isEmpty()){
        msg.append(QString(" (%1:%2)").arg(selectedInterface.type).arg(selectedInterface.id));
    } else
        msg.append(" (no interface)");
    sourceLable->setText(msg);
}
