//
//    Copyright 2013-2015 Ivan Filippov
//    Copyright 2013-2015 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "HrbMainWindow.h"
#include "ui_HrbMainWindow.h"

#include <QDateTime>
#include <QDebug>
#include <QElapsedTimer>
#include <QFileInfo>
#include <QHostAddress>
#include <QHostInfo>
#include <QHBoxLayout>
#include <QLabel>
#include <QMessageBox>
#include <QScrollBar>
#include <QSettings>
#include <QShortcut>
#include <QStringList>
#include <QTimer>
#include <QCheckBox>

#include "base/ConfigSelector.h"
#include "configuration-manager/ConfigurationManager.h"
#include "remote-control-server/RemoteControlServerStatus.h"

#include "device-discover/DiscoverDialog.h"
#include "widgets/LogViewWidget.h"
#include "m-stream-output/MStreamStatisticDialog.h"

#include "TriggerSetupDialog.h"
#include "StatisticDialog.h"
#include "HrbSignalRouter.h"
#include "HrbModel.h"
#include "TtbModel.h"
#include "TtbModelColumns.h"


HrbMainWindow::HrbMainWindow(QWidget *parent)
    : BaseMainWindow(parent)
    , ui(new Ui::HrbMainWindow)
    , configuration_manager(new ConfigurationManager(this))
    , discoverDialog(new DiscoverDialog(this))
    , statisticDialog(new StatisticDialog(this))
    , mStreamStatisticDialog(new MStreamStatisticDialog(this))
    , labelFsmState(new QLabel(this))
    , hrbModel(new HrbModel(this))
    , ttbModel(new TtbModel(this))
{
    ui->setupUi(this);
    setupMenu(ui->menuFile, ui->menuHelp, ui->menuOptions);
    ui->tabWidget->addTab(new LogViewWidget(ui->tabWidget), "Log");

    ui->tableHrb->setModel(hrbModel);
    ui->tableTtb->setModel(ttbModel);
    initTtbTable();
    updateLabelRemoteState();

    auto signalRouter = &HrbSignalRouter::instance();
    signalRouter->setMainWindow(this);
    signalRouter->setStatisticDialog(statisticDialog);
    signalRouter->setMStreamStatisticDialog(mStreamStatisticDialog);
    signalRouter->setConfigurationManager(configuration_manager);
    signalRouter->setDiscoverDialog(discoverDialog);

    statusBar()->addWidget(labelFsmState);
}

HrbMainWindow::~HrbMainWindow()
{
    delete ui;
}

void HrbMainWindow::setConfig(const ConfigKey &, const HrbAppConfig &newConfig)
{
    if(config == newConfig)
        return;

    statisticDialog->clearStatistic();
    config = newConfig;
    mStreamStatisticDialog->setDeviceList(
                config.getDeviceIndexEnabledMap(DeviceFilter::HRB));
    updateGuiVal();
}

void HrbMainWindow::onRemoteStatusUpdated(const RemoteControlServerStatus &st)
{
    underRemoteCtrl = st.portInUse;
    updateControlState();
    updateLabelRemoteState();
}

void HrbMainWindow::preparingRun(const RcRunId &_runId)
{
    runId = _runId;
    updateLabelRemoteState();
}

void HrbMainWindow::onFsmStateChange(const FsmState &state)
{
    fsmState = state;
    labelFsmState->setText("FSM: " + fsmStateToString(fsmState));
    updateControlState();
    updateLabelRemoteState();
}

void HrbMainWindow::configSelectorChanged(const ConfigIndexName &cin)
{
    auto t = ConfigSelector(DaqModule(DaqModule::Type::HrbSystem),
                            cin).toWindowTitle();
    if(windowTitle() != t)
        setWindowTitle(t);

}

void HrbMainWindow::on_pushButtonStart_clicked()
{
    emit performLocalStart();

}

void HrbMainWindow::on_pushButtonStop_clicked()
{
    emit performLocalStop();
}

void HrbMainWindow::on_actionTriggerSetup_triggered()
{
    TriggerSetupDialog *dialog = getTriggerSetupDialog();
    if(!dialog) {
        dialog = new TriggerSetupDialog(this);
    }
    dialog->show();
}

void HrbMainWindow::on_actionBoardsDetails_triggered()
{
    statisticDialog->show();
}

void HrbMainWindow::on_actionMStream_statistic_triggered()
{
    mStreamStatisticDialog->show();
}

void HrbMainWindow::on_actionPrint_FSM_inputs_triggered()
{
    emit printFsmInputs();
}

void HrbMainWindow::on_actionFindDevices_triggered()
{
    discoverDialog->show();


//    AssignIndexesDialog assignDialog(devices, hostMap, this);
//    assignDialog.init(newSystemSetup, true);
//    if(assignDialog.exec() != QDialog::Accepted)
//        return;
//    newSystemSetup = assignDialog.getSystemSetup();
//    assignDialog.init(newSystemSetup, false);
//    if(assignDialog.exec() != QDialog::Accepted)
//        return;
//    newSystemSetup = assignDialog.getSystemSetup();

//    removeAllDevices();
//    lastConfig = config;
//    config = newSystemSetup;
//    addDevices();
//    updateProgramDescription();
}

void HrbMainWindow::on_spinBoxThreshold_valueChanged(double val)
{
    if(blockControl) return;
    config.setThreshold(val);
    sendConfig();
}

void HrbMainWindow::on_spinBoxLatency_valueChanged(int val)
{
    if(blockControl) return;
    config.defaultHrb.hrb.latency = val/8;
    config.spreadGlobalCfg();
    sendConfig();
}

void HrbMainWindow::on_spinBoxWindow_valueChanged(int val)
{
    if(blockControl) return;
    config.defaultHrb.hrb.window = val/8;
    config.spreadGlobalCfg();
    sendConfig();
}

void HrbMainWindow::on_spinBoxMinHitDuration_valueChanged(int val)
{
    if(blockControl) return;
    config.defaultHrb.hrb.minHitDuration = val/2;
    config.spreadGlobalCfg();
    sendConfig();
}

void HrbMainWindow::on_actionAllow_Remote_Control_toggled(bool checked)
{
    if(blockControl) return;
    config.allowRemoteControl = checked;
    sendConfig();
}

void HrbMainWindow::on_actionShow_average_busy_time_toggled(bool showAveraged)
{
    //todo: move to local setting
    if(blockControl) return;
    config.showAveragedBusy = showAveraged;
    sendConfig();
}

void HrbMainWindow::updateControlState()
{
    bool inStateErrReady = fsmState == FsmState::Ready || fsmState == FsmState::Error;

    ui->pushButtonStart->setEnabled((fsmState == FsmState::Ready) && !underRemoteCtrl);
    ui->pushButtonStop->setEnabled((fsmState == FsmState::Run) && !underRemoteCtrl);

    ui->actionFindDevices->setEnabled(inStateErrReady);
    ui->actionConfiguration_manager->setEnabled(inStateErrReady);
    ui->actionReset_MStream->setEnabled(inStateErrReady);
    ui->actionSend_soft_reset->setEnabled(inStateErrReady);

    ui->actionAllow_Remote_Control->setEnabled(!underRemoteCtrl ||
                                               fsmState != FsmState::Run);

//    ui->actionTriggerSetup->setEnabled(inStateErrReady);
    hrbModel->setEnabled(inStateErrReady);
    ttbModel->setEnabled(inStateErrReady);

    ui->groupBoxGlobalCtrl->setEnabled(inStateErrReady);
    TriggerSetupDialog *dialog = getTriggerSetupDialog();
    if(dialog) {
        dialog->setControlEnable(inStateErrReady);
    }

}

void HrbMainWindow::sendConfig() const
{
    emit configChanged(config);
}

TriggerSetupDialog * HrbMainWindow::getTriggerSetupDialog() const
{
    for(auto c : children()) {
        auto d = qobject_cast<TriggerSetupDialog*>(c);
        if(d)
            return d;
    }
    return nullptr;
}

void HrbMainWindow::initTtbTable()
{
#if QT_VERSION >= 0x050000
    ui->tableTtb->horizontalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
#else
    ui->tableTtb->horizontalHeader()->setResizeMode(QHeaderView::ResizeToContents);
#endif

    ui->tableTtb->hideColumn(COL_TTB_IP);
    ui->tableTtb->hideColumn(COL_TTB_TIME);
}

void HrbMainWindow::on_actionReset_MStream_triggered() const
{
    emit mstream_reset_all(0);
}

void HrbMainWindow::on_actionSend_soft_reset_triggered() const
{
    emit softReset();
}

void HrbMainWindow::on_actionConfiguration_manager_triggered() const
{
    configuration_manager->show();
}

void HrbMainWindow::saveAppGuiSettings(QSettings &s) const
{
    BaseMainWindow::saveAppGuiSettings(s);
    statisticDialog->saveSettings(s);
}

void HrbMainWindow::loadAppGuiSettings(const QSettings &s)
{
    BaseMainWindow::loadAppGuiSettings(s);
    statisticDialog->loadSettings(s);
}

void HrbMainWindow::updateGuiVal()
{
    ++blockControl;
    ui->spinBoxThreshold->setValue(config.getThreshold());
    ui->spinBoxLatency->setValue(config.defaultHrb.hrb.latency*8);
    ui->spinBoxWindow->setValue(config.defaultHrb.hrb.window*8);
    ui->spinBoxMinHitDuration->setValue(config.defaultHrb.hrb.minHitDuration*2);
    ui->actionAllow_Remote_Control->setChecked(config.allowRemoteControl);
    ui->actionShow_average_busy_time->setChecked(config.showAveragedBusy);
    --blockControl;
}

void HrbMainWindow::updateLabelRemoteState() const
{
    auto l = ui->labelUnderRemoteCtrl;
    if(!underRemoteCtrl) {
        l->hide();
        return;
    }
    QString txt;
    QColor c;
    switch (fsmState) {
    case FsmState::Starting:
        txt = QString("Starting run %1 %2...")
                .arg(runId.runIndex)
                .arg(runId.runNumber);
        c = 0xff9a3c;
        break;
    case FsmState::Run:
        txt = QString("Run started: %1 %2")
                .arg(runId.runIndex)
                .arg(runId.runNumber);
        c = Qt::darkGreen;
        break;
    case FsmState::Stopping:
        txt = QString("Stopping run %1 %2...")
                .arg(runId.runIndex)
                .arg(runId.runNumber);
        c = 0xff9a3c;
        break;
    default:
        txt = "Under remote control";
    }
    l->setText(txt);
    auto p = l->palette();
    p.setColor(l->foregroundRole(), c);
    l->setPalette(p);
    if(!l->isVisible())
        l->show();
}
