/*
**    Copyright 2013-2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "TtbDevice.h"

#include <QDateTime>
#include <QVector>

#include "LegacyDeviceModule.h"
#include "TTBRegs.h"
#include "UtcTimestampModule.h"
#include "lib-common/logDefinesPrefix.h"
#include "regio/regop.h"

static const quint32 link_mask = (1<<TTB_CH_COUNT)-1;

static quint32 getLinkMask(QMap<int, bool> chEn)
{
    quint32 mask=0;
    for (int i=0; i<TTB_CH_COUNT; i++)
        if(chEn[i])
            mask |= 1<<i;
    return mask;
}


TtbDevice::TtbDevice(const DeviceIndex &index, QObject *parent)
    : ModularMregDevice(index, status, parent)
    , status(index)
    , legacyDevice(createModule<LegacyDeviceModule>())
    , utcTimestampModule(createModule<UtcTimestampModule>())
{}

bool TtbDevice::writeConfig()
{
    bool ok = ModularMregDevice::writeConfig();
    if (!isConnected())
        return false;
    if (!isConfigCacheValid() || configCache.ttb != config.ttb) {
        ok &= legacyDevice->regWrite16(REG_TTB_CSR_LINK_EN, getLinkMask(config.ttb.chEn));
        //    regWrite(REG_TTB_RJTTC_ERRDISABLE_CTRL, link_mask);
    }
    clearStatistics();

    return ok;
}

bool TtbDevice::readTtbStatus()
{
    quint16 ttbStatus;
    RegOpVector op;
    op.RegRead16(REG_TTB_CSR_STATUS, &ttbStatus);
    op.RegRead32(REG32_TTB_CSR_SPILL_NUM, &status.spillNum);
    op.RegRead32(REG32_TTB_CSR_EVENT_NR, &status.eventNR);
    op.RegRead64(REG64_TTB_GLOB_EVENT_NUM, &status.globEvNum);

    op.RegRead16(REG_TTB_CSR_LINK_EN, &status.linkEn);
    op.RegRead16(REG_TTB_CSR_LINK_UP, &status.linkUp);
    op.RegRead16(REG_TTB_RJTTC_ERRDISABLE_CTRL, &status.errDisableEn);
    op.RegRead16(REG_TTB_CSR_LINK_ERRDISABLE, &status.linkErrdisable);
    op.RegRead16(REG_TTB_CSR_LINK_BUSY_BIT, &status.linkBusy);

    if (!legacyDevice->regOpExecRebased(op))
        return false;

//    qDebug() << "TTB status: " << hex << ttbStatus;
    status.ettc_clock_ok  = (ttbStatus & 0x1);
    status.ettc_ok        = (ttbStatus & 0x2);
    status.rjttc_clock_ok = (ttbStatus & 0x4);
    status.xoff           = (ttbStatus & 0x8);
    status.ettc_spill     = (ttbStatus & 0x10);
    if (!status.ettc_clock_ok){
//        sendLog("ETTC disconnected");
    } else
        if (!status.ettc_ok)
            logWarning << "ETTC error";
//    if (status.xoff)
//        sendLog("XOFF");
//    if (status.ettc_spill)
//        sendLog("ETTC SPILL");

    status.systemTs = utcTimestampModule->readClock();
//    const quint64 TS_FREQ = 125000000;
//    double system_ts_real = (double)system_ts / (double)TS_FREQ;
//    quint64 system_ts_s = (quint64)system_ts_real;
//    quint64 system_ts_us_part = 1000000 * (system_ts_real - system_ts_s);
//    QDateTime date = QDateTime::fromTime_t(system_ts_s);

//    qDebug() << QString("EVENT # %1, TS %2.%3")
//                .arg(event_nr, 10)
//                .arg(date.toString("yyyy-MM-dd hh:mm:ss"))
//                .arg(system_ts_us_part,6, 10, QChar('0'))
//            ;


    status.linkEn &= link_mask;
    status.linkUp &= link_mask;
    status.errDisableEn &= link_mask;
    status.linkErrdisable &= link_mask;
    status.linkBusy &= link_mask;

    //    QString str_link_en         = QString("Link en:       %1").arg(link_en, TTB_NCH, 2, QChar('0'));
//    QString str_link_up         = QString("Link up:       %1").arg(link_up, TTB_NCH, 2, QChar('0'));
//    QString str_errdisable_en   = QString("Errdisable en: %1").arg(errdisable_en, TTB_NCH, 2, QChar('0'));
//    QString str_link_errdisable = QString("Link errdis:   %1").arg(link_errdisable, TTB_NCH, 2, QChar('0'));
//    QString str_link_busy       = QString("Link busy bit: %1").arg(link_busy, TTB_NCH, 2, QChar('0'));

//    qDebug() << str_link_en;
//    qDebug() << str_link_up;
//    qDebug() << str_errdisable_en;
//    if (link_errdisable)
//        sendLog(str_link_errdisable);
//    qDebug() << str_link_busy;
    return true;
}


bool TtbDevice::readStatistics()
{
//    QMap<int, QString> labels;
//    for (int i=0; i<REG_TTB_COUNTERS_NUM; i++)
//        labels[i] = QString("[%1]").arg(i); // default
//    labels[0] = "ETTC_ERROR";
//    labels[1] = "TTC_TRIG";
//    labels[4] = "ETH_RX_ERROR";
//    labels[5] = "ETH_TX_ERROR";
//    labels[6] = "ETH_REGIO_ERROR";
//    labels[7] = "ETH_REGIO_TIMEOUT";
//    for (int i=0; i<TTB_NCH; i++)
//        labels[16+i] = QString("LINK_UP[%1]").arg(i);
//    for (int i=0; i<TTB_NCH; i++)
//        labels[32+i] = QString("LINK_ERROR[%1]").arg(i);
//    for (int i=0; i<TTB_NCH; i++)
//        labels[48+i] = QString("LINK_TIMEOUT[%1]").arg(i);
//    for (int i=0; i<TTB_NCH; i++)
//        labels[64+i] = QString("LINK_ERRDISABLE[%1]").arg(i);

    TtbStatistic &statistic = status.statistic;

    RegOpVector op;

    op.RegRead32(REG_TTB_COUNTERS_BASE + REG32_TTB_ETTC_ERROR, &statistic.ettcError);
    op.RegRead32(REG_TTB_COUNTERS_BASE + REG32_TTB_TTC_TRIG, &statistic.ttcTrig);
    op.RegRead32(REG_TTB_COUNTERS_BASE + REG32_TTB_ETH_RX_ERROR, &statistic.ethRxError);
    op.RegRead32(REG_TTB_COUNTERS_BASE + REG32_TTB_ETH_TX_ERROR, &statistic.ethTxError);
    op.RegRead32(REG_TTB_COUNTERS_BASE + REG32_TTB_ETH_REGIO_ERROR, &statistic.ethRegioError);
    op.RegRead32(REG_TTB_COUNTERS_BASE + REG32_TTB_ETH_REGIO_TIMEOUT, &statistic.ethRegioTimeout);


    for(int ch=0;ch<TTB_CH_COUNT;++ch) {
        auto offset = REG_TTB_COUNTERS_BASE + ch*2;
        op.RegRead32(offset + REG32_TTB_STAT_LINK_UP, &statistic.linkUp[ch]);
        op.RegRead32(offset + REG32_TTB_STAT_LINK_ERROR, &statistic.linkError[ch]);
        op.RegRead32(offset + REG32_TTB_STAT_LINK_TIMEOUT, &statistic.linkTimeout[ch]);
        op.RegRead32(offset + REG32_TTB_STAT_LINK_ERR_DISABLE, &statistic.linkErrDisable[ch]);
        op.RegRead32(offset + REG32_TTB_STAT_LINK_ERR_DISABLE_SET, &statistic.linkErrDisableBusySet[ch]);
        op.RegRead32(offset + REG32_TTB_STAT_LINK_ERR_DISABLE_UNSET, &statistic.linkErrDisableBusyUnset[ch]);
        op.RegRead32(offset + REG32_TTB_STAT_LINK_ERR_DISABLE_MSTREAM, &statistic.linkErrDisableBusyMStream[ch]);
    }

    op.RegRead64(REG64_TTB_CNT_SPILL, &statistic.spillCnt);

    // busy counters
    op.RegRead64(REG64_TTB_CNT_BUSY_TTB, &statistic.ttbBusyCnt);
        for (int i=0; i<TTB_CH_COUNT; i++) {
        op.RegRead64(REG64_TTB_CNT_BUSY_CH0+i*4, &statistic.chBusyCnt[i]);
    }

    if (!legacyDevice->regOpExecRebased(op)) {
        qWarning() << "Failed to read TTB statistics";
        return false;
    }

    return true;
}


void TtbDevice::clearStatistics()
{
    RegOpVector r;
    r.RegWrite16(REG_CONTROL, 1);
    r.RegWrite16(REG_CONTROL, 0);
    if (!legacyDevice->regOpExecRebased(r))
        logDebug << "clearStatistics";
    else
        logCritical << "failed to clear statistic";
}

void TtbDevice::softReset()
{
    qWarning() << __PRETTY_FUNCTION__<<"todo";
}

void TtbDevice::evNumRequest(int checkIndex)
{
    qWarning() << __PRETTY_FUNCTION__<<"todo";
}

bool TtbDevice::readStatus()
{
    if (!ModularMregDevice::readStatus())
        return false;
    if(!isConnected()) return false;

    if (!readTtbStatus()) return false;
    if (!readStatistics()) return false;

    emit statusUpdated(status);
    return true;
}

void TtbDevice::on_deviceConnected()
{
    ModularMregDevice::on_deviceConnected();
    utcTimestampModule->setClock();

    status.fw = getDeviceFwVersion();
    status.fwStr = getFirmwareVersionStr();
    status.hostStr = getAddressStr();
    emit connectedStateChanged(getDeviceIndex(), true);
    logInfo << QString("Device online (%1) fw=%2")
               .arg(getAddressStr()).arg(getFirmwareVersionStr());
}

void TtbDevice::on_deviceDisconnected()
{
    ModularMregDevice::on_deviceDisconnected();

    emit connectedStateChanged(getDeviceIndex(), false);
    logWarning << QString("Device offline");
}
