#include "EvbGenCore.h"
#include "GenServerConfig.h"
#include "GenServer.h"
#include "GenStartupOptions.h"
#include "EvbGenConfigDb.h"

#include "base/RunDocument.h"
#include "base/RunDocumentConverter.h"
#include "base/Globals.h"
#include "base/ThreadController.h"

#include <QCoreApplication>
#include <QCryptographicHash>
#include <QDataStream>
#include <QDebug>
#include <QElapsedTimer>
#include <QFile>
#include <QJsonDocument>
#include <iostream>
#include <QStringList>
#include <stdio.h>
#include <QObject>
#include <QTimer>
#include <QThread>
#include <limits>
#include <math.h>
#include "EvbGenSignalRouter.h"
#include "EvbGenConfigDispatcher.h"

const int LOOP_TIME_LIMIT = 500;
const int WHOLE_TIMEOUT_MS = 15000;
namespace {
QString humanReadableSize(double size)
{
    static constexpr const char* list[] = {
        "B", "KiB", "MiB", "GiB", "TiB"
    };
    static constexpr const size_t n = sizeof(list)/sizeof(char*);

    size_t i= 0;
    for(; i<n-1; ++i) {
        if(size<1024)
            break;
        size /= 1024;
    }
    return QString("%1 %2").arg(size,0,'f',2).arg(list[i]);
}
}

EvbGenCore::EvbGenCore(const GenStartupOptions &start_opt, QObject *parent)
    : BaseAppCore(parent)
    , statusTimer(new QTimer)
    , configDb(new EvbGenConfigDb(this))
    // , configDispatcher(new EvbGenConfigDispatcher(this))
    , fsm(new EvbGenFsm(this))
{
    qRegisterMetaType<ProgramDescription>();

    programIndex = Globals::getInitialProgramIndex();

    QString msg;
    // if(!generator_config.read_config()){
    //     msg = QString("Read configuration ") + "\033[91m" + QString("fail") + "\033[0m";
    //     add_log(msg);
    //     return;
    // }
    // else {
    //     msg = QString("Read configuration ") + "\033[32m" + QString("success") + "\033[0m";
    //     add_log(msg);
    // }

    // setEvbGenConfig(start_opt.cfgBase);

    //---status manager---
    connect(statusTimer, &QTimer::timeout, this, &EvbGenCore::printStats);
    statusTimer->setInterval(1e3);

    //--- prepare program description config---
    prog_desc.options["devSer"] = "Generator-" + QString::number(random_between(0,1000));
    prog_desc.options["devId"] = "Generator-" + QString::number(random_between(0,1000));

    //---TCPServer connections---
    qRegisterMetaType<GenServerStatus>();
    qRegisterMetaType<GenServerConfig>();


    auto &signalRouter = EvbGenSignalRouter::instance();
    signalRouter.setCore(this);
    signalRouter.setMainFsm(fsm);
    signalRouter.setEvbGenConfigDb(configDb);
    signalRouter.connectAll();

    auto &configDispatcher = EvbGenConfigDispatcher::instance();
    // configDispatcher.subscribeCS(this);
    configDispatcher.subscribeCS(configDb);
    configDispatcher.subscribeEvbGenConfig(this);
    configDispatcher.subscribeEvbGenConfig(configDb);

    //---prepare qrand---
    QDateTime current_time_for_qrand = QDateTime::currentDateTime();
    int seedBase = current_time_for_qrand.currentMSecsSinceEpoch();
    qsrand(seedBase);

    QTimer::singleShot(0, this, &EvbGenCore::init);

    // QTimer::singleShot(0, [=, this]() {
    //     setClientsCount(start_opt.numberOfClients);
    // });
 }

EvbGenCore::~EvbGenCore()
{
    add_log("~Generator()");

    emit closeAll();
    emit closeProgram(prog_desc);
}

void EvbGenCore::setClientsCount(int cnt)
{
    if(clientsCount == cnt)
        return;

    for(auto i=clientsCount; i<cnt; ++i)
        addGenServer(i);
    while(clientsCount>cnt) {
        auto lastId = --clientsCount;
        emit deleteGen(lastId);
    }

    fsm->setGenNum(clientsCount);
}

void EvbGenCore::setEvbGenConfig(const EvbGenConfig &newConfig)
{
    if(config == newConfig)
        return;
    config = newConfig;
    if(!check_config(config.cfgBase)) {
        qFatal("check_config failed");
        auto msg = QString("Check configuration ") + "\033[91m" + QString("fail") + "\033[0m";
        add_log(msg);
    }
    setClientsCount(config.numberOfClients);
}

void EvbGenCore::performLocalStart()
{
    if(fsm->getState() != FsmState::Ready) {
        const auto &msg = QString("Application isn't in '%1' state (curState='%2')")
        .arg(fsmStateToString(FsmState::Ready))
            .arg(fsmStateToString(fsm->getState()));
        qWarning() << msg;
        return;
    }

    emit startGen(rcRunId, config.cfgBase);
    fsm->setRequestState(FsmState::Starting);
}

void EvbGenCore::performLocalStop()
{
    if(fsm->getState() != FsmState::Run) {
        const auto &msg = QString("Application isn't in '%1' state (curState='%2')")
        .arg(fsmStateToString(FsmState::Run))
            .arg(fsmStateToString(fsm->getState()));
        qWarning() << msg;
        return;
    }

    emit requestWorkCompleted();
    fsm->setRequestState(FsmState::Stopping);
}

void EvbGenCore::onRemoteStartRequest(const RcRunId &runId_, const QVariant &startEvNum)
{
    if(fsm->getState() != FsmState::Ready) {
        const auto &msg = QString("Application isn't in '%1' state (curState='%2')")
                              .arg(fsmStateToString(FsmState::Ready))
                              .arg(fsmStateToString(fsm->getState()));
        emit sendRunWarning(rcRunId, msg);
        return;
    }

    qInfo() << "=== Start request" << " runId:" << runId_ << " startEvNum:" << startEvNum;
    rcRunId = runId_;

    emit startGen(rcRunId, config.cfgBase);
    fsm->setRequestState(FsmState::Starting);
}

void EvbGenCore::onRemoteStopRequest(const RcRunId &runId_, const QVariant &stopEvNum)
{
    // return;
    bool success=true;
    if(rcRunId.runIndex != runId_.runIndex) {
        qWarning().noquote()
            << QString("Got stop request with wrong run index: runIndex (%1); remote control runIndex (%2)")
                   .arg(rcRunId.runIndex)
                   .arg(runId_.runIndex);
        success = false;
    }
    if(success && rcRunId.runNumber != runId_.runNumber){
        qWarning().noquote()
            << QString("Got stop of unknown run: runNumber (%1); remote control runNumber (%2)")
                   .arg(rcRunId.runNumber)
                   .arg(runId_.runNumber);
        success = false;
    }

    if(fsm->getState() != FsmState::Run) {
        const auto &msg = QString("Application isn't in '%1' state (curState='%2')")
                              .arg(fsmStateToString(FsmState::Run))
                              .arg(fsmStateToString(fsm->getState()));
        emit sendRunWarning(rcRunId, msg);
        if(fsm->getState() != FsmState::Ready && fsm->getState() != FsmState::Error)
            fsm->setRequestState(FsmState::Error);
        return;
    }

    emit requestWorkCompleted();
    fsm->setRequestState(FsmState::Stopping);
}

void EvbGenCore::onFsmStateChange(const FsmState &state, const FsmState &oldState)
{
    switch (state) {
    case FsmState::Ready:
        statusTimer->stop();
        break;
    case FsmState::Starting:
        fsm->setRequestState(FsmState::Run);
        statusTimer->start();
        break;
    case FsmState::Run:
        sendRunDocument();
        break;
    case FsmState::Stopping:
        fsm->setRequestState(FsmState::Ready);
        break;
    default:;
    }
}

void EvbGenCore::genStatusUpdated(const GenServerStatus &st)
{
    getStats[st.id] = st;
}

void EvbGenCore::init()
{
    initEvbGenProgDescr();

    emit allowRemoteControl(true);
}

// void EvbGenCore::remoteStatusUpdated(const RemoteControlServerStatus &status)
// {

// }

// void GeneratorCore::onFsmStateChange(const FsmState &state, const FsmState &oldState)
// {

// }

bool EvbGenCore::isRun() const
{
    return fsm->getState() == FsmState::Run;
}

const BaseFsm *EvbGenCore::getFsm() const
{
    return fsm;
}

void EvbGenCore::initEvbGenProgDescr()
{
    progDescr.setClientIndex({Globals::getDaqModule(), programIndex});
    emit updateProgramDescription(progDescr);
}

// void GeneratorCore::prepare_fire_plan()
// {
//     fire_tact = 500;
//     fire_number = 0;
//     events_in_fire = (generator_config.spill_events * fire_tact) / generator_config.spill_size;
//     fires_in_spill = generator_config.spill_events / fire_tact;
//     total_events = 0;

//     for(int i=0; i<225000; ++i){
//         data.append(i);
//     }

//     data_size = data.size();

// //    config.debug_print();
// //    qInfo() << "=== fire number: " << fire_number;
// //    qInfo() << "=== events_in_fire: " << events_in_fire;
// //    qInfo() << "=== fires_in_spill: " << fires_in_spill;
// }

// void GeneratorCore::work()
// {
//     qInfo() << "==== work";
//     if ((status == "READY") && (fire_number == 0)){
//         status = "STARTED";
//         emit sendState("STARTED", true);
//         emit completedStart(true, generator_config.run_index, generator_config.run_number);
//         emit config_transfer_to_rc(generator_config.root.get_json().toBinaryData());
//         qInfo() << "=== send state STARTED";
//     }

// //    if(status == "STOPPED"){
// //        status = "STOPPED";
// //        emit sendState("STOPPED", true);
// ////        QCoreApplication::quit();
// ////        return;
// //    }

//     sleep(2);
//     emit sendState("STARTED", true);
//     emit completedStart(true, generator_config.run_index, generator_config.run_number);

//     QDateTime fire_time_befor = QDateTime::currentDateTime();
//     qInfo() << "Fire  before " << fire_number << fire_time_befor.toString("mm ss zzz");

//     for(int i=0; i< 300000; ++i){
//         sendEvent();
//     }

//     QDateTime fire_time_after = QDateTime::currentDateTime();
//     qInfo() << "Fire number " << fire_number << fire_time_after.toString("mm ss zzz");
//     qInfo() << "=== data_size: " << data_size;
//     qInfo() << "total_events: " << total_events;
//     qInfo() << "total_size: " << total_size;

// //    emit sendState("STOPPED", true);

// //    qInfo() << "fire_tact: " << fire_tact;
//     qInfo() << "fire interval: " << fire_time_befor.msecsTo(fire_time_after);

//    emit completedStop(true, run_index, run_number);

//    config->get_from_db();
//    qint64 stop_count = eventsInFire +(config.spill_count * config->eventsInSpill);
//    for(qint64 i=evNumber; i<stop_count; ++i){
//        sendEvent();
//        ++evNumber;
//    }

//    //---work in spill---
//    if(fire_number < fires_in_spill){
//        // timer
//        QDateTime fire_time = workTime.addMSecs(fire_number*10 + config.spill_count*spill_period + 10);
//        QDateTime current_time = QDateTime::currentDateTime();
//        if(current_time.msecsTo(fire_time) >= 0){
//            common_timer->setInterval(current_time.msecsTo(fire_time));
//        }
//        else{
//            common_timer->setInterval(0);
//        }
//        common_timer->start();
//    }

//    //---work in interspill---
//    if((fire_number == fires_in_spill)&&((config.spill_count+1) < config->numberOfSpills)){
//        QDateTime spill_end_time = workTime.addMSecs(numberOfSpills_count * spill_period + config->spillSize);
//        QDateTime current_time = QDateTime::currentDateTime();
//        int spill_late = current_time.msecsTo(spill_end_time);
//        if (spill_late >= 0){
//            addLog(QString("Spill #%1/%2; events=%3 ")
//                   .arg(numberOfSpills_count +1).arg(config->numberOfSpills).arg(evNumber));
//            common_timer->setInterval(config->interSpillSize + spill_late);
//        }
//        else{
//            spill_late = spill_late * (-1);
//            addLog(QString("Spill #%1/%2; events=%3; was late on %4 msecs")
//                   .arg(numberOfSpills_count +1).arg(config->numberOfSpills).arg(evNumber).arg(spill_late));
//            common_timer->setInterval(config->interSpillSize - spill_late);
//        }
//        ++numberOfSpills_count;
//        fire_number = 0;
//        common_timer->start();
//        emit sendState(QString("EvNum %1").arg(evNumber), true);
//    }

//    //---end of work---
//    if((fire_number == fires_in_spill) && (numberOfSpills_count+1 == config->numberOfSpills)){
//        QDateTime spill_end_time = workTime.addMSecs(numberOfSpills_count * spill_period + config->spillSize);
//        QDateTime current_time = QDateTime::currentDateTime();
//        int spill_late = current_time.msecsTo(spill_end_time);
//        if (spill_late >= 0){
//            addLog(QString("Spill #%1/%2; events=%3 ")
//                   .arg(numberOfSpills_count +1).arg(config->numberOfSpills).arg(evNumber));
//        }
//        else{
//            spill_late = spill_late * (-1);
//            addLog(QString("Spill #%1/%2; events=%3; was late on %4 msecs")
//                   .arg(numberOfSpills_count +1).arg(config->numberOfSpills).arg(evNumber).arg(spill_late));
//        }
//        addLog(QString("Work is done correctly. Total events: %1").arg(evNumber));
//        emit sendState(QString("EvNum %1").arg(evNumber), true);
//        requestWorkCompleted();
//    }
//    ++fire_number;
// }

void EvbGenCore::sendEvent()
{
    // //---prepare qrand---
    // device_hdr.length = data_size;
    // event_hdr.length = sizeof(MpdDeviceHdr) + data_size;
    // ++event_hdr.evNum;

    // gen_server->addData((char *)&event_hdr, sizeof(MpdEventHdr));
    // gen_server->addData((char *)&device_hdr, sizeof(MpdDeviceHdr));
    // gen_server->addData(data, data_size);

    // total_size += sizeof(MpdEventHdr) + event_hdr.length;
    // total_events += 1;
}

void EvbGenCore::sendRunDocument() const
{
    auto cfgObj = QVariant::fromValue(config).value<QJsonObject>();
    RunDocument rd(rcRunId, configDb->getClientIndex(), cfgObj);
    emit sendRunDocumentToRc(rd);
}

void EvbGenCore::addGenServer(int id)
{
    auto gen_server = new GenServer(id, programIndex.getIndex());

    connect(this, &EvbGenCore::closeAll, gen_server, &GenServer::deleteLater);
    connect(this, &EvbGenCore::deleteGen, gen_server, &GenServer::deleteGen);
    connect(this, &EvbGenCore::startGen, gen_server, &GenServer::startGen);
    connect(this, &EvbGenCore::requestWorkCompleted, gen_server, &GenServer::requestWorkCompleted);
    connect(gen_server, &GenServer::statusUpdated, this, &EvbGenCore::genStatusUpdated);

    connect(gen_server, &GenServer::updateProgramDescription, this, &EvbGenCore::updateProgramDescription);
    connect(gen_server, &GenServer::closeProgram, this, &EvbGenCore::closeProgram);

    connect(gen_server, &GenServer::fsmStateChanged, fsm, &EvbGenFsm::setGenState);
    if(Globals::getDebugMode()) {
        auto &configDispatcher = EvbGenConfigDispatcher::instance();
        connect(&configDispatcher, &EvbGenConfigDispatcher::evbGenServerConfigChanged,
                gen_server, &GenServer::setEvbGenServerConfig);
    }

    gen_server->setDropDataOnFreeSocket(true);
    new ThreadController(gen_server, this);
    ++clientsCount;
}

//---dont used slots yet---
int EvbGenCore::random_between(int low, int high)
{
    QDateTime current_time= QDateTime::currentDateTime();
    uint seedBase = current_time.currentMSecsSinceEpoch();
    qsrand(seedBase);
    return (qrand() % ((high + 1) - low) + low);
}

bool EvbGenCore::check_config(const GenServerConfig &cfg) const
{
    if(cfg.spill_count <= 0){
        add_log("Not valid configuration: spill count <= 0");
        return false;
    }
    if(cfg.spillMs <= 0){
        add_log("Not valid configuration: spill size <= 0");
        return false;
    }
    if(cfg.interspillMs < 0){
        add_log("Not valid configuration: interspill size < 0");
        return false;
    }
    if(cfg.getEvCntPerSpill() < 0){
        add_log("Not valid configuration: spill size <= 0");
        return false;
    }
    if(cfg.event_length <= 0){
        add_log("Not valid configuration: event length <= 0");
        return false;
    }
    return  true;
}

void EvbGenCore::add_log(QString log) const
{
    qInfo()<<log.toStdString().c_str();
}

void EvbGenCore::printStats()
{
    if(fsm->getState() != FsmState::Run)
        return;

    double runProgress = 100;
    quint32 runEvNum = config.cfgBase.spill_count * config.cfgBase.getEvCntPerSpill();
    // qInfo() << '\n' << __PRETTY_FUNCTION__;
    quint64 minEvNum = -1;
    for(auto i=getStats.cbegin(); i!=getStats.cend(); ++i) {
        minEvNum = std::min(minEvNum, i->total_events);
        runProgress = std::min(runProgress, 100.*i->total_events/runEvNum);
        continue;

        QStringList str;
        str << QString("[%1]").arg(i->id, 2, 10, QChar('0'));
        str << "loop:" << QString("%L1").arg(i->workLoop);
        str << "total_events:" << QString("%L1").arg(i->total_events);

        str << "total_size:" << humanReadableSize(i->total_size);
        str << "numberBytesSent:" << humanReadableSize(i->numberBytesSent);

        if(i->numberBytesDropped)
            str << "numberBytesDropped:" << QString("%L1").arg(i->numberBytesDropped);
        qInfo() << str.join(' ');
    }
    emit runProgressChanged(minEvNum, runEvNum, runProgress);
}

// trash
//if((config.spill_size % 10 != 0) ||
//  (config.interspill_size % 10 !=0)){
//    addLog("Spill or interSpil size not a multiple of 10 ");
//    return;
//}
//if (config.spill_events == 0){
//    addLog("0 events in spill");
//    return;
//}
//if (config.spill_count == 0){
//    addLog("0 spills");
//    return;
//}

//    //--- check generator config---;

//    emit status_manager_signal("init ready");

    //---prepare work time - for debug---
//    QDateTime current_time = QDateTime::currentDateTime();
//    workTime = current_time.addSecs(2);
//    QString workTime_str = workTime.toString("yyyy.MM.dd.hh.mm.ss");
//    status = "READY";
//    emit timeToStart(workTime_str);

//    QTimer *t = new QTimer();
//    t->setInterval(3000);
//    t->start();
//    connect(t, &QTimer::timeout, this, &Generator::work, Qt::QueuedConnection);
//    connect(t, &QTimer::timeout, this, &Generator::work, Qt::DirectConnection);

//    QTimer::singleShot(3000, this, &Generator::work);



//    emit closeProgram(prog_desc);
//    deleteLater();

    /*
    emit closeProgram(progDesc);
    emit deleteAll();

    if(common_timer){
        delete common_timer;
        common_timer = nullptr;
    }

    QElapsedTimer timer;
    timer.start();

    if(!threadQuiting(pnpServerThread, WHOLE_TIMEOUT_MS, &timer))
        qWarning()<<"Error on closing PNP Server";
    if(!threadQuiting(tcpServerThread, WHOLE_TIMEOUT_MS, &timer))
        qWarning()<<"Error on closing TCP Server";
    if(!threadQuiting(remoteControlThread, WHOLE_TIMEOUT_MS, &timer))
        qWarning()<<"Error on closing TCP Server";
        */


//void Generator::timeToStart(QString str)
//{
//    if(status != "READY"){
////        addLog(QString("Generators %1 status is not READY, and will be finished:").arg(argument));
//        foreach (QString str, status_manager.keys()) {
//            qInfo() << str << status_manager[str];
//            if((str == "init")&&(status_manager[str] == "READY")){
//                 addLog("Initialization was done correctly");
//            }
//            if((str == "init")&&(status_manager[str] == "not ready")){
//                 addLog("Initialization fails");
//            }
//            if((str == "tcp")&&(status_manager[str] == "READY")){
//                 addLog("TCP client is ready");
//            }
//            if((str == "tcp")&&(status_manager[str] == "not ready")){
//                 addLog("TCP client fails");
//            }
//            if((str == "rc")&&(status_manager[str] == "READY")){
//                 addLog("Remote control is ready");
//            }
//            if((str == "rc")&&(status_manager[str] == "not ready")){
//                 addLog("Remote control fails");
//            }
//        }
//        return;
//    }

//    fire_number = 0;
//    fires_in_spill = (config->spillSize / 10) - 1;

//    numberOfSpills_count = 0;
//    spill_period = config->spillSize + config->interSpillSize;
//    evNumber = 0;
//    eventsIn10Msec = ((double) 10. * config->eventsInSpill) / config->spillSize;

//    QDateTime current_time = QDateTime::currentDateTime();
//    workTime = QDateTime::fromString(str, "yyyy.MM.dd.hh.mm.ss");
//    int msecs_to = current_time.msecsTo(workTime);

//    common_timer->setSingleShot(true);
//    common_timer->setInterval(msecs_to);
    //    common_timer->start();
//}
