#include "Combiner.h"

#include "params.h"
#include "Analizer.h"
#include "EvbSignalRouter.h"
#include "EvbTlvTypes.h"
#include "FileWriter.h"
#include "FileWriterStats.h"
#include "OutputTcpServer.h"
#include "RingBuffer.h"
#include "TcpClient.h"

#include "base/DeviceId.h"
#include "base/SoftwareDevices.h"
#include "base/RunDocumentConverter.h"
#include "base/threadQuit.h"

#include <QElapsedTimer>
#include <QJsonDocument>
#include <QSemaphore>
#include <QSharedPointer>
#include <QThread>
#include <QTimer>
#include "CombinerFsm.h"

namespace {
    const int OUTPUT_TCP_PORT = 0; // 33399 or 0 for auto-detect
    const quint64 LOOP_TIMEOUT = 100;

    const int START_STOP_TIMEOUT_MS = 1000;
    const int SUPPRESS_LOG_LAGGING_LIMIT = 10;
    const int EV_INCOMPLET_LIMIT = 10;

    QByteArray allineData(const QByteArray &data)
    {
        auto allinedData(data);
        if(int rest = (allinedData.size()%4))
            allinedData.append(4-rest, 0);
        return allinedData;
    }
}

Q_DECLARE_METATYPE(QVector<Analizer*>)

Combiner::Combiner()
    : CombinerFsm()
    , workLoopTimer(new QTimer(this))
    , statusTimer(new QTimer(this))
    , writerRingBuf(new RingBuffer<char>(0))
    , fileWriter(new FileWriter(writerRingBuf))
    , fileWriterThread(new QThread(this))
{
    setObjectName(__func__);
    AbstractConfigConverter<RunDocument>::registerTypes();
    qRegisterMetaType<CombinerStats>();
    qRegisterMetaType<COMBINER_OUT>();
    qRegisterMetaType<OutputTcpStatus>();
    qRegisterMetaType< QVector<Analizer*> >();
    qRegisterMetaType< QMap<int, QString> >();
    connect(workLoopTimer, &QTimer::timeout, this, &Combiner::workLoop);
    workLoopTimer->setSingleShot(true);
    workLoopTimer->setInterval(0); // it's ok, work can't completed faster then workTimeoutMs

    fileWriterThread->setObjectName("fileWriterThread");
    fileWriter->moveToThread(fileWriterThread);
    fileWriterThread->start();
    connect(fileWriter, &FileWriter::destroyed, fileWriterThread, &QThread::quit);
    connect(fileWriter, &FileWriter::statusUpdated, this, &Combiner::fileWriterStatsUpdated);
    connect(fileWriter, &FileWriter::statusUpdated, this, &Combiner::onFileWriterStatsUpdated);
    connect(fileWriter, &FileWriter::fileOpened, this, &Combiner::writerFileOpened);
    connect(fileWriter, &FileWriter::error, this, &Combiner::outputError);
    connect(this, &Combiner::resetOutputs, fileWriter, &FileWriter::clear);
    connect(this, &Combiner::deleteFileWriter, fileWriter, &FileWriter::deleteLater);
    connect(this, &Combiner::startFileRun, fileWriter, &FileWriter::startRun);
    connect(this, &Combiner::closeFileRun, fileWriter, &FileWriter::closeRun);

    connect(this, &BaseFsm::stateChanged, this, &Combiner::onFsmStateChanged);

    connect(statusTimer, &QTimer::timeout, this, &Combiner::sendStatus);
    statusTimer->setSingleShot(false);
    statusTimer->setInterval(1000);

    connect(this, &Combiner::requestForFileSplit,
            fileWriter, &FileWriter::requestForSplit);
    connect(fileWriter, &FileWriter::fileSplitted,
            this, &Combiner::fileSplitted);

    auto evbSignalRouter = &EvbSignalRouter::instance();
    evbSignalRouter->setCombiner(this);
    evbSignalRouter->setCombinerFsm(this);
    evbSignalRouter->setFileWriter(fileWriter);

    QTimer::singleShot(0, this, &Combiner::init);
}

Combiner::~Combiner()
{
    if(enabled)
        setRequestState(FsmState::Ready);

    QElapsedTimer timer;
    timer.start();
    emit deleteFileWriter();
    emit deleteTcp(0);

    bool ok=true;
    for (int index : outputTcpServers.keys()) {
        if(!threadQuiting(outputTcpServerThreads[index], 1000, &timer)){
            qWarning() << QString("Failed to stop output TcpServer id=%1").arg(index);
            ok=false;
        }
    }
    if(!threadQuiting(fileWriterThread, 1000, &timer)) {
        qWarning() << "Failed to stop FileWriter thread";
        ok=false;
    }
    if(ok)
        qDebug() << "All parts of Combiner were stoped.";
    else
        qWarning() << "Some output TcpServers failed to stop.";
    delete writerRingBuf;
}

void Combiner::initOutputBuffers(const EvbCombinerConfig &cfg)
{
    int size = cfg.outputBufSize;
    preferedBufSize = (size<=0) ? FILE_WRITER_BUFFER_SIZE : size;
}

void Combiner::onFsmStateChanged(const FsmState &cur, const FsmState &prev)
{
    switch (cur) {
    case FsmState::Run:
    {
        qDebug() << __PRETTY_FUNCTION__ << "enablinig combiner";
        enabled = true;
        workLoopTimer->start();
        qDebug() << "Combiner is started.";

        stats.period.newPeriod();
        break;
    }
    case FsmState::Error:
    {
        emit resetOutputs();
        if((prev == FsmState::Run || prev == FsmState::PrepareStopping) && outToFile) {
            QString errStr = QString("Can't wait end of work of file writer; there are some data pending (%L1 bytes)")
                    .arg(writerRingBuf->getContinuosDataSize(0));
            qWarning() << errStr;
            emit error(errStr);
        }
        if(prev != FsmState::Ready && prev != FsmState::Init)
            onStoppingAction();
        break;
    }
    default:;
    }
}

void Combiner::setCombRequestState(FsmState state, const QVariant &arg)
{
    switch (state) {
    case FsmState::Run:
        runId = arg.value<RcRunId>();
        qDebug() << QString("Combiner starting run %1").arg(runId.toString());
        break;

    case FsmState::PrepareStopping:
        setStopEvNum(arg.value<quint64>());
        setExpectedEvNum(expectedEvNum);
        break;
    default:;
    }
    setRequestState(state);
}

void Combiner::init()
{
    stats.period.newPeriod();
    int monitorId = addTcpServer();
    Q_ASSERT(monitorId==0);
    monitorTcpServer = outputTcpServers.first();
    resizeBuffer();
    emit startTcpServer(monitorId, true);
    // tcpStatusUpdated from monitor TCP will invoke this method
//    sendStatus();
}

void Combiner::onStartingAction()
{
    startRecordWritten = false;
    runDocumentWritten = false;
    if(outToFile) {
        fileSplitId = 0;
        auto writerIndex = getWriterIndex();
        std::optional<int> splitId;
        if(splitConfig.hasAnyEn())
            splitId = fileSplitId;
        emit startFileRun(runId, writerIndex, splitId);
    } else if(outToTcpServer) {
        emit startingTcpWork();
    }

    statusTimer->start();
    sendStatus();
}

void Combiner::reset()
{
    resetEventIncomplet();
    for(auto &d : ccds) {
        d.evMI = nullptr;
        d.suppresLimit = ClientCombinerData::SuppresLimitMax;
    }
    warnEvNumStepFitFail = true;
    warnEvNumShift = true;
    emit resetOutputs(); // todo: add feedback

#ifdef RESET_ROUGH
    // All work with RingBuffers performed in Analizer and TcpClient
#else
    for (RingBuffer<EventMetaInfo> *rb : eventRingBuffers) {
        int eventsPending = rb->getDataSize();
        rb->trashData(eventsPending);
    }
#endif
    for(auto &ccd : ccds)
        ccd.rbDataSpace = 0;

    initFlag=true;
    Q_ASSERT(!enabled); // enabled = false;
    expectedEvNum = evNumberCfg.startEvNum;
    stats.clear();

    sendStatus();
}

void Combiner::destroyClient(const ClientIndex &clientId)
{
    if(!clientId.isEmpty()) {
        ccds.remove(clientId);
        clientIds.remove(clientId);
        nonEvBasedClients.remove(clientId);
    }
    updateClientSizeCache();
}

void Combiner::setCombinerConfig(const EvbCombinerConfig &cfg)
{
    if(combCfg == cfg)
        return;
    combCfg = cfg;

    if(preferedBufSize != cfg.outputBufSize) {
        waitingResize = true;
        preferedBufSize = cfg.outputBufSize;
        resizeBuffer();
    }

    CombinerFsm::setOutputList(cfg);

    outToFile = cfg.combinerOut==COMBINER_FILE;
    outToTcpServer = cfg.combinerOut==COMBINER_TCP;

    auto tcpNum = outToTcpServer ? cfg.combinerOutTcpNum : 0;
    for(int i=outputTcpServers.size(); i<tcpNum+1; ++i) {
        int id = addTcpServer();
        emit startTcpServer(id, true);
        qInfo() << QString("Add out TCP id:%1").arg(id);
    }
    for(int i=outputTcpServers.size()-1; i>tcpNum; --i ) {
        qInfo() << QString("Remove out TCP id:%1").arg(i);
        emit deleteTcp(i);
        if(!threadQuiting(outputTcpServerThreads[i], 1000)){
            qWarning() << QString("failed to stop output TcpServer id=%1").arg(i);
            outputTcpServerThreads[i]->terminate();
        }
        stats.tcpStatus.remove(i);
        outputTcpServers.remove(i);
        outputTcpServerThreads.remove(i);
    }
    emit infOutputUpdated(cfg.infOutputEn);
    resizeBuffer();
    sendStatus();
}

void Combiner::workLoop()
{
    if(eventsWithDebug)
        qInfo()<<"[Combiner] start of workLoop";

    if(waitingResize)
        resizeBuffer();

    if(!enabled)
        return;

    if(!isRunRecWritten()){
        workLoopTimer->start();
        return;
    }
    writeRunDocument();

    if(checkFileSplitting()){
        workLoopTimer->start();
        return;
    }

    QElapsedTimer eventMinTime, loopTimer, workTimer;
    const qint64 workTimeoutMs = getState() == FsmState::PrepareStopping ? 100 : 1000;
    const qint64 loopTimeoutMs = getState() == FsmState::PrepareStopping ? 10 : 100;
    qint64 leftTimeMs;
    bool loopExpired, eventExpired;
    int gotClients;

    workTimer.start();
    if(initFlag){
        processInit();
        if(initFlag) {
            // process other pending signals and try again
            workLoopTimer->start();
            return;
        }
    }
    if(stats.period.isNextPeriod(workTimer)) {
        emit evSizePeriodUpdated(stats.period);
        stats.period.newPeriod();
    }

    while(!initFlag && !workTimer.hasExpired(workTimeoutMs)) {
        if(splitInProgress)
            break;
        if(eventsWithDebug)
            qInfo()<<"[Combiner] while loop";

        eventMinTime.invalidate();
        loopTimer.start();
        loopExpired = false;
        eventExpired = false;

        //quick check eventMinTime
        for(auto id : clientIds) {
            auto &ccd = ccds[id];
            while(updateClientsEventInfo(&ccd)) {
                auto curEv = ccd.evMI;
                if(curEv->evNum<expectedEvNum) {
                    if(eventsWithDebug) {
                        qInfo() << "trash with lower evNumber=" << curEv->evNum
                                << " expectedEvNum:" << expectedEvNum
                                << "clientId:" << id;
                    }
                    if(suppressLogLagging < SUPPRESS_LOG_LAGGING_LIMIT) {
                        qWarning()<<QString("1.Trash lagging event evNum(%1)<expected(%2) from %3")
                                    .arg(curEv->evNum).arg(expectedEvNum).arg(id.toString());
                        if(++suppressLogLagging == SUPPRESS_LOG_LAGGING_LIMIT)
                            qWarning() << "suppressLogLagging";
                    }
                    ccd.removeEv();
                    stats.clientsStat[id].trashedEvents++;
                    continue;
                }
                if(!eventMinTime.isValid() || curEv->createTime.msecsTo(eventMinTime)>0){
                    eventMinTime = curEv->createTime;
                    if(eventsWithDebug)
                        qInfo()<<"The earliest ev from "<<id.toString();
                }
                break;
            }
        }

        // At this step eventMinTime (if valid) contains min time of event which evNum >= expectedEvNum
        // Timing loop
        gotClients=0;
        for(auto id : evBasedClients) {
            auto &ccd = ccds[id];
            leftTimeMs=0;
            while(true) {
                if(!updateClientsEventInfo(&ccd, leftTimeMs)){
                    if(leftTimeMs) {
                        loopExpired = true;
                        break; // try to get event from following client
                    }
                }

                auto curEv = ccd.evMI;
                if(curEv) {
                    if(curEv->evNum < expectedEvNum) {
                        addTrashStat(&ccd, curEv->evNum);
                        ccd.removeEv();
                        stats.clientsStat[id].trashedEvents++;
                        // try to receive correct ev in loop timeout
                        continue;
                    }

                    // curEv->evNum >= expectedEvNum
                    if(curEv->evNum == expectedEvNum) {
                        ++gotClients;
                    } else if(eventsWithDebug)
                        qInfo()<<QString("Got event with event numer (%1) greater that expected evNum (%2) from %3")
                              .arg(curEv->evNum).arg(expectedEvNum).arg(id.toString());

                    if(!eventMinTime.isValid() || curEv->createTime.msecsTo(eventMinTime)>0) {
                        eventMinTime = curEv->createTime;
                    }
                    break;
                }

                if(loopExpired) {
                    leftTimeMs = 0;
                } else {
                    qint64 timeToLoopExpire = loopTimeoutMs-loopTimer.elapsed();
                    if(eventMinTime.isValid()) {
                        qint64 timeToEvExpire = EVENT_EXPIRE_TIMEOUT_MS-eventMinTime.elapsed();
                        if(timeToEvExpire < 0)
                            eventExpired = true;
                        leftTimeMs = qMin(timeToEvExpire, timeToLoopExpire);
                    } else
                        leftTimeMs = timeToLoopExpire;
//                    qInfo()<<leftTimeMs<<timeToEvExpire<<timeToLoopExpire;
                    if(leftTimeMs < 0){
                        loopExpired = true;
                        leftTimeMs = 0;
                    }
                }
                if(leftTimeMs==0)
                    break;
            }
        }

        // At this moment we either collected all clients with proper evNum,
        // either workLoop timeout expired
        // or some client expired (eventMinTime)
        if(gotClients) {
            if(gotClients == clientSizeCache ||
               eventExpired                  ||
               eventMinTime.hasExpired(EVENT_EXPIRE_TIMEOUT_MS)) {
                if(gotClients == clientSizeCache) {
                    resetEventIncomplet();
                } else {
                    if(evIncompletCnt < EV_INCOMPLET_LIMIT) {
//                        EventMetaInfo* curEv;
                        QString msg = QString("Send incomplete event; curEvNum=%1; missing:").arg(expectedEvNum);
                        for(auto id : evBasedClients) {
                            auto &ev = ccds[id].evMI;
                            if(ev==nullptr || ev->evNum!=expectedEvNum) {
                                msg.append(id.toString()).append(',');
                            }
                        }
                        qWarning() << msg;
                        if(++evIncompletCnt == EV_INCOMPLET_LIMIT)
                            qWarning() << "prevEventIncomplet suppresed";
                    }
                }

                flushEvent();
            } else {
                // try to complete event in next work loop
            }
//            eventMinTime.invalidate();
        } else if(eventMinTime.isValid() && eventMinTime.hasExpired(EVENT_EXPIRE_TIMEOUT_MS)) {
            //todo is this a error?
            qWarning() << QString("Got expired event with EvNum greater than expectedEvNum %1; start initializing...")
                          .arg(expectedEvNum);
            initFlag=true;
//            eventsWithDebug=50;
//            qInfo()<<"\tInit eventsWithDebug";
        }
    }

    workLoopTimer->start();
}

void Combiner::sendStatus()
{
    stats.expectedEvNum = expectedEvNum;
    for(auto i : clientIds) {
        auto &ccd = ccds[i];
        Q_ASSERT(!ccd.index.isEmpty());
        EventMetaInfo * curEv = ccd.evMI;
        if(curEv && curEv->type==MPD_SYNC_NORM)
            stats.hasPendingEvent = true;
        auto &cs = stats.clientsStat[i];
        auto rb = ccd.rb;
        cs.freeEvPart = rb->getFreeSize();
        cs.preparedEvPart = rb->getDataSize();
        cs.flushedEvPart = rb->getTrashSize();
    }
    if(outToTcpServer) {
        stats.fwFullSize = 0;
        stats.fwDataPart = 0;
        for (const OutputTcpServer *tcpServer : outputTcpServers.values()) {
            if(tcpServer->getIndex()==0)
                continue;
            stats.fwFullSize += tcpServer->getRingBuffer()->getAllSize();
            stats.fwDataPart += tcpServer->getRingBuffer()->getDataSize();
        }
        if(stats.fwFullSize)
            stats.fwDataPart /= stats.fwFullSize;
    } else if(outToFile) {
        stats.fwDataPart = 100.*writerRingBuf->getDataSize()/writerRingBuf->getAllSize();
        stats.fwFullSize = writerRingBuf->getAllSize();
    } else {
        stats.fwFullSize = 0;
        stats.fwDataPart = 0;
    }

    stats.fwPrefSize = preferedBufSize;
    stats.timeStamp.start();
    emit statusRaw(stats);

    for (auto &ccd : ccds) {
        const ClientIndex &index = ccd.index;
        auto &ts = ccd.trashStat;
        for (quint64 _expectedEvNum : ts.keys()) {
            const EventsRange &range = ts[_expectedEvNum];
            if(range.first == range.second)
                qWarning() << QString("Lagging event evNum(%1) was trashed from %2; expected(%3)")
                            .arg(range.first).arg(index.toString()).arg(_expectedEvNum);
            else
                qWarning() << QString("Lagging events evNum(%1-%2) were trashed from %3; expected(%4)")
                            .arg(range.first).arg(range.second).arg(index.toString()).arg(_expectedEvNum);
        }
        ts.clear();
    }
}


void Combiner::tcpStatusUpdated(const OutputTcpStatus &s)
{
    bool immediateSend = true;
    if(stats.tcpStatus.contains(s.index)) {
        OutputTcpStatus &prevS = stats.tcpStatus[s.index];
        immediateSend = s.listening!=prevS.listening
                || s.localPort!=prevS.localPort
                || s.connected!=prevS.connected
                || s.peer.hostName!=prevS.peer.hostName;
    }

    stats.tcpStatus[s.index] = s;
    setOutputTcpConnected(s.index, s.connected);
    setOutputTcpDone(s.index, s.isEmpty);
    if(immediateSend)
        sendStatus();
}

void Combiner::tcpFinished(int tcpId, uint fId)
{
    Q_UNUSED(fId)
    setOutputTcpDone(tcpId, true);
}

void Combiner::writerFileOpened(bool opened, const RcRunId &_runId)
{
    if(runId == _runId) {
        fileTimer.start();
        writtenEvNum = 0;
        setFileOpened(opened);
    } else {
        qWarning() << QString("Got %1 with wrong runId=%2 (curRunId=%3)")
                      .arg(__func__, _runId.toString(), runId.toString());
    }
}

void Combiner::fileSplitted(uint id)
{
    Q_ASSERT(id == fileSplitId);
    splitInProgress = false;
    writtenEvNum = 0;
    auto lastFileTime = fileTimer.restart();
    auto t = QString::asprintf("Planned time: %ds; Actual time: %lldms (%+lld)",
                               splitConfig.timeFileSplit, lastFileTime,
                               lastFileTime-splitConfig.timeFileSplit*1000);
    qDebug() << __PRETTY_FUNCTION__ << t;
    writeFileSplitRecrd(true);
    writeRunDocument();
}

void Combiner::outputError(int index, const QString &errStr)
{
    if(getState() == FsmState::Configure)
        setRequestState(FsmState::Error);

    if(index == -1){
        // FileWriter
        qWarning() << "Got error in FileWriter:" << errStr;
    } else if(index){
        // Tcp Data Output
        qWarning() << QString("Got error in Tcp Server #%1 for data output:")
                      .arg(index) << errStr;
    } else {
        // Monitor
        qWarning() << "Got error in Tcp Server for monitor data" << errStr;
    }
    // emit error for all exept monitor
    if(index)
        emit error(errStr);
}

void Combiner::onFileWriterStatsUpdated(const FileWriterStats &s)
{
    if(!outToFile || !splitConfig.fileSizeEn)
        return;
    if(splitInProgress)
        return;
    if(s.writtenFileSize > splitConfig.fileSizeSplit) {
        qInfo() << QString("[%1] File split triggered: file size:%L2B limit:%L3B")
                   .arg(objectName()).arg(s.writtenFileSize).arg(splitConfig.fileSizeSplit);
        sendSplitRequest();
    }
}

void Combiner::doReqStatePreparations()
{
    auto reqSt = getRequestState();
    if(performedStateAction == reqSt)
        return;
    performedStateAction = reqSt;
    qDebug() << __PRETTY_FUNCTION__ << "reqSt:" << reqSt;

    switch (reqSt) {
    case FsmState::Configure:
        reset();
        break;
    case FsmState::Run:
        onStartingAction();
        break;
    case FsmState::Ready:
        onStoppingAction();
        break;
    default:;
    }
}

void Combiner::processInit()
{
    QElapsedTimer eventMinTimer;
    QElapsedTimer initTimer;
    initTimer.start();
    std::optional<quint32> minNumber;
    static constexpr const qint64 INIT_TIMEOUT_MS = LOOP_TIMEOUT/2;
    qint64 loopTimeout;
    bool hasSomeTime = true;

    for(auto i : clientIds) {
        loopTimeout = hasSomeTime ? INIT_TIMEOUT_MS : 0;
        auto &ccd = ccds[i];
        if(updateClientsEventInfo(&ccd, loopTimeout)) {
            const auto &ev = ccd.evMI;
            eventMinTimer = ev->createTime;
            minNumber = ev->evNum;
            break;
        }
        hasSomeTime = false;
    }

    if(!eventMinTimer.isValid()){
        // There is no any pending event and updateClientsEventInfo timed out
        return;
    }

    // Got at least one event from some client
    for(auto i : clientIds) {
        loopTimeout = hasSomeTime ? INIT_TIMEOUT_MS-initTimer.elapsed() : 0;
        if(loopTimeout<=0) {
            loopTimeout = 0;
            hasSomeTime = false;
        }
        auto &ccd = ccds[i];
        if(updateClientsEventInfo(&ccd, loopTimeout)) {
            auto ev = ccd.evMI;
            if(ev->createTime.msecsTo(eventMinTimer)>0)
                eventMinTimer = ev->createTime;
            if(!minNumber || *minNumber > ev->evNum)
                minNumber = ev->evNum;
        }
    }

    if(minNumber) {
        initFlag = false;
        expectedEvNum = *minNumber;
        qInfo() << QString("Initialied: minNumber=%1").arg(expectedEvNum);
    }
}

void Combiner::flushSingleRecord(ClientCombinerData *ccd)
{
    const ClientIndex &index = ccd->index;
    const EventMetaInfo* ev = ccd->evMI;
    if(ev == nullptr)
        return;

    const quint32 &allSize = static_cast<uint>(ev->size1 + ev->size2);
    bool ok;
    const auto &outBuffers = getOutBuffersForRecord(allSize, true, &ok);
    if(!ok)
        return;

    if((ev->type == MPD_SYNC_EOB || ev->type == MPD_SYNC_NORM) && combCfg.fixTlvFormat) {
        if(ev->type == MPD_SYNC_EOB) {
            mpdHdr.sync = MPD_SYNC_EOB_TLV;
        } else {
            mpdHdr.sync = MPD_SYNC_NORM_TLV;
        }
        mpdHdr.length = allSize - sizeof(MpdHdr);
        // send header
        writeData(outBuffers, reinterpret_cast<char*>(&mpdHdr), sizeof(MpdHdr));
        // send evNum or reserved field
        const char* evNumField = reinterpret_cast<const char*>(ev->type == MPD_SYNC_EOB
                                                         ? &expectedEvNum
                                                         : &(ev->evNum));
        writeData(outBuffers, evNumField, sizeof(expectedEvNum));
        // send payload
        sendClientPayload(outBuffers, ev);
    } else {
        writeData(outBuffers, ev->data1, ev->size1);
        if(ev->size2) {
            writeData(outBuffers, ev->data2, ev->size2);
        }
    }
    auto &cs = stats.clientsStat[index];
    cs.flushedSize += allSize;

    switch(ev->type) {
    case MPD_SYNC_NORM:
    case MPD_SYNC_NORM_TLV:
        ++cs.eventBlocks;
        break;
    case MPD_SYNC_EOB:
    case MPD_SYNC_EOB_TLV:
        ++stats.eob;
        ++cs.statBlocks;
        break;
    case MPD_SYNC_JSON_TLV:
        ++stats.jsonBlocks;
        ++cs.jsonBlocks;
        break;
    }
    ccd->removeEv();
//    qInfo() << "EOB: curEvNum="<<expectedEvNum;

    if(eventsWithDebug) {
        qInfo()<<QString("Flushed single event cl:%1 type:0x%2")
                 .arg(index.toString()).arg(ev->type,0,16);
//                 .arg(singleIndex.toString()).arg(getMpdSyncWordName(ev->type));
        --eventsWithDebug;
    }
}

void Combiner::flushEvent()
{
    quint32 allSize = sizeof(MpdEventHdr);
    QElapsedTimer minEvTime;
    bool completEvent = true;
    int gotClients = 0;

    // 1st step: Calculate total event size
    for(auto i : evBasedClients) {
        auto ev = ccds[i].evMI;
        if(ev == nullptr || ev->evNum != expectedEvNum)
            continue;

        ++gotClients;
        allSize += ev->size1 + ev->size2-sizeof(MpdEventHdr);
        if(combCfg.timeDistrEn && completEvent) {
            minEvTime = minEvTime.isValid() ? std::min(ev->createTime, minEvTime)
                                            : ev->createTime;
        }
    }

    // 2nd step: Deside to what RingBuffers we should write a data
    bool ok;
    const auto &outBuffers = getOutBuffersForRecord(allSize, false, &ok);
    if(!ok)
        return;

    // 3rd step: write a data
    // 3-1 step: write ev header
//    QString mess = QString("write ev:%1;").arg(mpdDataHdr.evNum);
    if(combCfg.fixTlvFormat) {
        mpdHdr.sync = MPD_SYNC_NORM_TLV;
        mpdHdr.length = allSize - sizeof(MpdHdr);
        writeData(outBuffers, reinterpret_cast<char*>(&mpdHdr), sizeof(MpdHdr));
        writeData(outBuffers, reinterpret_cast<char*>(&expectedEvNum), sizeof(expectedEvNum));
    } else {
        mpdEventHdr.evNum = expectedEvNum;
        mpdEventHdr.length = allSize - sizeof(MpdEventHdr);
        writeData(outBuffers, reinterpret_cast<char*>(&mpdEventHdr), sizeof(MpdEventHdr));
    }

    // 3-2 step: write ev device blocks
    stats.lastEventClients.clear();
    for(auto i : evBasedClients) {
        auto &ccd = ccds[i];
        auto &ev = ccd.evMI;

        if(ev == nullptr) {
            completEvent = false;
            auto &cs = stats.clientsStat[i];
            if(!cs.lastMissingEvNum.has_value() || expectedEvNum-cs.lastMissingEvNum.value()>1000) {
//                qWarning()<<"There is no event from "<<getClientNameByIndex(index)<<" evNum="<<expectedEvNum;
            }
            cs.lastMissingEvNum = expectedEvNum;
            continue;
        } else if(ev->evNum != expectedEvNum) {
            completEvent = false;
            if(ccd.suppresLimit > 0) {
                --ccd.suppresLimit;
                auto w = qWarning();
                w << QString("%1 has event with wrong evNum(%2) expectedEvNum(%3)")
                                  .arg(i.toString()).arg(ev->evNum).arg(expectedEvNum);
                if(ccd.suppresLimit==0) {
                    w << " (err suppresed)";
                }
            }
            continue;
        }
        if(ccd.suppresLimit==0)
            qInfo() << QString("%1 'wrong evNum' warnMsg restored").arg(objectName());
        ccd.suppresLimit = ClientCombinerData::SuppresLimitMax;

        auto &cs = stats.clientsStat[i];
        if(combCfg.timeDistrEn && completEvent) {
            const qint64 &dt = minEvTime.msecsTo(ev->createTime); // ev creation time relative to earliest
            cs.timeDistr = cs.timeDistr.value_or(0) + dt;
        }
        Q_ASSERT(!stats.lastEventClients.contains(i));
        stats.lastEventClients.insert(i);
//        mess.append(QString("%1 ").arg(i.toString()));

        cs.flushedSize += ev->size1+ev->size2;
        ++cs.eventBlocks;
        sendClientPayload(outBuffers, ev);

//        auto rb = ccd.rb;
//        qDebug()<<"trash flushed event="<<ev->evNum;
//        qDebug()<<"Combiner: anal="<<clientIds[index]<<" trashData "
//                           << "event pos="<<rb->getReaderPos()
//                           << " evNum="<<rb->getReadPointer()->evNum
//                           << ev->evNum;

        ccd.removeEv();
    }

    if(completEvent) {
        if(combCfg.timeDistrEn)
            ++stats.timeDistrCnt;
        stats.period.addEvent(allSize);
        ++stats.fullEvents;
        if(suppressLogLagging) {
            suppressLogLagging = 0;
            qWarning() << "suppressLogLagging canceled";
        }
    } else {
        ++stats.partedEvents;
    }
    ++stats.clientNumDistr[gotClients];
//    qDebug()<<mess;

    ++stats.regularEvents;
    if(eventsWithDebug) {
        qInfo()<<QString("Flushed combined event evNum:%1").arg(expectedEvNum);
        --eventsWithDebug;
    }
    ++writtenEvNum;
    if(outToFile) {
        checkEvNumSplit();
    }
    expectedEvNum += evNumberCfg.step;
    checkStopEvNum();
}

QList<RingBuffer<char> *> Combiner::getOutBuffersForRecord(quint32 allSize,
                                                           bool forceFirstOut,
                                                           bool *ok)
{
    QList<RingBuffer<char> *> outBuffers;

    RingBuffer<char> *primaryBuffer = nullptr;
    if(outToFile) {
        primaryBuffer = writerRingBuf;
    } else if(outToTcpServer) {
        auto outId = forceFirstOut ? 1 : static_cast<int>(expectedEvNum)%
                                     (outputTcpServers.size()-1)+1;
        primaryBuffer = outputTcpServers[outId]->getRingBuffer();
    }

    if(primaryBuffer) {
        if(primaryBuffer->waitFreeSpace(static_cast<int>(allSize), LOOP_TIMEOUT)) {
            outBuffers << primaryBuffer;
            *ok = true;
        } else {
            *ok = false;
            if(!noFreeSpaceError)
                qCritical() << "Can't write event, there is no free space in buffer";
            noFreeSpaceError = true;
        }
    } else {
        *ok = true;
        noFreeSpaceError = false;
    }

    const auto &monitorBuffer = outputTcpServers[0]->getRingBuffer();
    if(stats.tcpStatus[0].connected &&
            (monitorBuffer->getFreeSize() >= static_cast<int>(allSize))) {
        outBuffers << monitorBuffer;
        ++stats.monitorEvCount;
//        sendLog(QString("send evNum=%1").arg(mpdDataHdr.evNum));
    }
    return outBuffers;
}

void Combiner::sendClientPayload(const QList<RingBuffer<char> *> &outBuffers,
                                 const EventMetaInfo* ev)
{
    int headerSize1 = qMin(ev->size1, ev->headerSize);
    if(ev->size1-headerSize1>0) {
        writeData(outBuffers, ev->data1+headerSize1, ev->size1-headerSize1);
    } else
        Q_ASSERT(ev->size1-headerSize1==0);

    if(ev->size2) {
        int headerSize2 = static_cast<int>(sizeof(MpdEventHdr))-headerSize1;
        Q_ASSERT(headerSize2>=0 && headerSize2<sizeof(MpdEventHdr));
        writeData(outBuffers, ev->data2+headerSize2, ev->size2-headerSize2);
    }
}

void Combiner::writeData(const QList<RingBuffer<char> *> &outBuffers, const char *src, int size)
{
    Q_ASSERT(enabled);
    for(auto b : outBuffers)
        writeData(b, src, size);
}

void Combiner::writeData(RingBuffer<char> *buffer, const char *src, int size)
{
    if(buffer->getFreeSize()<size)
        qWarning() << "Start Combiner::writeData without enough space";
    Q_ASSERT_X(buffer->getFreeSize()>=size,
               "writeData",
               QThread::currentThread()->objectName().toStdString().c_str());
    const int part1=qMin(size, buffer->getContinuosFreeSize());
    memcpy(buffer->getWritePointer(), src, static_cast<uint>(part1));
    buffer->writeData(part1);
    const int part2=size-part1;
    if(part2){
        memcpy(buffer->getWritePointer(), src+part1, static_cast<uint>(part2));
        buffer->writeData(part2);
    }
}

struct FuncTimeTest
{
    FuncTimeTest(const ClientIndex &index, QElapsedTimer *_t, int *cnt)
        : index(index)
        , t(_t)
        , cnt(cnt){}
    ~FuncTimeTest(){
        auto v = t->elapsed();
        static qint64 max=0;
        if(v>max) {
            qInfo() << __PRETTY_FUNCTION__ << v << *cnt << index;
            max = v;
        }
    }

    ClientIndex index;
    QElapsedTimer *t;
    int *cnt;
};

QElapsedTimer updateClientsTimer2;
int updateClientsCnt;
bool Combiner::updateClientsEventInfo(ClientCombinerData *ccd, int timeout)
{
    const ClientIndex &index = ccd->index;
    auto &ev = ccd->evMI;
    if(ev != nullptr)
        return true;
//    updateClientsTimer2.start();
//    updateClientsCnt = 0;

//    FuncTimeTest tt(ccd->index, &updateClientsTimer2, &updateClientsCnt);
    auto rb = ccd->rb;
    int &rbDataSpace = ccd->rbDataSpace;
    auto &clientStat = stats.clientsStat[index];

    bool ret = false;
    int loopTimeout=0;

    // 1st loop with 0 timeout; following with incoming 'timeout'
    while(true) {
        if(rbDataSpace < 1) {
            // update evMetaInfo cache
            if(rb->waitDataSpace(1, loopTimeout))
                rbDataSpace = rb->getDataSize();
        }
        if(rbDataSpace >= 1) {
            ev = rb->getReadPointer();
            clientStat.allBlocks++;
            if(ev->type != MPD_SYNC_NORM && ev->type != MPD_SYNC_NORM_TLV){
                flushSingleRecord(ccd);
                if(loopTimeout) {
                    if(updateClientsTimer.hasExpired(timeout))
                        break;
                }
                continue;
            }

            restoreRealEvNum(index, ev);

            if(evNumberCfg.step > 1) {
                // EvB works in parallel
                if(ev->evNum % evNumberCfg.step != evNumberCfg.getEvNumIndex()) {
                    if(warnEvNumStepFitFail) {
                        warnEvNumStepFitFail = false;
                        qWarning()<<QString("Flush event which evNum(%1) % evNumStep(%2) != EvB_index(%3)")
                                    .arg(ev->evNum).arg(evNumberCfg.step).arg(evNumberCfg.getEvNumIndex());
                    }
                    flushSingleRecord(ccd);
                    continue;
                }
            }
            clientStat.lastEventNum = ev->evNum;
            if(eventsWithDebug)
                qInfo() << "updateClientsEventInfo " << index.toString()
                        << " evNumber=" << ev->evNum;

            ret = true;
            break;
        }

        if(loopTimeout) {
            if(updateClientsTimer.hasExpired(timeout))
                break;
        } else {
            if(timeout) {
                loopTimeout = timeout;
                updateClientsTimer.start();
            } else
                break;
        }
    }

    return ret;
}

void Combiner::restoreRealEvNum(const ClientIndex &index, EventMetaInfo *ev)
{
    // Restore real evNum with expectedEvNum (can be moved to Analyzer)
    static constexpr const quint32 BASE_SHIFT = 1<<24;
    static constexpr const quint32 BASE_MASK = BASE_SHIFT-1;
    quint32 shiftEvNum = expectedEvNum & ~BASE_MASK;
    ev->evNum &= BASE_MASK; // trunk to single mask
    const quint32 &cutExpected = expectedEvNum & BASE_MASK;
    if(cutExpected<ev->evNum && (ev->evNum-cutExpected)>(BASE_MASK/2)) {
        if(shiftEvNum) {
            shiftEvNum -= BASE_SHIFT;
        }
        if(warnEvNumShift) {
            qCritical() << QString("Seems lagging event (%1) with too big differance: "
                        "expected:%2 cutExpected:%3 curEv:%4")
                           .arg(index.toString()).arg(expectedEvNum)
                           .arg(cutExpected).arg(ev->evNum);
            warnEvNumShift = true;
        }
    } else if(cutExpected>ev->evNum && (cutExpected-ev->evNum)>(BASE_MASK/2)) {
        shiftEvNum += BASE_SHIFT;
        if(warnEvNumShift) {
            qCritical() << QString("Seems subsequent event (%1) with too big differance: "
                        "expected:%2 cutExpected:%3 curEv:%4")
                           .arg(index.toString()).arg(expectedEvNum)
                           .arg(cutExpected).arg(ev->evNum);
            warnEvNumShift = true;
        }
    }
    ev->evNum += shiftEvNum;
}

int Combiner::addTcpServer()
{
    const int id = outputTcpServers.size();
    OutputTcpServer *server = new OutputTcpServer(id);
    QThread *thread = new QThread(this);
    thread->setObjectName(id ? QString("outTcpServers #%1").arg(id) : "monitor Tcp");
    server->moveToThread(thread);
    thread->start();
    connect(this, &Combiner::startTcpServer, server, &OutputTcpServer::startListen);
    connect(this, &Combiner::deleteTcp, server, &OutputTcpServer::deleteTcp);
    connect(this, &Combiner::resetOutputs, server, &OutputTcpServer::clear);
    connect(server, &OutputTcpServer::statusUpdated, this, &Combiner::tcpStatusUpdated);
    connect(server, &OutputTcpServer::destroyed, thread, &QThread::quit);
    connect(this, &Combiner::startingTcpWork, server, &OutputTcpServer::startWork);
    connect(this, &Combiner::closeTcpWork, server, &OutputTcpServer::finishWork);
    connect(server, &OutputTcpServer::workFinished, this, &Combiner::tcpFinished);
    if(id) {
        connect(server, &OutputTcpServer::error, this, &Combiner::outputError);
        connect(this, &Combiner::infOutputUpdated, server, &OutputTcpServer::setInfOutput);
    } else
        connect(this, &Combiner::setMonitorPortNum, server, &OutputTcpServer::setPreferedPortNum);
    outputTcpServers[id] = server;
    outputTcpServerThreads[id] = thread;
    return id;
}

bool Combiner::writeRunDocument()
{
    if(runDocumentWritten)
        return true;
    if(!runDoc)
        return false;
    if(!outToFile)
        return false;

    Q_ASSERT(enabled);
    auto runDocObj = QVariant::fromValue(*runDoc).value<QJsonObject>();
    QJsonObject obj{{"config", runDocObj}};
    const auto &data = QJsonDocument(obj).toJson(QJsonDocument::Compact);
    auto allinedData = allineData(data);

    MpdDeviceHdr mpdDeviceHdr;
    mpdDeviceHdr.deviceSerial = SERIAL_RUN_DOCUMENT;
    mpdDeviceHdr.deviceId = DEVICE_ID_SOFTWARE;
    mpdDeviceHdr.length = allinedData.size();

    static const quint32 reservedField = 0;
    MpdHdr mpdHdr;
    mpdHdr.sync = MPD_SYNC_JSON_TLV;
    mpdHdr.length = mpdDeviceHdr.length + sizeof(MpdDeviceHdr) + sizeof(reservedField);

    MpdRawDataDevRecord curRec;
    curRec.addData(&mpdHdr, sizeof(mpdHdr));
    curRec.addData(&reservedField, sizeof(reservedField));
    curRec.addData(&mpdDeviceHdr, sizeof(mpdDeviceHdr));
    curRec.addData(allinedData.data(), mpdDeviceHdr.length);

    bool ok;
    const auto &outBuffers = getOutBuffersForRecord(curRec.size(), true, &ok);
    if(ok) {
        writeData(outBuffers, curRec.data.data(), curRec.size());
        runDocumentWritten = true;
    }
    return ok;
}

bool Combiner::writeStartStopRecrd(bool isStart)
{
    SartStopRecord r(isStart, runId.runNumber, runId.runIndex);
    return writeComplexRecrd(r);
}

bool Combiner::writeFileSplitRecrd(bool isStart)
{
    FileSplitRecord r(isStart, runId, fileSplitId, getWriterIndex());
    return writeComplexRecrd(r);
}

bool Combiner::writeComplexRecrd(const ComplexTlvRecord &r)
{
    Q_ASSERT(enabled);
    bool ok;
    const auto &outBuffers = getOutBuffersForRecord(r.totSize(), true, &ok);
    if(ok)
        writeData(outBuffers, r.getData().data(), r.totSize());
    return ok;
}

void Combiner::addTrashStat(ClientCombinerData *ccd, quint32 trashEvNum)
{
    auto &ts = ccd->trashStat;
    if(ts.contains(expectedEvNum)){
        EventsRange &range = ts[expectedEvNum];
        if(range.second+1 != trashEvNum)
            qWarning() << QString("Gap in trashing events (%1-%2) for same expected event number (%3)")
                          .arg(range.second).arg(trashEvNum).arg(expectedEvNum);
        range.second = trashEvNum;
    } else
        ts[expectedEvNum] = EventsRange(trashEvNum, trashEvNum);
}

void Combiner::resizeBuffer()
{
    Q_ASSERT(preferedBufSize>0);
    bool allResized=true;
    if(outToTcpServer) {
        constexpr const static int MIN_OUT_SIZE = 300e6;
        int nextTcpSize = preferedBufSize/(outputTcpServers.size()-1);
        nextTcpSize = qMax(nextTcpSize, MIN_OUT_SIZE);
        for (int id : outputTcpServers.keys()) {
            if(id==0)
                continue; // Skip monitor
            RingBuffer<char> *tcpRb = outputTcpServers[id]->getRingBuffer();
            if(tcpRb->getAllSize() == preferedBufSize)
                continue;

            bool ok=false;
            if(tcpRb->getFreeSize()==tcpRb->getAllSize()) {
                int oldSize = tcpRb->getAllSize();
                ok = tcpRb->resize(nextTcpSize);
                if(ok)
                    qInfo() << QString("TCP-buffer(%1) resized from %2MB to %3MB")
                               .arg(id).arg(1e-6*oldSize,0,'f',3)
                               .arg(1e-6*nextTcpSize,0,'f',3);
            }
            allResized &= ok;
        }
    } else {
        if(writerRingBuf->getFreeSize()==writerRingBuf->getAllSize()) {
            int oldSize = writerRingBuf->getAllSize();
            if(oldSize != preferedBufSize) {
                allResized = writerRingBuf->resize(preferedBufSize);
                if(allResized)
                    qInfo() << QString("File-buffer resized from %2MB to %3MB")
                               .arg(1e-6*oldSize,0,'f',3)
                               .arg(1e-6*preferedBufSize,0,'f',3);
            } else
                allResized = true;

        } else
            allResized = false;
    }
    if(allResized) {
        emit outputBufferSizeUpdated(0, preferedBufSize);
        waitingResize = false;
    }
}

void Combiner::checkStopEvNum()
{
    if(getState() != FsmState::PrepareStopping)
        return;
    setExpectedEvNum(expectedEvNum);
}

void Combiner::onStoppingAction()
{
    runDoc.reset();
    workLoopTimer->stop();
    if(getState()!= FsmState::Error) {
        if(!writeStartStopRecrd(false)) {
            //try again after event loop processing
            QTimer::singleShot(0, this, &Combiner::onStoppingAction);
            return;
        }
    }
    qDebug() << __PRETTY_FUNCTION__ << "disablinig combiner";
    enabled = false;
    if(outToFile) {
        emit closeFileRun();
    } else if(outToTcpServer) {
        ++finishId;
        // 0 reserved for error state
        if(finishId==0)
            ++finishId;
        emit closeTcpWork(finishId);
    }

    statusTimer->stop();
    sendStatus();
}

void Combiner::updateClientSizeCache()
{
    evBasedClients = clientIds-nonEvBasedClients;
    clientSizeCache = evBasedClients.size();
}

bool Combiner::isRunRecWritten()
{
    if(!startRecordWritten && writeStartStopRecrd(true))
        startRecordWritten = true;
    return startRecordWritten;
}

bool Combiner::checkFileSplitting()
{
    if(splitInProgress)
        return true;

    if(!outToFile)
        return false;

    if(checkTimeSplit())
        return true;
    if(checkEvNumSplit())
        return true;

    return false;
}

bool Combiner::checkTimeSplit()
{
    if(!outToFile || !splitConfig.timeEn)
        return false;

    if(fileTimer.elapsed() > splitConfig.timeFileSplit*1000){
        qInfo() << QString("[%1] File split triggered: time elapsed:%2ms limit:%3s")
                   .arg(objectName()).arg(fileTimer.elapsed()).arg(splitConfig.timeFileSplit);
        sendSplitRequest();
        return true;
    }
    return false;
}

bool Combiner::checkEvNumSplit()
{
    if(!outToFile || !splitConfig.evNumEn)
        return false;

    if(writtenEvNum > splitConfig.evNumFileSplit) {
        qInfo() << QString("[%1] File split triggered: event count:%L2 limit:%3")
                   .arg(objectName()).arg(writtenEvNum).arg(splitConfig.evNumFileSplit);
        sendSplitRequest();
        return true;
    }
    return false;
}

void Combiner::sendSplitRequest()
{
    if(!writeFileSplitRecrd(false))
        return; // try again on next workLoop
    splitInProgress = true;
    ++fileSplitId;
    qDebug() << __PRETTY_FUNCTION__ << "fileSplitId:" << fileSplitId;
    emit requestForFileSplit(fileSplitId);
}

std::optional<int> Combiner::getWriterIndex() const
{
    if(evNumberCfg.step == 1)
        return {};
    return static_cast<int>(evNumberCfg.getEvNumIndex());
}

void Combiner::resetEventIncomplet()
{
    if(!evIncompletCnt)
        return;

    evIncompletCnt=0;
    qWarning() << "prevEventIncomplet suppresed canceled; curEvNum="<<expectedEvNum;
//    qInfo() << "Resume to complete event; curEvNum="<<expectedEvNum;
}

void Combiner::updateAnalizerList(const ClientIndexSet &clients,
                                  const ClientIndexSet &nonEvBasedClients_,
                                  const QVector<Analizer *> &analizers)
{
    stats.clear();
    stats.clientsStat.clear();

    for(auto it=ccds.begin(); it!=ccds.end();) {
        if(clients.contains(it.key())) {
            it->evMI = nullptr;
            ++it;
         } else {
            it = ccds.erase(it);
        }
    }
    clientIds = clients;
    nonEvBasedClients = nonEvBasedClients_;
    updateClientSizeCache();
    for (int i=0;i<analizers.size();++i) {
        const Analizer * a = analizers[i];
        auto index = a->getClientIndex();
        auto &ccd = ccds[index];
        ccd.index = index;
        ccd.rb = a->getEventRingBuff();
        ccd.rbDataSpace = a->getEventRingBuff()->getDataSize();
    }
    qDebug() << QString("clientNumbers=%1").arg(clients.size());
}

void Combiner::gotRunDocument(const RunDocument &rd)
{
    if(runDoc == rd)
        return;
    if(runId != rd.runId)
        return;
    runDoc = rd;
}

void Combiner::setEvbEvNumberConfig(const EvbEvNumberConfig &cfg)
{
    if(evNumberCfg == cfg)
        return;

    evNumberCfg = cfg;
    expectedEvNum = evNumberCfg.startEvNum;
    qInfo() << QString("Set expected EvNum:%1; step:%2")
               .arg(expectedEvNum).arg(evNumberCfg.step);
}

void Combiner::setFileSplitterConfig(const EvbFileSplitterConfig &newConfig)
{
    if(splitConfig == newConfig)
        return;
    splitConfig = newConfig;
}

void ClientCombinerData::removeEv()
{
    evMI = nullptr;
    rb->trashData();
    --rbDataSpace;
}
