#include "CombinerFsm.h"

#include <QDebug>
#include <QLoggingCategory>

#include "EvBConfig.h"

namespace {
Q_DECLARE_LOGGING_CATEGORY(fsm)
Q_LOGGING_CATEGORY(fsm, "afi.fsm.EvB.Combiner")
}

CombinerFsm::CombinerFsm(QObject *parent):
    BaseFsm(parent)
{
    setObjectName("CombinerFsm");
    recreateInput(new CombinerFsmInput());
}

bool CombinerFsm::isTransitionAllowed(const FsmState &toState) const
{
    switch (getState()) {
    case FsmState::Ready:
        if(toState == FsmState::Configure) return true;
        break;
    case FsmState::Configure:
        if(toState == FsmState::Run) return checkStart();
        break;
    case FsmState::Run:
        if(toState == FsmState::PrepareStopping) return checkStopEvNum();
        if(toState == FsmState::Ready) return checkStop();
        break;
    case FsmState::PrepareStopping:
        if(toState == FsmState::Ready) return checkStop();
        break;
    default:;
    }
    return BaseFsm::isTransitionAllowed(toState);
}

void CombinerFsm::setOutputList(const EvbCombinerConfig &cfg)
{
    qDebug(fsm) << __PRETTY_FUNCTION__ << objectName()
                << cfg.combinerOut << cfg.combinerOutTcpNum;

    auto i = getInput();
    QSet<int> outputs;
    switch (cfg.combinerOut) {
    case COMBINER_FILE:
        outputs << 0;
        break;
    case COMBINER_TCP:
        for(int i=0; i<cfg.combinerOutTcpNum; ++i)
            outputs << i+1;
        break;
    default:;
    }
    i->combinerOutType = cfg.combinerOut;
    i->outputList = outputs;
    evaluate();
}

void CombinerFsm::setOutputTcpConnected(int id, bool connected)
{
    auto input = getInput();
    if(!input->outputList.contains(id))
        return;

    if(input->outputConnected.value(id,false) != connected) {
        input->outputConnected[id] = connected;
        evaluate();
    }
}

void CombinerFsm::setOutputTcpDone(int id, bool rbIsEmpty)
{
    auto input = getInput();
    if(!input->outputList.contains(id))
        return;

    if(input->outputDone.value(id, false) != rbIsEmpty) {
        input->outputDone[id] = rbIsEmpty;
        if(getState() == FsmState::Stopping)
            evaluate();
    }
}

void CombinerFsm::setStopEvNum(quint64 stopEvNum)
{
    qDebug(fsm) << __PRETTY_FUNCTION__ << objectName() << stopEvNum;
    getInput()->stopEvNum = stopEvNum;
    if(getState() == FsmState::PrepareStopping)
        evaluate();
}

void CombinerFsm::setExpectedEvNum(quint64 expectedEvNum)
{
    getInput()->expectedEvNum = expectedEvNum;
    if(getState() == FsmState::PrepareStopping) {
        qDebug(fsm) << __PRETTY_FUNCTION__ << objectName() << expectedEvNum;
        evaluate();
    }
}

void CombinerFsm::setFileOpened(bool fileOpened)
{
    getInput()->fileOpened = fileOpened;
    evaluate();
}

bool CombinerFsm::evaluate()
{
    doReqStatePreparations();
    if(BaseFsm::evaluate())
        return true;

    if(getState() == FsmState::Error) {
        if(!hasAnyError()) {
            setState(FsmState::Ready);
            return true;
        }
    } else {
        if(hasAnyError()) {
            setState(FsmState::Error);
            return true;
        }
    }

    return false;
}

void CombinerFsm::printInputString(QDebug dbg) const
{
    dbg << *getInput();
}

qint64 CombinerFsm::stateTimeoutMs(const FsmState &state) const
{
    switch (state) {
    case FsmState::Configure:
    case FsmState::PrepareStopping:
    case FsmState::Stopping:
        return 10e3;
    default:
        return BaseFsm::stateTimeoutMs(state);
    }
}

qint64 CombinerFsm::requestTimeoutMs() const
{
    switch (getRequestState()) {
    case FsmState::Configure:
    case FsmState::Ready:
        return getState()==FsmState::Error ? -1 : 1e3;
    case FsmState::Run:
    case FsmState::PrepareStopping:
    case FsmState::Stopping:
        return 10e3;
    default:;
    }
    return BaseFsm::requestTimeoutMs();
}

bool CombinerFsm::hasAnyError() const
{
    if(!allOutputConnected())
        return true;

    switch (getState()) {
    case FsmState::Error:
    case FsmState::Ready:
        return !checkStop();
    case FsmState::Run:
    case FsmState::PrepareStopping:
        if(getRequestState()==FsmState::Ready)
            return false;
        return false;

    default:
        return false;
    }
}

bool CombinerFsm::canAcceptReqState() const
{
    switch (getRequestState()) {
    case FsmState::Run: return checkStart();
    case FsmState::PrepareStopping: return checkStopEvNum();
    case FsmState::Ready: return checkStop();
    default: return true;
    }
}

bool CombinerFsm::checkStart() const
{
    const auto input = getInput();
    if(input->combinerOutType == COMBINER_FILE) {
        return input->fileOpened;
    } else if(input->combinerOutType == COMBINER_TCP) {
        return allOutputConnected();
    } else
        return true;
}

bool CombinerFsm::checkStop() const
{
    const auto input = getInput();
    if(input->combinerOutType == COMBINER_FILE) {
        return !input->fileOpened;
    } else if(input->combinerOutType == COMBINER_TCP) {
        return allOutputDone();
    } else
        return true;
}

bool CombinerFsm::allOutputConnected() const
{
    const auto &i = getInput();
    if(i->combinerOutType != COMBINER_TCP)
        return true;

    for(const auto &id : i->outputList)
        if(i->outputConnected.value(id, false) == false)
            return false;
    return true;
}

bool CombinerFsm::allOutputDone() const
{
    for(const auto &id : getInput()->outputList)
        if(getInput()->outputDone.value(id, false) == false)
            return false;
    return true;
}

bool CombinerFsm::checkStopEvNum() const
{
    // Last event block was transmitted to output
    return (getInput()->stopEvNum == 0) ||
        (getInput()->expectedEvNum > getInput()->stopEvNum);
}

QDebug operator<<(QDebug dbg, const CombinerFsmInput &in)
{
    dbg << BaseFsmInput(in);
    dbg << "setupKey:" << in.setupKey << ";";
    dbg << "fileOpened:" << in.fileOpened << ";";
    dbg << "stopEvNum:" << in.stopEvNum << ";";
    dbg << "expectedEvNum:" << in.expectedEvNum << ";";
    dbg << "combinerOutType:" << in.combinerOutType << ";";
    dbg << "outputList:" << in.outputList << ";";
    dbg << "outputConnected:" << in.outputConnected << ";";
    dbg << "outputDone:" << in.outputDone << ";";
    return dbg;
}
