#ifndef MSCDEVICE_H
#define MSCDEVICE_H

#include "ModularMregDevice.h"
#include "cal/CalResult.h"
#include "data/MscHistData.h"
#include "status/MscBuildParams.h"
#include "status/MscStatus.h"

#include <QMetaType>

class QTimer;

class Ddr3StreamFifoModule;
class ModularDeviceConfig;
class MscCycleCountersModule;
class MscDacCalibrator;
class MscHistModule;
class MscInputModule;
class MscStreamModule;

struct PlotHistHistory
{
    MscHistData getBoundedHist(std::optional<int64_t> lastGateWidth) const
    {
        int min = std::max(minPos.has_value()?(minPos.value()-1) : 0, 0);
        int max = std::min(maxPos.has_value()?(maxPos.value()) : hist.size(), hist.size());
        min=0;
        auto ret = hist.mid(min, max-min);

        if(!lastGateWidth.has_value())
            return ret;

        double lastGateWidthNs = 1e-9*static_cast<double>(lastGateWidth.value());
        std::optional<double> lastFilledTime;
        if(!hist.isEmpty())
            lastFilledTime = hist.last().x;
        if(!lastFilledTime.has_value() || lastFilledTime.value() < lastGateWidthNs) {
            ret.push_back({lastFilledTime.value_or(0), 0});
            ret.push_back({lastGateWidthNs, 0});
        }
        return ret;
    }

    uint32_t gateCnt = 0;
    quint16 ramAddr = 0;
    std::optional<int> minPos;
    std::optional<int> maxPos;
    MscHistData hist; // [gate] = hist
};

class MscDevice : public ModularMregDevice
{
    Q_OBJECT
public:
    explicit MscDevice(const DeviceIndex &index, QObject *parent = nullptr);
    ~MscDevice() override;

signals:
    void deviceConnectedFw(DeviceIndex index, QString fw) const;
    void histUpdated(const DeviceIndex &index, const MscDevHists &) const;
    void statusUpdated(const MscStatus &stat) const;
    void buildParametersUpdated(const DeviceIndex &index,
                                const MscBuildParams &params) const;
    void deviceConfigChanged(const DeviceIndex &index,
                             const ModularDeviceConfig &config) const; // dummy signal
    void cycleStatusUpdated(const DeviceIndex &index, const MscCycleStatus &) const;
    void calResultChanged(const CalResult &newCalResult);

public slots:
    void startCalibration(const ConfigKey &cal_key);
    void setCalResult(const CalResult &newCalResult);

protected:
    bool readStatus() override;
    bool writeConfig() override;
    void enforceConfigConstraints(ModularDeviceConfig *config) override;
    virtual void createModules() override;
    virtual void initModules() override;

protected slots:
    void on_deviceConnected() override;

private slots:
    void readHists();
    std::optional<MscHistData> readTimeHist(quint16 hId);
    std::optional<MscHistData> readPlotHist(quint16 hId);

private:
    bool readMscStatus();
    bool updateInputAndHistStatus();
    std::optional<CalResult> performCalibration();
    std::optional<DacCalResult> calibrateDac();

    MscStatus status;
    CalResult calResult;
    QMap<int, PlotHistHistory> prevPlotHists; // per hId
    MscBuildParams buildParams;

    QTimer * const histReadTimer;

    QSharedPointer<MscInputModule> inputMod;
    QSharedPointer<MscHistModule> histMod;
    QSharedPointer<MscStreamModule> streamMod;
    QSharedPointer<MscCycleCountersModule> cycleMod;
    QSharedPointer<Ddr3StreamFifoModule> ddrMod;
    MscDacCalibrator * const dacCalibrator;
};

#endif // MSCDEVICE_H
