/*
**    Copyright 2014-2015 Ivan Filippov
**    Copyright 2014-2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef TRIGDEVICE_H
#define TRIGDEVICE_H

#include "FirmwareBuildParams.h"
#include "ModularMregDevice.h"
#include "RcTrigDeviceConfig.h"
#include "TrigStatisticsWriter.h"
#include "TrigStatus.h"
#include "RunControlRegs.h"
#include "base/ConfigKey.h"
#include "base/SpillPlotData.h"
#include "base/TriggerHistogramData.h"

#include <optional>
#include <QVector>
#include <QMap>
#include <QMetaType>

class LegacyDeviceModule;
class MscHistModule;
struct SpillConfig;
struct TrigSourceConfig;

constexpr int TRIG_STATUS_POLL_TIME_MS_DEFAULT = 200;

class TrigDevice : public ModularMregDevice
{
    Q_OBJECT
public:
    explicit TrigDevice(const DeviceIndex &deviceIndex, QObject *parent = nullptr);

    static double getClockFreqHz() { return 62500000; }
    static double getTrigTimerMinFreqHz() { return getClockFreqHz() / (quint32)(-1); }
    static double getTrigTimerMaxFreqHz() { return getClockFreqHz() / 2; }

signals:
    void firmwareBuildParamsChanged(const FirmwareBuildParams &);
    void updateStatus(const TrigStatus &status);
    void spillPlotUpdated(const SpillPlotData &);
    void triggerHistogramUpdated(const TriggerHistogramData &);
    void trigSpillStatisticsUpdated(const DeviceIndex &index, const TrigSpillStatistics &);
    void trigSrcCntUpdated(quint8 trigScrCnt);
    void topEvNumResponse(int checkIndex, quint64 evNum);
    void setStatusDumpDirName(const QString &);
    void setStatusDumpEnable(bool);

public slots:
    void writeSetup(const RcTrigDeviceConfig &cfg, const ConfigKey &setupSeqNum);
    void resetAll();
    void preRunReset(const QVariant &startEvNum);
    bool clearStats();
    void evNumRequest(int checkIndex, bool local);

protected slots:
    void trigStatusPoll();
    virtual void on_deviceConnected() override;
    virtual void on_deviceDisconnected() override;

protected:
    bool readStatus() override;
    void enforceConfigConstraints(ModularDeviceConfig *config) override;
    virtual void createModules() override;

private:
    quint8 getTrigSrcCount() const;
    std::optional<FirmwareBuildParams> readFirmwareBuildParams();
    int spillPlotSize() const { return 1 << fbp.spill_plot_ram_addr_bits;}
    int triggerHistogramSize() const {return 1 << fbp.trig_hist_ram_addr_bits;}
    mlink::RegOpVector setControlStateCmd(bool trigOpen, bool softSpill);
    mlink::RegOpVector setTrigTimerFreqCmd(double freq);
    mlink::RegOpVector setTrigRandomFreqCmd(double freq);
    mlink::RegOpVector setTrigRandomDeadTimeCmd(double seconds);
    mlink::RegOpVector setTrigCmd(const ModularDeviceConfig &trig);

    bool writeTrigSource(const TrigSourceConfig &trig);
    bool writeTrigCtrlStatConfig(const TrigCtrlStatConfig &trig);
    void setStatusPollTimeMs(int ms);

    std::optional<quint64> readGlobEvNum();
    void evNumReset(const QVariant &startEvNum);
    std::optional<SpillPlotData> readSpillPlotRam();
    std::optional<TriggerHistogramData> readTriggerTimeHistogramRam();
    void analyzeDeviceStatistics(const TrigSpillStatistics &data);
    bool writeSetup_internal(const RcTrigDeviceConfig &lastConfig, const RcTrigDeviceConfig &cfg, bool force = true);
    bool readTrigStatus();
    std::optional<RcCounters> readRcCounters();
    std::optional<TrigStatCounters> readStatCounters();
    std::optional<TrigCoreStatus> readTrigCoreStatus();
    void debugPrintStatus(const TrigStatus &status);
    mlink::RegOpVector regOpGetCoreStatus(TrigCoreStatus *result) const;
    mlink::RegOpVector regOpGetWrStatus(TrigWrStatus *result) const;
    std::optional<TrigWrStatus> readTrigWrStatus();
    bool updateCounters();

    QSharedPointer<MscHistModule> histMod;
    TrigStatisticsWriter tsw;
    RcTrigDeviceConfig lastConfig;
    ConfigKey lastSetupSeqNum;
    QTimer * const trigStatusTimer;
    FirmwareBuildParams fbp;
    TrigStatus status;
    reg_spill_ctrl_t regCtrl;
    int statusPollTimeMs = TRIG_STATUS_POLL_TIME_MS_DEFAULT;
    TrigSpillStatistics trigSpillStatistics;
    std::optional<QPair<quint32,quint32>> lastSpdInf;

    QSharedPointer<LegacyDeviceModule> legacyDevice;
};

#endif // TRIGDEVICE_H
