#include "TqdcMainWindow.h"
#include "ui_TqdcMainWindow.h"

#include <QMetaType>
#include <QMessageBox>
#include <QPlainTextEdit>
#include <QSettings>
#include <QTimer>
#include <QSplitter>

#include "CalibrationDialog.h"
#include "TqdcSignalRouter.h"
#include "DeviceChTree/DeviceChModel.h"

#include "base/Globals.h"
#include "configuration-manager/ConfigurationManager.h"
#include "device-modules/TdcCore/TdcCoreStatus.h"

namespace {
const int STATUS_UPDATE_DELAY = 1000;

void setWidgetWarn(QWidget *w, const QStringList &warnList)
{
    bool en = w->isEnabled();
    auto c = en ? (warnList.isEmpty() ? QColor()         : QColor(0xFF8800)) :
                  (warnList.isEmpty() ? QColor(0xbfbfbf) : QColor(0xFFCC44));
    auto p = w->palette();

    auto b = p.brush(w->foregroundRole());
    if(p.color(w->foregroundRole()) != c) {
        p.setColor(w->foregroundRole(), c);
        w->setPalette(p);
    }

    QString tt = warnList.join('\n');
    if(w->toolTip() != tt)
        w->setToolTip(tt);
}
}

TqdcMainWindow::TqdcMainWindow(QWidget *parent) :
      ReadoutAppMainWindow(parent)
    , calibrationDialog(new CalibrationDialog(this))
    , ui(new Ui::TqdcMainWindow)
    , treeModel(new DeviceChModel(this))
    , status_update_timer(new QTimer(this))
{
    ui->setupUi(this);
    setupMenu(ui->menuFile, ui->menuHelp, ui->menuOptions, ui->menuTools);
/*
    auto splitter = new QSplitter(ui->centralWidget);
    splitter->setOrientation(Qt::Vertical);
    splitter->addWidget(ui->upperFrame);
    splitter->addWidget(ui->lowerFrame);
    ui->centralWidget->layout()->addWidget(splitter);
*/
    status_update_timer->setSingleShot(true);
    status_update_timer->setInterval(STATUS_UPDATE_DELAY);

    ui->deviceTreeView->setTreeModel(treeModel);

    configuration_manager->setDefaultConfig(BaseDeviceAppConfig());

    init_main_window();
    updateControlState();

    auto &signalRouter = TqdcSignalRouter::instance();
    signalRouter.setMainWindow(this);
}

TqdcMainWindow::~TqdcMainWindow()
{
    emit closeAll();
    delete ui;
}

DeviceStatusTableViewFancy *TqdcMainWindow::getDeviceStatusTableView()
{
    return ui->deviceStatusTableView;
}

DeviceChTreeView *TqdcMainWindow::getDeviceChTreeView() const
{
    return ui->deviceTreeView;
}

TrigCtrlStatWidget *TqdcMainWindow::getTrigCtrlStatWidget() const
{
    return ui->trigCtrlStatWidget;
}

RunLogicWidget *TqdcMainWindow::getRunLogicWidget() const
{
    return ui->runLogicWidget;
}

FirFilterWidget *TqdcMainWindow::getFirFilterWidget() const
{
    return ui->firFilterWidget;
}

WaveBlcWidget *TqdcMainWindow::getWaveBlcWidget() const
{
    return ui->waveBlcWidget;
}

TtlIoWidget *TqdcMainWindow::getWidgetTtl() const
{
    return ui->widgetTtl;
}

void TqdcMainWindow::init_main_window()
{
    ui->labelAdcPreWin->hide();
    ui->labelAdcPreWinNs->hide();
    ui->spinBoxAdcPreWin->hide();
    ui->comboBoxEdges->addItem("Leading", 1);
    ui->comboBoxEdges->addItem("Trailing", 2);
    ui->comboBoxEdges->addItem("Lead+Trail", 3);

    {
        auto cb = ui->comboBoxZsMode;
        QSignalBlocker blocker(cb);
        cb->addItem("Disabled", static_cast<int>(WaveRecConfig::ZsDisabled));
        cb->addItem("Analog", static_cast<int>(WaveRecConfig::ZsAnlg));
        cb->addItem("Digital", static_cast<int>(WaveRecConfig::ZsDig));
    }
}

void TqdcMainWindow::saveAppGuiSettings(QSettings &s) const
{
    ui->deviceTreeView->saveAppGuiSettings(s);
}

void TqdcMainWindow::loadAppGuiSettings(const QSettings &s)
{
    ui->deviceTreeView->loadAppGuiSettings(s);
}

void TqdcMainWindow::updateControlState()
{
    bool inStateErrReady = curFsmState == FsmState::Ready || curFsmState == FsmState::Error;

    ui->pushButtonStart->setEnabled(curFsmState == FsmState::Ready && !underRemoteCtrl);
    ui->pushButtonStop->setEnabled(curFsmState == FsmState::Run && !underRemoteCtrl);

    controlStateEnable((inStateErrReady || Globals::getDebugMode())
                       && curFsmState != FsmState::Calibrate);

    ui->actionReset_MStream->setEnabled(inStateErrReady);
    ui->actionSend_soft_reset->setEnabled(inStateErrReady);
    ui->deviceStatusTableView->setControlEnabled(inStateErrReady);
}

void TqdcMainWindow::deviceStatusUpdated(const ModularDeviceStatus &status)
{
    dev_status[status.getDeviceIndex()] = status;
    deviceStatusUpdatedRequest();
}

void TqdcMainWindow::deviceStatusUpdatedRequest()
{
    deviceStatusUpdatedResponce(); // FIXME: remove me
    if (! status_update_timer->isActive()) {
        deviceStatusUpdatedResponce();
        status_update_timer->start();
    }
}

void TqdcMainWindow::deviceStatusUpdatedResponce()
{
    for(auto &s : dev_status.keys()){
        const auto &status = dev_status[s];
        const auto &deviceIndex = status.getDeviceIndex();
        if(!config.knownConfigs.contains(deviceIndex) ||
                !config.knownConfigs.value(deviceIndex).enabled) {
            dev_status.remove(s);
            continue;
        }

        emit setStatus(status);

        treeModel->deviceStatusUpdated(status);
        ui->deviceTreeView->deviceStatusUpdated(status);
        if(status.connected) {
            QStringList statusList;
            if (status.tdcCoreStatus.has_value()) {
                const auto &tdcCoreStatus = status.tdcCoreStatus.value();
                for(int i=0; i<tdcCoreStatus.hptdcStatus.size(); ++i) {
                    if(!tdcCoreStatus.hptdcStatus[i].getSummary().isEmpty()) {
                        statusList << QString("HPTDC #%1:\n%2").arg(i)
                                          .arg(tdcCoreStatus.hptdcStatus[i].getSummary());
                    }
                }
            }
            if (status.desStatus.has_value() && !status.desStatus.value().isOk()) {
                statusList << QString("DesStatus: %1").arg(status.desStatus.value().toString());
            }

            QString str = statusList.isEmpty() ? "Ok"
                                               : statusList.join('\n');
            if (status.isRunFinished()) {
                //TODO: move phasing stop logic to core
                //TODO: get stopEvNum in status after closing triggers
                if (status.trigInfo.has_value()) {
                    stopEvNum = status.trigInfo.value().eventNumber - 1;
                    emit prepareForStop(stopEvNum);
                } else {
                    stopEvNum = 0;
                }
                // qInfo().noquote() <<
                //     "[MainWindow]: Stop event number detected:" << stopEvNum;
            }
        }
    }
}

bool TqdcMainWindow::isControlBlocked() const
{
    return blockControl || underRemoteCtrl;
}

void TqdcMainWindow::updateMatchWinMaximum() const
{
    const QSettings settings;
    auto matchWinLimitEn = settings.value("matchWinLimitEn").toBool();

    int v = matchWinLimitEn ? config.defaultSetup.tqdcCore.latency : 670;
    ui->spinBoxMatchWin->setMaximum(std::min(v, 670));
}

void TqdcMainWindow::updateWdgetWarnings() const
{
    const auto &ds= config.defaultSetup.tqdcCore;
    const auto &wr= config.defaultSetup.waveRec;
    QStringList warnList;
    if(config.defaultSetup.waveRec.zsMode && ((ds.matchWin*3 + wr.readCellNumber) >= (670*3)))
        warnList << "MatchWin+AdcWin > 16us";

    setWidgetWarn(ui->comboBoxZsMode, warnList);
    setWidgetWarn(ui->labelAdcWinNs, warnList);

    if(ds.matchWin > ds.latency)
        warnList << "MatchWin > Latency";

    setWidgetWarn(ui->labelMatchWinNs, warnList);
}

void TqdcMainWindow::deviceInfoUpdated(DeviceIndex index, int maxSampleCount)
{
    auto cfgIt = config.knownConfigs.find(index);
    if(cfgIt == config.knownConfigs.end())
        return;
    auto &cfg = cfgIt.value().tqdcCore;
    int chNum = TqdcCoreConfig::getNumberOfChannels(index.getDevId());
    if (cfg.getNCh() != chNum) {
        cfg.setNCh(chNum);
        emit configChanged(config);
    }
    ui->spinBoxLatency->setMaximum(maxSampleCount ? maxSampleCount/3 : 2048); // 2083 taps - 50us
//    ui->spinBoxMatchWin -> setMaximum(maxSampleCount ? maxSampleCount/3 : 416);
    ui->spinBoxAdcWin->setMaximum(maxSampleCount ? maxSampleCount/2  : 625);
    updateGuiVal();
}

void TqdcMainWindow::setConfig(const ConfigKey &config_key, const BaseDeviceAppConfig &newConfig)
{
    if(config == newConfig)
        return;

    ReadoutAppMainWindow::setConfig(config_key, newConfig);
    updateGuiVal();
    updateControlState();
}

void TqdcMainWindow::copyAndSetupDevice()
{
    config.unify_config();
    setupDevice();
}

void TqdcMainWindow::setupDevice()
{
    if(blockControl) return;
    emit configChanged(config);
}

void TqdcMainWindow::updateGuiVal()
{
    ++blockControl;
    const auto &ds = config.defaultSetup;

    const QSettings settings;
    auto separateChEn = settings.value("separateChEn").toBool();
    auto matchWinLimitEn = settings.value("matchWinLimitEn").toBool();
    {
        const auto &tqdcCore = ds.tqdcCore;
        const auto &wrCfg = ds.waveRec;

        ui->spinBoxAdcPreWin->setValue(wrCfg.adcPreWin);
        ui->spinBoxAdcWin->setValue(wrCfg.readCellNumber);
        ui->spinBoxLatency->setValue(tqdcCore.latency);
        updateMatchWinMaximum();
        ui->spinBoxMatchWin->setValue(tqdcCore.matchWin);
        updateWdgetWarnings();

        ui->deviceTreeView->setseparateChEn(separateChEn);

        const auto &waveRec = ds.waveRec;
        auto zsModeIndex = ui->comboBoxZsMode->findData(waveRec.zsMode);
        if(zsModeIndex != -1)
            ui->comboBoxZsMode->setCurrentIndex(zsModeIndex);
        ui->spinBoxAdcWin->setEnabled(waveRec.zsMode != WaveRecConfig::ZsDisabled);
        ui->spinBoxAdcPreWin->setEnabled(waveRec.zsMode != WaveRecConfig::ZsDisabled);

        ui->radioButtonGain1->setChecked(!tqdcCore.isGain4());
        ui->radioButtonGain4->setChecked(tqdcCore.isGain4());
        ui->actionAdd_HPTDC_event_header->setChecked(tqdcCore.hptdcHdr);
        ui->actionAdd_HPTDC_event_tralier->setChecked(tqdcCore.hptdcTrl);
        int edgeData = 0;
        if(tqdcCore.leadEdge) edgeData |= 1;
        if(tqdcCore.trailEdge) edgeData |= 2;
        ui->comboBoxEdges->setCurrentIndex(ui->comboBoxEdges->findData(edgeData));
    }

    ui->actionSeparate_channel_enable->setChecked(separateChEn);
    ui->actionDisable_match_win_limitations->setChecked(!matchWinLimitEn);

    --blockControl;
    updateControlState();
}

void TqdcMainWindow::controlStateEnable(bool enable)
{
    static bool curState = true;
    if(curState == enable)
        return;
    curState = enable;

    ui->groupBoxAdc->setEnabled(enable);
    ui->groupBoxTtl->setEnabled(enable);
    treeModel->setEnabled(enable);
    updateWdgetWarnings();
}

// ==== Slots from UI
void TqdcMainWindow::on_pushButtonStart_clicked()
{
    if(isControlBlocked()) return;
    if(!config.hasEnabledDevices()) {
        qWarning() << "There is no enabled device";
        return;
    }

    emit performLocalStart();
}

void TqdcMainWindow::on_pushButtonStop_clicked()
{
    if(isControlBlocked()) return;
    emit performLocalStop();
}

void TqdcMainWindow::on_radioButtonGain1_toggled(bool checked)
{
    on_radioButtonGain4_toggled(!checked);
}

void TqdcMainWindow::on_radioButtonGain4_toggled(bool checked)
{
    auto gain = checked ? TqdcCoreConfig::Gain4 : TqdcCoreConfig::Gain1;
    config.defaultSetup.tqdcCore.setGain(gain);
    for(auto &ds : config.knownConfigs)
        ds.tqdcCore.setGain(gain);
    copyAndSetupDevice();
}

void TqdcMainWindow::on_comboBoxZsMode_currentIndexChanged(int index)
{
    auto mode = static_cast<WaveRecConfig::ZsMode>(ui->comboBoxZsMode->itemData(index).toInt());
    auto zsEn = mode != WaveRecConfig::ZsDisabled;
    ui->spinBoxAdcWin->setEnabled(zsEn);
    ui->spinBoxAdcPreWin->setEnabled(zsEn);

    config.defaultSetup.waveRec.zsMode = mode;
    copyAndSetupDevice();
    updateWdgetWarnings();
}

void TqdcMainWindow::on_spinBoxLatency_valueChanged(int lat)
{
    ui->labelLatencyNs->setText(QString("%1 ns").arg(lat*24));
    auto &ds = config.defaultSetup;
    ds.tqdcCore.latency = lat;

    const QSettings settings;
    auto matchWinLimitEn = settings.value("matchWinLimitEn").toBool();
    if (matchWinLimitEn)
        updateMatchWinMaximum();
    updateWdgetWarnings();
    if(blockControl) return;
    copyAndSetupDevice();
}

void TqdcMainWindow::on_spinBoxMatchWin_valueChanged(int win)
{
    ui->labelMatchWinNs->setText(QString("%1 ns").arg(win*24));
    config.defaultSetup.tqdcCore.matchWin = win;
    updateWdgetWarnings();
    if(blockControl) return;
    copyAndSetupDevice();
}

void TqdcMainWindow::on_spinBoxAdcWin_valueChanged(int val)
{
    ui->labelAdcWinNs->setText(QString("%1 ns").arg(val*8));

    config.defaultSetup.waveRec.readCellNumber = val;
    if(blockControl) return;
    copyAndSetupDevice();
}

void TqdcMainWindow::on_spinBoxAdcPreWin_valueChanged(int preWin)
{
    ui->labelAdcPreWinNs->setText(QString("%1 ns").arg(preWin*8));
    config.defaultSetup.waveRec.adcPreWin = preWin;
    if(blockControl) return;
    copyAndSetupDevice();
}

void TqdcMainWindow::on_actionCalibrationData_triggered()
{
    calibrationDialog->show();
}

void TqdcMainWindow::on_actionSeparate_channel_enable_toggled(bool checked)
{
    QSettings settings;
    settings.setValue("separateChEn", checked);
    emit configChanged(config);
    updateGuiVal();
}

void TqdcMainWindow::on_actionDisable_match_win_limitations_toggled(bool checked)
{
    QSettings settings;
    settings.setValue("matchWinLimitEn", !checked);
    updateGuiVal();
}

void TqdcMainWindow::on_actionReset_MStream_triggered()
{
    emit mstream_reset_all(true);
}

void TqdcMainWindow::on_actionSend_soft_reset_triggered()
{
    emit softReset();
}

void TqdcMainWindow::on_actionPrint_FSM_inputs_triggered()
{
    emit printFsmInputs();
}

void TqdcMainWindow::on_actionTxt_data_format_triggered()
{
    QString s;
    s += "<h2>TQDC2 Txt Data format</h2>";
    s += "<p>File constructs with following lines:</p>";
    s += "<ul>";
    s += "<li>Ev <i>&lt;evNumber&gt;</i> <i>&lt;TAI_sec&gt;</i>.<i>&lt;TAI_ns&gt;</i></li>";
    s += "<li>Tdc <i>&lt;ch&gt;</i>: <i>&lt;tdcLeadTime&gt;</i></li>";
    s += "<li>Adc <i>&lt;ch&gt;</i>: <i>&lt;adcTime&gt;</i> <i>&lt;sample1&gt;</i> <i>&lt;sample2&gt;</i> ... <i>&lt;sampleN&gt;</i></li>";
    s += "</ul>";
    s += "<p>where:</p>";
    s += "<dl>";
    s += "<dt><b>TAI time</b></dt>";
    s += "<dd><i>Precise arrival time of the trigger</i></dd>";
    s += "<dt><b>ch</b></dt>";
    s += "<dd><i>Channel number in range 0-15</i></dd>";
    s += "<dt><b>tdcLeadTime</b></dt>";
    s += "<dd><i>HPTDC time in 0.024/1.024ns steps from trigger (TAI time)</i></dd>";
    s += "<dt><b>adcTime</b></dt>";
    s += "<dd><i>Timestamp of the first sample in 8ns steps from trigger (TAI time)</i></dd>";
    s += "<dt><b>sample</b></dt>";
    s += "<dd><i>ADC sample in signed 16-bit integer</i></dd>";
    s += "</dl>";

    QMessageBox::about (this, "Txt data format", s);
}

void TqdcMainWindow::on_actionRaw_data_format_triggered()
{
    QDialog d;
    QPlainTextEdit scrollArea;
    QLabel htmlLabel;

    QFile htmlFile(":/rawDataFormat.html");
    QString html;
    if(htmlFile.open(QIODevice::ReadOnly)){
        html = QString(htmlFile.readAll());
        htmlFile.close();
        if(html.isEmpty())
            html = "<b>Channel map file is empty.</b>";
    } else
        html = "<b>There is no channel map file.</b>";

    scrollArea.appendHtml(html);
    htmlLabel.setText(html);

    QVBoxLayout *mainLayout = new QVBoxLayout(&d);
//    mainLayout->addWidget(&scrollArea);
    mainLayout->addWidget(&htmlLabel);
    d.setLayout(mainLayout);
    d.exec();

//    QMessageBox::about (this, "Raw data format", html);
}

void TqdcMainWindow::on_actionAdd_HPTDC_event_header_toggled(bool checked)
{
    config.defaultSetup.tqdcCore.hptdcHdr = checked;
    copyAndSetupDevice();
}

void TqdcMainWindow::on_actionAdd_HPTDC_event_tralier_toggled(bool checked)
{
    config.defaultSetup.tqdcCore.hptdcTrl = checked;
    copyAndSetupDevice();
}

void TqdcMainWindow::on_comboBoxEdges_currentIndexChanged(int index)
{
    bool ok;
    int v = ui->comboBoxEdges->itemData(index).toInt(&ok);
    Q_ASSERT(ok);

    config.defaultSetup.tqdcCore.leadEdge = v&1;
    config.defaultSetup.tqdcCore.trailEdge = v&2;
    copyAndSetupDevice();
}
