#include "CalConfig.h"

#include <QCoreApplication>
#include <QDebug>
#include <QDir>
#include <QDirIterator>
#include <QSettings>

#include "base/DeviceId.h"
#include "base/DeviceIndex.h"
#include "histogramm/HistMngrWidget.h"

namespace {
const QString INL_PREFIX = "inl_corr";

inline QString getSectionStr(const DeviceIndex &devIndex, const QString &sectionType)
{
    return QString("%1-%2-%3")
        .arg(DeviceIndex::getDeviceIdName(devIndex.getDevId()))
            .arg(devIndex.getSerial(), 8, 16, QChar('0'))
            .arg(sectionType);
}
}

void CalConfig::calibrationUpdated(QMap<DeviceIndex, DeviceCalibration> newCal, QString baseDir)
{
    for(auto devCal : newCal.values()) {
        if(devCal.inl.isEmpty()) continue;
        QString fileName = QString("%1%2%3%2%1-%3.ini")
                               .arg(DeviceIndex::getDeviceIdName(devCal.index.getDevId()))
                .arg(QDir::separator())
                .arg(devCal.index.getSerialStr().toUpper());
        if(!baseDir.isEmpty()){
            fileName.prepend(QString("%1%2").arg(baseDir).arg(QDir::separator()));
        }
        QFileInfo dir(fileName);
        dir.dir().mkpath(".");
        QSettings *s = baseDir.isNull() ? new QSettings() : new QSettings(fileName, QSettings::IniFormat);

        qInfo() << s->fileName();
        QString group = getInlSectionStr(devCal.index);
        s->beginGroup(group);
        for(auto ch : devCal.inl.keys()) {
            QMap<QString, QString> statMap;
            QList<int> temp_ids = devCal.minTemp.keys();
            temp_ids << devCal.maxTemp.keys();
            for(int temp_id : temp_ids.toSet()) {
                if(devCal.minTemp.contains(temp_id))
                    statMap[QString("min_temp_%1").arg(temp_id)] = QString::number(devCal.minTemp[temp_id], 'f', 1);
                if(devCal.maxTemp.contains(temp_id))
                    statMap[QString("max_temp_%1").arg(temp_id)] = QString::number(devCal.maxTemp[temp_id], 'f', 1);
            }

            if(!devCal.fw_versions.isEmpty()) {
                QStringList list;
                for(uint rev : devCal.fw_versions)
                    list << QString::number(rev);
                statMap["fw_rev"] = list.join(" ");
            }
            if(!devCal.bmc_versions.isEmpty()) {
                QStringList list;
                for(uint rev : devCal.bmc_versions)
                    list << QString::number(rev);
                statMap["bmc_rev"] = list.join(" ");
            }

            const QVector<double> &inlSrc = devCal.inl[ch];
            if(inlSrc.isEmpty())
                continue;

            quint64 dataCnt = static_cast<quint64>(std::accumulate(inlSrc.begin(), inlSrc.end(), 0));
            if(dataCnt==0)
                continue;

            QVector<double> corr = HistMngrWidget::makeDNLCorrectionTable(inlSrc);
            qInfo() << QString("Inl for mod:%1 ch:%2: %3 entries")
                       .arg(devCal.index.getIdent())
                       .arg(ch).arg(dataCnt);


            QList<QVariant> sVars;
            if(corr.size() != 1024)
                continue;

            for(int i=0; i<1024; i++)
                sVars.push_back(QString::number(corr[i], 'f', 2));
            s->setValue(QString::number(ch), sVars);

            QStringList statList;
            statList << QString("hits:%L1").arg(dataCnt, 11);
            for(auto key : statMap.keys()) {
                statList << QString("%1: %2").arg(key, statMap[key]);
            }

            s->setValue(QString("stat_%1").arg(ch,2,10,QChar('0')), statList.join("; "));
        }
        s->endGroup();
        s->sync();
        delete s;
    }
}

void CalConfig::reloadCalibration(const QString &separateInlFolder)
{
    reloadCalibration();

    if(!separateInlFolder.isEmpty()) {
        QStringList nameFilters;
        nameFilters << "*.ini" << "*.conf";
        QDirIterator it(separateInlFolder, nameFilters, QDir::Files | QDir::NoDotDot, QDirIterator::Subdirectories);
        while(it.hasNext()) {
            auto fn = it.next();

            QSettings sSep(fn, QSettings::IniFormat);
            CalConfig::reloadCalibration(sSep);
        }
    }
    QSettings s;
    CalConfig::reloadCalibration(s);
}


void CalConfig::reloadCalibration()
{
    allCal.clear();
    QSettings s(QSettings::IniFormat, QSettings::UserScope, QCoreApplication::organizationName(), "tdc2_cal");
    reloadCalibration(s);
}

void CalConfig::reloadCalibration(QSettings &s)
{
    const QString settingFileName = s.fileName();
    qInfo() << "CalConfig::reloadCalibration from:\n" << settingFileName;
    bool ok;
    for(const QString &group: s.childGroups()) {
        QString groupType;
        DeviceIndex devIndex = getDevIndexByGroup(group, groupType);
        if(devIndex.isNull())
            continue;

        DeviceCalibration &devCal = allCal[devIndex];
        s.beginGroup(group);

        if(group.endsWith(INL_PREFIX)) {
            const int nch = s.value("nch", 16).toInt(&ok);
            if(ok && nch) {
                int ch;
                QVector<double> corr;

                QList<int> read_ch;
                for(const QString &key: s.childKeys()) {
                    corr.reserve(1024);
                    ch = key.toInt(&ok);
                    if(ok) {
                        const QList<QVariant> &corrList = s.value(key).toList();
                        for(const QVariant &c: corrList) {
                            corr << c.toDouble(&ok);
                            if(!ok) {
                                qWarning()<< "Bad inl corr format in config";
                                break;
                            }
                        }
                        if(ok && corr.size()==1024) {
                            devCal.inl[ch] = corr;
                            devCal.inlSource[ch] = settingFileName;
                            read_ch << ch;
                        }

                        corr.clear();
                    }
                }
                std::sort(read_ch.begin(), read_ch.end());
                QString chMaskStr;
                QPair<int,int> range;
                range.first = 1000;
                for(int ch : read_ch) {
                    if(range.first == 1000) {
                        //first range
                        range.first=ch;
                        range.second=ch;
                    } else if(range.second+1 == ch) {
                        // continuation of range
                        range.second=ch;
                    } else {
                        // new range
                        if(range.first == range.second)
                            chMaskStr.append(QString(" %1").arg(range.first));
                        else
                            chMaskStr.append(QString(" %1-%2").arg(range.first).arg(range.second));
                        range.first=ch;
                        range.second=ch;
                    }
                }
                // Flush last range if exists
                if(range.first != 1000) {
                    if(range.first == range.second)
                        chMaskStr.append(" ").append(QString::number(range.first));
                    else
                        chMaskStr.append(QString(" %1-%2").arg(range.first).arg(range.second));
                }
                qInfo() << settingFileName << "Got inl" << devIndex.getIdent() << chMaskStr;
            }
        }

        s.endGroup();
    }
}

QString CalConfig::getInlSectionStr(const DeviceIndex &devIndex)
{
    return getSectionStr(devIndex, INL_PREFIX);
}

DeviceIndex CalConfig::getDevIndexByGroup(QString groupStr, QString &type)
{
    DeviceIndex index;
    if(groupStr.endsWith(INL_PREFIX)){
        type = INL_PREFIX;
    } else {
        return index;
    }
    QString devIndexStr = groupStr.left(groupStr.length()-type.length()-1); // "TQDC16VS_E-046F0C75"
    QString devType = devIndexStr.left(devIndexStr.length()-9); // '-' + 8 chars of serial
    QString devSerial = devIndexStr.right(8); // "046F0C75"
    bool ok;
    quint32 serial = devSerial.toUInt(&ok, 16);
    if(ok) {
        quint8 id=1;
        for(; id!=0; ++id) {
            if(devType == getDeviceTypeName(id)) {
                break;
            }
        }
        if(id)
            return DeviceIndex(static_cast<DeviceId>(id), serial);

        if(serial)
            return DeviceIndex(DEVICE_ID_AnyDevice, serial);
    }
    return DeviceIndex();
}
