#include "TcpClient.h"

#include <QtNetwork>
#include <QDebug>
#include <QSemaphore>

#if QT_VERSION < 0x050000
Q_DECLARE_METATYPE(QAbstractSocket::SocketError)
#endif

#include "histogramm/HistParams.h"

namespace {
const qint64 SOCKET_READ_BUFFER_SIZE = 1e7; // 10MB
}

TcpClient::TcpClient(QSemaphore *_analyzerSem, QObject *parent) :
    QObject(parent),
    analyzerSem(_analyzerSem),
    enabled(false),
    tcpSocket(new QTcpSocket(this)),
    hostPort(0),
    reconnectTimer(this),
    delayReadTimer(this),
    statusTimer(this),
    connectingAttempts(0),
    dataRecieved(0),
    lastDataRecieved(0)
{
    qRegisterMetaType<QAbstractSocket::SocketError>();
    tcpSocket->setReadBufferSize(SOCKET_READ_BUFFER_SIZE);

    connect(tcpSocket, &QTcpSocket::readyRead, this, &TcpClient::readData, Qt::QueuedConnection);
    connect(tcpSocket, &QTcpSocket::connected, this, &TcpClient::sendStatus, Qt::QueuedConnection);
    connect(tcpSocket, &QTcpSocket::disconnected, this, &TcpClient::disconnected, Qt::QueuedConnection);
    connect(tcpSocket, SIGNAL(error(QAbstractSocket::SocketError)),
            this, SLOT(displayError(QAbstractSocket::SocketError)), Qt::QueuedConnection);
    connect(&reconnectTimer, &QTimer::timeout, this, &TcpClient::requestNewData, Qt::QueuedConnection);
    reconnectTimer.setSingleShot(true);

    delayReadTimer.setSingleShot(true);
    delayReadTimer.setInterval(1000);
    connect(&delayReadTimer, &QTimer::timeout, this, &TcpClient::readData, Qt::QueuedConnection);

    statusTimer.setSingleShot(false);
    connect(&statusTimer, &QTimer::timeout, this, &TcpClient::sendStatus);
    statusTimer.start(1000);
    statusPeriod.start();
}


TcpClient::~TcpClient()
{
    stopReadOut();
    delete tcpSocket;
}


void TcpClient::requestNewData()
{
    try{
    if(!enabled) return;

    if(tcpSocket->isValid() &&
            (tcpSocket->state() == QTcpSocket::ConnectedState ||
             tcpSocket->state() == QTcpSocket::ConnectingState) )
        return;
    tcpSocket->abort();
    tcpSocket->connectToHost(hostAddr, hostPort);
    ++connectingAttempts;
    if(tcpSocket->waitForConnected(1000)){
        qInfo()<<"TcpClient connected.";
    } else {
        if(connectingAttempts<=1){
            QString mess("TcpClient can't connected.");
            qInfo()<<mess;
            addLog(mess);
            reconnectTimer.start(1000);
        } else
            reconnectTimer.start(5000);
    }
    } catch(...){
        qInfo()<<"Exception in TcpClient::requestNewData()";
    }
    sendStatus();
}


void TcpClient::readData()
{
    try {
    if(!enabled) return;
    int acquireCnt = 0;
    while(!analyzerSem->tryAcquire(1, 100)) {
        if(acquireCnt==0) {
            qWarning()<<QString("[%1] tcp read delayed (analyzer is slow)")
                     .arg(QTime::currentTime().toString("hh:mm:ss:zzz"));
        }
        if(++acquireCnt>10) {
            delayReadTimer.start(200);
            return;
        }
    }

    qint64 avail = tcpSocket->bytesAvailable();
    if(!avail) {
        analyzerSem->release();
//        qInfo()<<"no data "<<queryFreeSize.available();
        // readData() will be invoqed by readyRead() signal
        return;
    }

    QVector<QByteArray> dataArray;
    qint64 sumSize=0;
    while(avail) {
        qint64 freeSizeLeft = std::min(avail, PACK_SIZE_LIMIT-sumSize);
        QByteArray reciveData = tcpSocket->read(freeSizeLeft);
        if(reciveData.size() < freeSizeLeft) {
            qWarning() << "We got data lesser that expected.";
        }
        dataArray.push_back(reciveData);
        sumSize += reciveData.size();
        if(sumSize >= PACK_SIZE_LIMIT)
            break;

        avail = tcpSocket->bytesAvailable();
//        if(avail){
//            qInfo()<<"double recive";
//        }
    }
    dataRecieved += sumSize;
//    if(++counter%10==0){
//        qInfo() <<QString("analyzer=%1   writer=%2").arg(analyzerQueueSize->available(),3).arg(writerQueueSize->available(),3);
//    }

    if(dataArray.size()==1) {
        emit dataAccuired(dataArray[0]);
    } else {
        QByteArray summedData;
        for (QByteArray data: dataArray) {
            summedData.append(data);
        }
        emit dataAccuired(summedData);
    }

    } catch(...) {
        qWarning()<<"Exception in TcpClient::readData()";
        qWarning()<<"Possible loss of semaphores in TcpClient::readData";
    }
}


void TcpClient::displayError(QAbstractSocket::SocketError socketError)
{
    switch (socketError) {
    case QAbstractSocket::RemoteHostClosedError:
        qInfo() << "Remote hoste closed connection.";
        addLog("Remote hoste closed connection.");
        break;
    case QAbstractSocket::HostNotFoundError:
        qInfo() << "The host was not found. Please check the "
                    "host name and port settings.";
        addLog("The host was not found. Please check the "
               "host name and port settings.");
        break;
    case QAbstractSocket::ConnectionRefusedError:
        if(connectingAttempts<=1){
            qInfo() << tr("The connection was refused by the peer. "
                           "Make sure the data server is running, "
                           "and check that the host name and port "
                           "settings are correct.");
            addLog("The connection was refused by the peer. "
                   "Make sure the data server is running, "
                   "and check that the host name and port "
                   "settings are correct.");
        }
        break;
    default:
        qInfo() << tr("The following error occurred: %1.")
                    .arg(tcpSocket->errorString());
        addLog(tr("The following error occurred: %1.")
               .arg(tcpSocket->errorString()));
    }
}


void TcpClient::setAddress(QHostAddress addr, quint16 port)
{
    if(hostAddr != addr || hostPort != port) {
        if(!addr.isNull() && tcpSocket->isValid() &&
                (tcpSocket->state() == QTcpSocket::ConnectedState ||
                 tcpSocket->state() == QTcpSocket::ConnectingState) ) {
            return;
        }

        stopReadOut();
        hostAddr = addr;
        hostPort = port;
        if(hostAddr.isNull()) {
            qInfo()<<"TDC2 TcpClient::setAddress(): stop readout";
        } else {
            qInfo()<<"TDC2 TcpClient::setAddress:"<<addr.toString()<<port;
            startReadOut();
        }
    }
}


void TcpClient::startReadOut()
{
    enabled = true;
//    requestNewData();
    reconnectTimer.start(0);
}


void TcpClient::stopReadOut()
{
    enabled = false;
    if(tcpSocket->state() != QTcpSocket::UnconnectedState)
        tcpSocket->disconnectFromHost();
    reconnectTimer.stop();
    sendStatus();
}

void TcpClient::disconnected()
{
    if(enabled) {
        qInfo()<<"Client was disconnected, trying to reconnect";
        connectingAttempts=0;
        reconnectTimer.start(1000);
    } else
        qInfo()<<"Client disconnected.";

    sendStatus();
}

void TcpClient::sendStatus()
{
    status.connected = enabled && tcpSocket!=nullptr && tcpSocket->state() == QTcpSocket::ConnectedState;

    if(status.connected) {
        status.host = tcpSocket->localAddress();
        status.port = tcpSocket->localPort();
    } else {
        status.host = {};
        status.port = 0;
    }
//    if(connOk)
//        qInfo() << "TcpClient::bytesAvailable" << 1e-6*tcpSocket->bytesAvailable() << "MB";
    status.dataRate = 1e3*(dataRecieved-lastDataRecieved)/statusPeriod.restart();
    lastDataRecieved=dataRecieved;

    emit statusUpdated(status);
//    qInfo() << "TcpClient: status "<< (connOk?"connected":"disconnected") << dataRecieved;
}
