#include "ViewerMainWindow.h"
#include "ui_ViewerMainWindow.h"

#include <QBoxLayout>
#include <QFileDialog>
#include <QHash>
#include <QHostAddress>
#include <QInputDialog>
#include <QLabel>
#include <QPlainTextEdit>
#include <QSemaphore>
#include <QSettings>
#include <QShortcut>
#include <QTreeWidgetItem>
#include <QThread>

#include "AfiViewerConfigDb.h"
#include "AfiViewerSignalRouter.h"
#include "AfiViewerConfigDispatcher.h"
#include "HptdcVoltageTest.h"
#include "CalibrationDialog.h"
#include "SaveInlDialog.h"
#include "file-reader-dialog/FileReaderDialog.h"
#include "base/ThreadController.h"
#include "base/threadQuit.h"
#include "add-pnp-dialog/AddPnpDialog.h"
#include "TcpClient.h"
#include "pnp-server/PnpDiscovery.h"
#include "InlCheckDialog.h"
#include "ViewerZmqPublisher.h"

#include "base/Globals.h"
#include "histogramm/DeviceHelper.h"
#include "histogramm/HistMngrCfg.h"
#include "histogramm/HistParams.h"
#include "histogramm/HistWorker.h"
#include "histogramm/TQDCAnalyzer.h"
#include "widgets/LogDialog.h"
#include "zmq-lib/ZmqDialog.h"

#include <qwt_plot.h>
#include <qwt_plot_curve.h>
#include <qxw/WaveformPlot.h>
#include "TdcTtvxsMapDialog.h"
#include "StatusDialog.h"

namespace {
enum tree_tdc_col {
    tree_tdc_name,
    tree_tdc_ch,
    tree_tdc_inv,
    tree_tdc_inl,
    tree_tdc_ttvxs,
    tree_tdc_raw_cnt,
    tree_tdc_tdc_hits,
    tree_tdc_msc_hits,
    tree_tdc_num_cols
};

const DaqModuleSet SOURCE_TYPES = {
    DaqModule(DaqModule::Type::EventBuilder),
    DaqModule(DaqModule::Type::MStream),
    DaqModule(DaqModule::Type::VmeDaq)
};
void setLabelColor(QLabel * const lab, QColor col) {
    QPalette p = lab->palette();
    p.setColor(lab->foregroundRole(), col);
    lab->setPalette(p);
}

QString inlMark(int st) {
    switch (st) {
    case  1: return QChar('V'); // V
    case -1: return QChar('X'); // X
    default: return QChar('?');
    }
}
QColor inlCol(int st) {
    switch (st) {
    case  1: return Qt::darkGreen;
    case -1: return Qt::red;
    default: return Qt::black;
    }
}
void updateItemInl(QTreeWidgetItem * item, int st){
    item->setText(tree_tdc_inl, inlMark(st));
    auto brush = item->foreground(tree_tdc_inl);
    brush.setColor(inlCol(st));
    item->setForeground(tree_tdc_inl, brush);
}
void updateItemTtvxs(QTreeWidgetItem * item, int st){
    item->setText(tree_tdc_ttvxs, inlMark(st));
    auto brush = item->foreground(tree_tdc_ttvxs);
    brush.setColor(inlCol(st));
    item->setForeground(tree_tdc_ttvxs, brush);
}

QList<ProgramInterface> getFreeProgramInterfaceList(const ProgramDescription &pd)
{
    QList<ProgramInterface> list;

    auto fillInt = [&](ProgramInterfaceType type) {
        for(ProgramInterface i : pd.interfaces) {
            if(!i.isFree || !i.enabled)
                continue;
            if(i.type == ProgramInterface::getTypeName(type)) {
                list << i;
            }
        }
    };

    fillInt(ProgramInterfaceMonitorData);
    if(pd.getDaqModule() == DaqModule(DaqModule::Type::MStream)
        || Globals::getDebugMode()) {
        fillInt(ProgramInterfaceDataFlow);
    }

    return list;
}

bool isFreeProg(const ProgramDescription &pd)
{
    return !getFreeProgramInterfaceList(pd).isEmpty();
}
}

ViewerMainWindow::ViewerMainWindow(AfiViewerStartupOptions startup_options, QWidget *parent) :
    BaseMainWindow(parent),
    ui(new Ui::ViewerMainWindow),
    analyzerSem(new QSemaphore(DATA_QUEUE_SIZE)),
    histSem(new QSemaphore(DATA_QUEUE_SIZE)),
    writerSem(new QSemaphore(DATA_QUEUE_SIZE)),
    zmqSem(new QSemaphore(DATA_QUEUE_SIZE)),
    configuration_manager(new ConfigurationManager(this)),
    afiViewerConfigDb(new AfiViewerConfigDb(startup_options, this)),
    afiViewerSignalRouter(new AfiViewerSignalRouter()),
    configDispatcher(new AfiViewerConfigDispatcher(this)),
    fileReaderDialog(new FileReaderDialog(analyzerSem, this)),
    dataSourceLabel(new QLabel(this)),
    dataRateLabel(new QLabel(this)),
    analyzerStatusLabel(new QLabel(this)),
    semStatusLabel(new QLabel(this)),
    vxsClockLabel(new QLabel(this)),
    formatErrorLabel(new QLabel(this)),
    logDialog(new LogDialog(this)),
    calibrationDialog(new CalibrationDialog(this)),
    inlCheckDialog(startup_options.inl ? new InlCheckDialog(this) : nullptr),
    tcpClient(new TcpClient(analyzerSem)),
    tcpClientThread(new QThread(this)),
    analyzer(new TQDCAnalyzer(analyzerSem, histSem, writerSem, zmqSem)),
    analyzerThread(new QThread(this)),
    worker(new HistWorker(histSem)),
    worker_thread(new QThread(this)),
    rootTreeItem(new QTreeWidgetItem(QStringList("All"))),
    rawDataPeriodicTimer(new QTimer(this)),
    statusTimer(new QTimer(this)),
    statusDialog(new StatusDialog(this)),
    zmqPublisher(new ViewerZmqPublisher(zmqSem)),
    zmq_thread(new QThread(this)),
    zmqDialog(new ZmqDialog(this))
{
    qRegisterMetaType<DeviceIndex>();
    qRegisterMetaType<QMap<DeviceIndex, QVector<bool> > >();
    qRegisterMetaType<TdcTtvxsMapType>();
    qRegisterMetaType<QHostAddress>();
    qRegisterMetaType<DeviceCalibration>();
    qRegisterMetaType<InlResults>();
    qRegisterMetaType<QMap<int,double> >();
    qRegisterMetaType<QMap<int,int> >();
    qRegisterMetaType<DecodedData>();
    qRegisterMetaType<AnalyzerStat>();
    qRegisterMetaType<AnalyzerStat>();
    qRegisterMetaType<InlCheckResults>();
    qRegisterMetaType<HistKey>();
    qRegisterMetaType<Histogramm>();
    qRegisterMetaType<QHash<HistKey, Histogramm>>();
    qRegisterMetaType<QMap<DeviceIndex,int>>();
    qRegisterMetaType<TcpClientStatus>();

    ui->setupUi(this);

    afiViewerSignalRouter->setMainWindow(this);
    afiViewerSignalRouter->setZmqPublisher(zmqPublisher);
    afiViewerSignalRouter->setZmqDialog(zmqDialog);
    afiViewerSignalRouter->setHistWorker(worker);
    afiViewerSignalRouter->setHistWidget(ui->widgetHistMngr);
    afiViewerSignalRouter->setConfigDispatcher(configDispatcher);
    afiViewerSignalRouter->setConfigDb(afiViewerConfigDb);
    afiViewerSignalRouter->setConfigurationManager(configuration_manager);
    afiViewerSignalRouter->connectAll();
    //TODO:
//    afiViewerSignalRouter->setRemoteControlServer(remoteControlServer);
    setupMenu(ui->menuFile, ui->menuHelp, ui->menuOptions);

    if(!startup_options.inl) {
        ui->menuHPTDC_INL->menuAction()->setVisible(false);
        ui->actionHPTDC_vol_check->setVisible(startup_options.inl);
    }
    ui->plainTextEditRawData->setFont(QFont("monospace"));

    statusBar()->addWidget(dataSourceLabel,10);
    statusBar()->addWidget(dataRateLabel,3);
    statusBar()->addWidget(analyzerStatusLabel,10);
    statusBar()->addWidget(formatErrorLabel,3);
    statusBar()->addWidget(vxsClockLabel,5);
    statusBar()->addWidget(semStatusLabel,4);
    analyzerStatusLabel->setToolTip("Analyzer status");
    setLabelColor(formatErrorLabel, Qt::red);
    semStatusLabel->setToolTip("Queue's semaphores");

    setWindowTitle(QCoreApplication::applicationName());

    QShortcut *shortcut = new QShortcut(QKeySequence(Qt::CTRL+Qt::Key_L), this);
    connect(shortcut, &QShortcut::activated, logDialog, &QWidget::show);

    analyzer->moveToThread(analyzerThread);
    tcpClient->moveToThread(tcpClientThread);
    worker->moveToThread(worker_thread);
    zmqPublisher->moveToThread(zmq_thread);

    analyzerThread->setObjectName("Analizer");
    tcpClientThread->setObjectName("TcpClient");
    worker_thread->setObjectName("Worker");
    zmq_thread->setObjectName("0MQ thread");

    connect(analyzer, &QObject::destroyed, analyzerThread, &QThread::quit);
    connect(tcpClient, &QObject::destroyed, tcpClientThread, &QThread::quit);
    connect(worker, &QObject::destroyed, worker_thread, &QThread::quit);
    connect(zmqPublisher, &QObject::destroyed, zmq_thread, &QThread::quit);
    connect(worker, &HistWorker::workerStatUpdated, statusDialog, &StatusDialog::updateWorkerStats);
    connect(analyzer, &TQDCAnalyzer::deviceDataUpdated, worker, &HistWorker::dataUpdated);
    connect(analyzer, &TQDCAnalyzer::deviceStatsUpdated, this, &ViewerMainWindow::updateHitStats);
    connect(analyzer, &TQDCAnalyzer::deviceStatsUpdated, statusDialog, &StatusDialog::updateHitStats);
    connect(analyzer, &TQDCAnalyzer::rawDataUpdated, this, &ViewerMainWindow::rawDataUpdated);
    connect(this, &ViewerMainWindow::rawDataEveryEvUpdated, analyzer, &TQDCAnalyzer::setRawDataEveryEv);
    connect(this, &ViewerMainWindow::requestRawData, analyzer, &TQDCAnalyzer::requestRawData);
    connect(this, &ViewerMainWindow::resetAll, analyzer, &TQDCAnalyzer::reset);
    connect(this, &ViewerMainWindow::resetAll, statusDialog, &StatusDialog::reset);
    connect(this, &ViewerMainWindow::resetAll, worker, &HistWorker::clearAll);
    connect(this, &ViewerMainWindow::resetAll, zmqPublisher, &ZmqPublisher::clearStat);
    connect(this, &ViewerMainWindow::resetAll, ui->widgetWaveform, &WaveformViewer::reset);
    connect(this, &ViewerMainWindow::set_can_skip, analyzer, &TQDCAnalyzer::set_can_skip);

    connect(fileReaderDialog, &FileReaderDialog::dataAccuired, analyzer, &TQDCAnalyzer::dataAccuired);
    connect(this, &ViewerMainWindow::closeFile, fileReaderDialog, &FileReaderDialog::onCloseFile);
    connect(this, &ViewerMainWindow::closeAll, tcpClient, &QObject::deleteLater);
    connect(this, &ViewerMainWindow::closeAll, analyzer, &QObject::deleteLater);
    connect(this, &ViewerMainWindow::closeAll, worker, &QObject::deleteLater);
    connect(this, &ViewerMainWindow::closeAll, zmqPublisher, &QObject::deleteLater);

    connect(this, &ViewerMainWindow::tcpAddressChanged, tcpClient, &TcpClient::setAddress);
    connect(tcpClient, &TcpClient::dataAccuired, analyzer, &TQDCAnalyzer::dataAccuired);
    connect(tcpClient, &TcpClient::statusUpdated, this, &ViewerMainWindow::tcpStatusUpdated);

    connect(analyzer, &TQDCAnalyzer::deviceDataUpdated,
            zmqPublisher, &ViewerZmqPublisher::publish_event);

    analyzerThread->start();
    tcpClientThread->start();
    worker_thread->start();
    zmq_thread->start();
    connect(analyzer, &TQDCAnalyzer::deviceDataUpdated,
            ui->widgetWaveform, &WaveformViewer::deviceDataUpdated);

    if(inlCheckDialog) {
        connect(inlCheckDialog, &InlCheckDialog::retry,
                ui->widgetHistMngr, &HistMngrWidget::inlCheckRequest);
        connect(ui->widgetHistMngr, &HistMngrWidget::inlCheckUpdated,
                inlCheckDialog, &InlCheckDialog::resultUpdated);
    }

    connect(rawDataPeriodicTimer, &QTimer::timeout, this, &ViewerMainWindow::rawDataPeriodicRequest);
    rawDataPeriodicTimer->setSingleShot(true);

    connect(this, &ViewerMainWindow::invertUpdated, analyzer, &TQDCAnalyzer::invertUpdated);
    connect(statusTimer, &QTimer::timeout, this, &ViewerMainWindow::updateStatus);

    statusTimer->setSingleShot(false);
    statusTimer->start(500);
    rawEventCntTimer.start();

    initTree();
    read_gui_config();
    updateSourceLabel();

    configuration_manager->setDefaultConfig(AfiViewerConfig());
}

ViewerMainWindow::~ViewerMainWindow()
{
    delete afiViewerSignalRouter;

    write_gui_config();

    emit closeAll();
    QElapsedTimer closeTimer;
    closeTimer.start();
    if(!threadQuiting(analyzerThread, 2000, &closeTimer)) {
        qWarning() << "Can't stop analyzer thread";
    }
    if(!threadQuiting(tcpClientThread, 2000, &closeTimer)) {
        qWarning() << "Can't stop TCP-readout thread";
    }
    if(!threadQuiting(worker_thread, 2000, &closeTimer)) {
        qWarning() << "Can't stop worker thread";
    }
    if(!threadQuiting(zmq_thread, 2000, &closeTimer)) {
        qWarning() << "Can't stop 0MQ thread";
    }

    delete ui;
}

FileReaderDialog *ViewerMainWindow::getFileReaderDialog() const
{
    return fileReaderDialog;
}

WaveformViewer *ViewerMainWindow::getWaveformViewer() const
{
    return ui->widgetWaveform;
}

void ViewerMainWindow::pnpCacheUpdated(const PnpInfoCache &cache)
{
    pnpCache = cache;
    for(auto pd : pnpCache) {
        if(pd.getClientIndex() == selectedProgram) {
            updateTcpInterface(pd);
        }
    }

    ui->actionConnect_to_source_program->setEnabled(hasFreeProgs());
}

void ViewerMainWindow::setConfig(const AfiViewerConfig &newConfig)
{
    static bool initial_setting = true;
    auto prevSource = config.source;
    bool doReloadCalib = config.useSeparateInl != newConfig.useSeparateInl
        || config.separateInlFolder != newConfig.separateInlFolder;
    config = newConfig;
    if(doReloadCalib) {
        reloadCalCalibration();
    }

    updateGuiToConfig();
    if(prevSource != config.source)
        onSourceStateChange(prevSource, config.source);

    updateSourceLabel();
    if(config.source == AfiViewerConfig::SOURCE_DATA::SOURCE_FILE) {
        if(initial_setting) {
            //todo: swap upper if (after histogrammer fix)
            initial_setting = false;
            fileReaderDialog->show();
            fileReaderDialog->setFileReaderConfig(config.file_reader_config);
            fileReaderDialog->reopenLastFile();
        }
    }
}

void ViewerMainWindow::on_actionConfiguration_manager_triggered()
{
    configuration_manager->show();
}

void ViewerMainWindow::configSelectorChanged(const ConfigIndexName &conf)
{
    setWindowTitle(ConfigSelector(DaqModule(DaqModule::Type::AfiViewer), conf)
                   .toWindowTitle());
}

void ViewerMainWindow::on_actionSet_source_file_triggered()
{
    ui->actionDisconnect_from_program->trigger();
    fileReaderDialog->show();
}

void ViewerMainWindow::on_actionConnect_to_source_program_triggered()
{
    if(!hasFreeProgs()) {
        ui->actionConnect_to_source_program->setEnabled(false);
        return;
    }

    AddPnpDialog dialog(pnpCache, false, this);
    dialog.set_function_filter(std::bind(&isFreeProg, std::placeholders::_1));
    PnpDiscovery::subscribeCacheUpdaed(&dialog);

    if(dialog.exec() != QDialog::Accepted)
        return;

    auto pd_list = dialog.getProgDescr();
    Q_ASSERT(!pd_list.isEmpty());
    if(pd_list.isEmpty()) return;
    ProgramDescription pd = pd_list.first();
    if(pd.getClientIndex() != selectedProgram || selectedInterface.isEmpty()) {
        auto list = getFreeProgramInterfaceList(pd);
        if(!list.isEmpty()) {
            ProgramInterface newInterface;
            if(list.size() > 1) {
                QStringList items;
                for(auto i : list) {
                    items << QString("%2#%3 %4:%5")
                                 .arg(i.type).arg(i.id)
                                 .arg(i.host.toString()).arg(i.port);
                }

                bool ok;
                QString item = QInputDialog::getItem(this, tr("Choose source interface..."),
                                                     tr("Source interface:"), items, 0, false, &ok);
                if(!ok)
                    return;
                int id = items.indexOf(item);
                newInterface = list.value(id);
            } else
                newInterface = list.first();

            qDebug().noquote()
                << QString("[MW] change source from %1 to %2 (%3)")
                    .arg(selectedProgram.toString()).arg(pd.getName()).arg(newInterface.type);
            disconnectFromProg();
            selectedProgram = pd.getClientIndex();
            selectedInterface = newInterface;
            config.source_program = selectedProgram;
            config.sourceIntType = selectedInterface.type;
            config.sourceIntId = selectedInterface.id;
            changeSourceState(AfiViewerConfig::SOURCE_DATA::SOURCE_STREAM);
            emit configChanged(config);
            emit rawDataEveryEvUpdated(config.rawDataEveryEv);
        }
        updateTcpInterface(pd);
        emit configChanged(config);
    } else {
        qWarning()<<"Selected program has no free interface";
    }
}

void ViewerMainWindow::on_actionDisconnect_from_program_triggered()
{
    disconnectFromProg();
    changeSourceState(AfiViewerConfig::SOURCE_DATA::SOURCE_IDLE);
}

void ViewerMainWindow::on_pushButtonResetAll_clicked()
{
    emit resetAll();
}

void ViewerMainWindow::tcpStatusUpdated(const TcpClientStatus &status)
{
    tcpClientStatus = status;
    if(config.source == AfiViewerConfig::SOURCE_DATA::SOURCE_STREAM)
        setSourceColor(tcpClientStatus.connected ? "darkGreen" : "red");

    dataRateLabel->setText(QString("%1 kB/s")
                           .arg(tcpClientStatus.dataRate/1024, 0, 'f', 2));
//    qInfo() << QString("Sem status: anal=%1; writer=%2; hist=%3")
//               .arg(analyzerSem->available())
//               .arg(writerSem->available())
//               .arg(histSem->available());
}

void ViewerMainWindow::updateHitStats(const AnalyzerStat &stats)
{
//    qInfo() << "Decoded events:" << stats.eventDecoded << stats.lastEvNum;
    updateTreeWithStats(stats);

    QString analStr = QString("Ev:%1").arg(stats.eventDecoded);
    if(lastAnalStat.eventDecoded+1 != stats.eventDecoded) {
        analStr += QString("(lastEv=%2)").arg(lastAnalStat.eventDecoded);
    }
    {
        qint64 dt = stats.decodeTimeNs-lastAnalStat.decodeTimeNs;
        qint64 ev = stats.eventDecoded-lastAnalStat.eventDecoded;
        qint64 size = stats.sizeDecoded-lastAnalStat.sizeDecoded;
//        qInfo() << "dt:" << dt;
        analStr.append(QString("rate(%1ev/s; %2MB/s)").arg(ev? (1e9*ev/dt) : 0.,0,'f',0)
                       .arg(size? (1e3*size/dt) : 0.,0,'f',2));
    }
//    if(stats.eventDecoded) {
//        analStr += QString("; Time per ev:%1us; Event: %2B; Size per sec: %3KB")
//                .arg(1e-3*stats.decodeTimeNs/stats.eventDecoded)
//                .arg(stats.sizeDecoded/stats.eventDecoded)
//                .arg(1e6*stats.sizeDecoded/stats.decodeTimeNs);
//    }

    analyzerStatusLabel->setText(analStr);
    if(stats.formatErrorCnt != lastAnalStat.formatErrorCnt) {
        formatErrorLabel->setVisible(stats.formatErrorCnt);
        formatErrorLabel->setText(QString("FormatErr:%L1").arg(stats.formatErrorCnt));
    }


    lastAnalStat = stats;
}

void ViewerMainWindow::rawDataUpdated(const QStringList &data)
{
    ++rawEventCnt;
    plaint_text = data;
    ui->plainTextEditRawData->setPlainText(data.join('\n'));
    rawDataReqDone = true;
    ui->plainTextEditRawData->setEnabled(rawDataReqDone);
    ui->pushButtonRequestRaw->setEnabled(!config.rawDataEveryEv && rawDataReqDone);

    if(!config.rawDataEveryEv && config.rawDataPeriodic) {
        rawDataPeriodicTimer->start();
    } else {
        rawDataPeriodicTimer->stop();
    }
}

void ViewerMainWindow::on_actionCalibrationData_triggered()
{
    calibrationDialog->show();
}

void ViewerMainWindow::on_actionReset_all_triggered()
{
    emit resetAll();
}

void ViewerMainWindow::inl_corrections_updated(InlResults res, bool allRes)
{
    if(allRes) {
        // actionHPTDC_vol_check
        HptdcVoltageTest test;
        QHash<HistKey, Histogramm> hist_container = ui->widgetHistMngr->getHists();
        test.print(config.file_reader_config.lastFile, lastAnalStat, hist_container,
                   &config.hist_manager_config, res);
        return;
    }
    SaveInlDialog dialog(this);
    dialog.setInlResults(res);
    if(dialog.exec()) {
        if(dialog.result() == QDialog::Accepted) {
            auto newInl = dialog.getInlResults();
            auto path = dialog.getPath();
            calConfig.calibrationUpdated(newInl, path);
        }
    }
}

void ViewerMainWindow::on_actionSave_INL_corrections_triggered()
{
    emit requestInlResults(false);
}

void ViewerMainWindow::on_actionShow_INL_check_triggered()
{
    if(inlCheckDialog)
        inlCheckDialog->show();
}

void ViewerMainWindow::on_actionSet_INL_separate_location_toggled(bool checked)
{
    if(checked) {
        auto dir = QFileDialog::getExistingDirectory(this, "Set folder with separate INL calibrations...", config.separateInlFolder);
        if(dir.isEmpty()) {
            ui->actionSet_INL_separate_location->setChecked(false);
            return;
        }
        config.separateInlFolder = dir;
        config.useSeparateInl = true;
    } else {
        config.useSeparateInl = false;
    }
    reloadCalCalibration();
    emit configChanged(config);
//    ui->widgetHistMngr->updateInlCorr(config.allCal);
    emit inlConfigUpdated(calConfig.allCal);
    updateInlMarks();
}

void ViewerMainWindow::on_actionHPTDC_vol_check_triggered()
{
    emit requestInlResults(true);
}

void ViewerMainWindow::on_actionSet_TDC_TTVXS_map_triggered()
{
    TdcTtvxsMapDialog dialog(lastAnalStat, config.tdc_ttvxs_map, this);
    if(dialog.exec()){
        config.tdc_ttvxs_map = dialog.getDevMap();
        emit configChanged(config);
        emit tdcTtvxMapUpdated(config.tdc_ttvxs_map);
        updateTtvxsMarks();
    }
}

void ViewerMainWindow::on_actionSatistic_dialog_triggered()
{
    statusDialog->show();
}

void ViewerMainWindow::on_action0MQ_publisher_triggered()
{
    zmqDialog->show();
}

void ViewerMainWindow::on_actionHPTDC_errors_triggered()
{
    QDialog *d = new QDialog(this, Qt::Dialog);
    QPlainTextEdit scrollArea;
    QLabel htmlLabel;

    QFile htmlFile(":/rawDataFormat.html");
    QString html;
    if(htmlFile.open(QIODevice::ReadOnly)){
        html = QString(htmlFile.readAll());
        htmlFile.close();
        if(html.isEmpty())
            html = "<b>Channel map file is empty.</b>";
    } else
        html = "<b>There is no channel map file.</b>";

    scrollArea.appendHtml(html);
    htmlLabel.setText(html);

    QVBoxLayout *mainLayout = new QVBoxLayout(d);
    //    mainLayout->addWidget(&scrollArea);
    mainLayout->addWidget(&htmlLabel);
    d->setLayout(mainLayout);
    d->show();

    //    QMessageBox::about (this, "Raw data format", html);
}

//void ViewerMainWindow::on_actionTxt_data_format_triggered()
//{
//    QString s;
//    s += "<h2>TDC2 Txt Data format</h2>";
//    s += "<p>File constructs with following lines:</p>";

//    s += "<ul>";
//    s += "<li>Ev <i>&lt;evNumber&gt;</i> <i>&lt;TAI_sec&gt;</i>.<i>&lt;TAI_ns&gt;</i></li>";
//    s += "<li>Tdc <i>&lt;ch&gt;</i>: <i>&lt;tdcLeadTime&gt;</i></li>";
//    s += "<li>Adc <i>&lt;ch&gt;</i>: <i>&lt;adcTime&gt;</i> <i>&lt;sample1&gt;</i> <i>&lt;sample2&gt;</i> ... <i>&lt;sampleN&gt;</i></li>";
//    s += "</ul>";
//    s += "<p>where:</p>";
//    s += "<dl>";
//    s += "<dt><b>TAI time</b></dt>";
//    s += "<dd><i>Precise arrival time of the trigger</i></dd>";
//    s += "<dt><b>ch</b></dt>";
//    s += "<dd><i>Channel number in range 0-15</i></dd>";
//    s += "<dt><b>tdcLeadTime</b></dt>";
//    s += "<dd><i>HPTDC time in 0.024/1.024ns steps from trigger (TAI time)</i></dd>";
//    s += "<dt><b>adcTime</b></dt>";
//    s += "<dd><i>Timestamp of the first sample in 8ns steps from trigger (TAI time)</i></dd>";
//    s += "<dt><b>sample</b></dt>";
//    s += "<dd><i>ADC sample in signed 16-bit integer</i></dd>";
//    s += "</dl>";

//    QMessageBox::about (this, "Txt data format", s);
//}

//void ViewerMainWindow::on_actionRaw_data_format_triggered()
//{
//    QDialog d;
//    QPlainTextEdit scrollArea;
//    QLabel htmlLabel;

//    QFile htmlFile(":/rawDataFormat.html");
//    QString html;
//    if(htmlFile.open(QIODevice::ReadOnly)){
//        html = QString(htmlFile.readAll());
//        htmlFile.close();
//        if(html.isEmpty())
//            html = "<b>Channel map file is empty.</b>";
//    } else
//        html = "<b>There is no channel map file.</b>";

//    scrollArea.appendHtml(html);
//    htmlLabel.setText(html);

//    QVBoxLayout *mainLayout = new QVBoxLayout(&d);
////    mainLayout->addWidget(&scrollArea);
//    mainLayout->addWidget(&htmlLabel);
//    d.setLayout(mainLayout);
//    d.exec();

////    QMessageBox::about (this, "Raw data format", html);
//}

void ViewerMainWindow::on_radioButtonRawEvery_toggled(bool checked)
{
    config.rawDataEveryEv = checked;
    emit configChanged(config);
    emit rawDataEveryEvUpdated(config.rawDataEveryEv);

    updateRawDataControl();
}

void ViewerMainWindow::on_pushButtonRequestRaw_clicked()
{
    rawDataReqDone = false;
    ui->pushButtonRequestRaw->setEnabled(!config.rawDataEveryEv && rawDataReqDone);
    ui->plainTextEditRawData->setEnabled(rawDataReqDone);
    emit requestRawData(config.rawDataLimit);
}

void ViewerMainWindow::on_spinBoxRawEvents_valueChanged(int limit)
{
    config.rawDataLimit = limit;
    emit configChanged(config);
}

void ViewerMainWindow::on_checkBoxRawPeriodical_toggled(bool checked)
{
    config.rawDataPeriodic = checked;
    emit configChanged(config);
    updateRawDataControl();
    if(!config.rawDataEveryEv && config.rawDataPeriodic)
        rawDataPeriodicTimer->start();
    else
        rawDataPeriodicTimer->stop();
}

void ViewerMainWindow::on_spinBoxRawTimer_valueChanged(int val)
{
    if(config.rawDataPeriod != val) {
        config.rawDataPeriod = val;
        emit configChanged(config);
    }
    rawDataPeriodicTimer->setInterval(config.rawDataPeriod*1000);
}

void ViewerMainWindow::rawDataPeriodicRequest()
{
    if(!config.rawDataEveryEv && rawDataReqDone && config.rawDataPeriodic)
        on_pushButtonRequestRaw_clicked();
}

void ViewerMainWindow::updateStatus()
{
    double rate = 1e3*rawEventCnt/rawEventCntTimer.restart();
    rawEventCnt=0;
    if(rate>100 && config.rawDataEveryEv) {
        qWarning() << QString("Cancel rawData on every event; rate:%1Hz").arg(rate);
        ui->radioButtonRawRequest->setChecked(true);
    }

    semStatusLabel->setText(QString("a:%1;h:%2;w:%3")
                            .arg(analyzerSem->available())
                            .arg(histSem->available())
                            .arg(writerSem->available()));
}

void ViewerMainWindow::write_gui_config()
{
    QSettings settings;
    settings.beginGroup("ViewerMainWindow");
    settings.setValue("geometry", saveGeometry());
    settings.setValue("windowState", saveState());
    settings.setValue("splitterState", ui->splitter->saveState());
    settings.setValue("tabIndex", ui->tabWidget->currentIndex());
    settings.endGroup();
}

void ViewerMainWindow::read_gui_config()
{
    QSettings settings;
    settings.beginGroup("ViewerMainWindow");
    restoreGeometry(settings.value("geometry").toByteArray());
    restoreState(settings.value("windowState").toByteArray());
    ui->splitter->restoreState(settings.value("splitterState").toByteArray());
    int tabCurrentIndex = settings.value("tabIndex").toInt();
    storageDir = settings.value("storageDir").toString();
    settings.endGroup();
    ui->tabWidget->setCurrentIndex(tabCurrentIndex);
}

void ViewerMainWindow::initTree()
{
    QTreeWidget *tree = ui->treeWidgetChCfg;
    tree->setColumnCount(tree_tdc_num_cols);
    QStringList headerLabels;
    headerLabels << "name"
                 << "abs.ch"
                 << "Inv"
                 << "INL"
                 << "TTVXS"
                 << "Raw Count"
                 << "TDC Hits"
                 << "MSC";
    tree->setHeaderLabels(headerLabels);
    tree->hideColumn(tree_tdc_ch);
    tree->hideColumn(tree_tdc_raw_cnt);
    tree->insertTopLevelItem(0, rootTreeItem);
    rootTreeItem->setExpanded(true);
    rootTreeItem->setFlags(rootTreeItem->flags()|Qt::ItemIsUserCheckable);
    rootTreeItem->setCheckState(tree_tdc_inv, Qt::PartiallyChecked);
    tree->resizeColumnToContents(tree_tdc_name);
    tree->resizeColumnToContents(tree_tdc_inv);
    tree->resizeColumnToContents(tree_tdc_inl);
    tree->resizeColumnToContents(tree_tdc_ttvxs);
}

void ViewerMainWindow::refreshModChNum(const AnalyzerStat &analStat)
{
    bool modified = false;
    for(auto it=analStat.dev_stat.cbegin(); it!=analStat.dev_stat.cend(); ++it) {
        const auto &index = it.key();
        auto modIt = modChNum.find(index);
        if(modIt == modChNum.end()) {
            auto nch = getTdcChNum(index);
            nch = std::max(nch, getAdcChNum(index));
            modIt = modChNum.insert(index, nch);
            modified = true;
        }
        auto &modNch = modIt.value();
        for(auto c : it->dev_ch_hits.keys()) {
            if(modNch < c+1) {
                modified = true;
                modNch = c+1;
            }
        }
    }
    for(auto i : modChNum.keys()) {
        if(!analStat.dev_stat.contains(i)) {
            modChNum.remove(i);
            modified = true;
        }
    }

    if(modified)
        updateModelList();
}

void ViewerMainWindow::updateModelList()
{
    QTreeWidget *tree = ui->treeWidgetChCfg;

    tree->blockSignals(true);
    for (QTreeWidgetItem *i: modTreeItem.values()) {
        rootTreeItem->removeChild(i);
        delete i;
    }
    modTreeItem.clear();

    int modChOffest = 0;
    int devCnt = 0;
    for (auto it = modChNum.cbegin(); it != modChNum.cend(); ++it) {
        const auto &index = it.key();
        const int modNch = it.value();
        QStringList list;
        list << QString("%1: %2").arg(++devCnt).arg(index.getIdent())
             << index.getSerialStr()
             << "" << "-" << "0" << "0";
        QTreeWidgetItem * const modItem = new QTreeWidgetItem(rootTreeItem, list);
        modItem->setFlags(modItem->flags()|Qt::ItemIsUserCheckable);
        modItem->setCheckState(tree_tdc_inv, Qt::Unchecked);
        modItem->setExpanded(false);
        modTreeItem[index] = modItem;
//        ui->widgetHistMngr->updateInlCorr(index, config.allCal.value(index).inl);
//        emit inlConfigUpdated(config.allCal);

        for(int ch=0; ch<modNch; ++ch)
            addChTreeItem(index, modItem, ch);

        modChOffest += modNch;
    }
    tree->blockSignals(false);
    tree->resizeColumnToContents(tree_tdc_name);
    tree->resizeColumnToContents(tree_tdc_ch);
    updateTreeItemStates();
    updateInlMarks();
    updateTtvxsMarks();
}

QTreeWidgetItem* ViewerMainWindow::addChTreeItem(const DeviceIndex &index, QTreeWidgetItem *modItem, int ch)
{
    QStringList list;
    list << QString::number(ch)
         << "<absCh>"
         << "" << "-" << "-" << "-";
    auto treeItem = new QTreeWidgetItem(modItem, list);
    treeItem->setData(tree_tdc_name, Qt::UserRole, QVariant::fromValue(index));
    treeItem->setData(tree_tdc_name, Qt::UserRole+1, ch);
    treeItem->setFlags(treeItem->flags()|Qt::ItemIsUserCheckable);
    treeItem->setCheckState(tree_tdc_inv,
                            config.invertMap.value(index).value(ch, false)
                            ? Qt::Checked :Qt::Unchecked);
    return treeItem;
}

void ViewerMainWindow::on_treeWidgetChCfg_itemChanged(QTreeWidgetItem *item, int column)
{
    if(column != tree_tdc_inv)
        return;
    //todo
//    bakConfig = config;
    Qt::CheckState state = item->checkState(tree_tdc_inv);
    if(state == Qt::PartiallyChecked)
        return;
    bool checked = (state == Qt::Checked);
    auto prev_invert_map = config.invertMap;
    if(item==rootTreeItem) {
        for (const DeviceIndex &index: modTreeItem.keys()) {
            if(modTreeItem[index]->checkState(tree_tdc_inv) != state)
                modTreeItem[index]->setCheckState(tree_tdc_inv, state);
        }
    } else {
        QTreeWidgetItem *p = item->parent();
        bool found=false;
        for (const DeviceIndex &index: modTreeItem.keys()){
            if(item==modTreeItem[index]) {
                // Module item
                found = true;
                auto &chInvertMap = config.invertMap[index];
                for(int ch=0; ch<chInvertMap.size(); ++ch)
                    chInvertMap[ch] = checked;
                ui->treeWidgetChCfg->blockSignals(true);
                for(int childIndex=0; childIndex<item->childCount(); ++childIndex) {
                    item->child(childIndex)->setCheckState(tree_tdc_inv, state);
                }
                ui->treeWidgetChCfg->blockSignals(false);
            } else if(p==modTreeItem[index]) {
                // Channel item
                found = true;
                bool ok;
                int modCh = item->data(tree_tdc_name, Qt::UserRole+1).toInt(&ok);
                auto &chInvertMap = config.invertMap[index];
                Q_ASSERT(ok);
//                Q_ASSERT(modCh<=ds.getNCh());
                if(chInvertMap.size()<modCh+1)
                    chInvertMap.resize(modCh+1);
                chInvertMap[modCh] = checked;
            }
            if(found) {
                updateTreeItemStates();
                break;
            }
        }
        Q_ASSERT(found);
    }
    if(prev_invert_map != config.invertMap) {
        emit invertUpdated(config.invertMap);
        emit configChanged(config);
        prev_invert_map = config.invertMap;
    }

//    updateGuiVal();
}

void ViewerMainWindow::updateTreeWithStats(const AnalyzerStat &analStat)
{
    ui->treeWidgetChCfg->setUpdatesEnabled(false);
    refreshModChNum(analStat);

    qint64 totRaw=0,totTdc=0;
    for (const DeviceIndex &index: analStat.dev_stat.keys()){
        const auto &dev_ch_hits = analStat.dev_stat[index].dev_ch_hits;
        qint64 modRaw=0,modTdc=0;
        if(!modTreeItem.contains(index))
            continue; // Reset was invoked recently

        QTreeWidgetItem *modItem = modTreeItem[index];
        for(int ch=0; ch<modChNum.value(index, 0); ++ch) {
            QTreeWidgetItem *chItem = nullptr;
            bool ok;
            for(auto i=0; i<modItem->childCount(); ++i) {
                auto item = modItem->child(i);
                if(item->data(tree_tdc_name, Qt::UserRole+1).toInt(&ok)==ch && ok) {
                    chItem = item;
                    break;
                }
            }
            if(!chItem)
                chItem = addChTreeItem(index, modItem, ch);
            auto hits = dev_ch_hits.value(ch);
            quint32 n = hits.tdc_hits;
            chItem->setText(tree_tdc_tdc_hits,
                            QString("%L1").arg(n));
            chItem->setText(tree_tdc_msc_hits,
                            QString("%L1").arg(hits.msc_hits));
            modTdc += n;
        }
        totRaw += modRaw;
        totTdc += modTdc;
        modItem->setText(tree_tdc_raw_cnt, QString::number(modRaw));
        modItem->setText(tree_tdc_tdc_hits, QString::number(modTdc));
    }
    rootTreeItem->setText(tree_tdc_raw_cnt, QString::number(totRaw));
    rootTreeItem->setText(tree_tdc_tdc_hits, QString::number(totTdc));
    ui->treeWidgetChCfg->setUpdatesEnabled(true);
}

void ViewerMainWindow::updateTreeItemStates()
{
    ui->treeWidgetChCfg->blockSignals(true);
    Qt::CheckState allSt = rootTreeItem->checkState(tree_tdc_inv);
    Qt::CheckState st = Qt::PartiallyChecked;
    bool firstMod=true;
    for (QTreeWidgetItem *i: modTreeItem.values()) {
        for(int chIndex=0; chIndex<i->childCount(); ++chIndex){
            Qt::CheckState childState = i->child(chIndex)->checkState(tree_tdc_inv);
            if(chIndex==0){
                st = childState;
                continue;
            }
            if(st!=childState){
                st = Qt::PartiallyChecked;
                break;
            }
        }
        i->setCheckState(tree_tdc_inv, st);
        if(firstMod) {
            allSt = st;
            firstMod = false;
        } else if(allSt != st)
            allSt = Qt::PartiallyChecked;
    }

    rootTreeItem->setCheckState(tree_tdc_inv, allSt);
    ui->treeWidgetChCfg->blockSignals(false);
}

void ViewerMainWindow::updateInlMarks()
{
    ui->treeWidgetChCfg->blockSignals(true);
    int allSt = 0;
    bool firstMod=true;
    const auto &calMap = calConfig.allCal;
    for (DeviceIndex index: modTreeItem.keys()) {
        int modSt = 0;
        QTreeWidgetItem * const modItem = modTreeItem[index];
        int nch = modItem->childCount();
        for(int ch=0; ch<nch; ++ch) {
            bool chHasInl = false;
            for(auto it=calMap.cbegin(); it!=calMap.cend(); ++it) {
                if(it.key() != index && it.key().getSerial() != index.getSerial())
                    continue;
                if(chHasInl=it->hasValidChInlCal(ch))
                    break;
            }
            auto chItem = modItem->child(ch);
            int chSt = chHasInl ? 1 : -1;
            updateItemInl(chItem, chSt);

            if(ch==0){
                modSt = chSt;
            }
            if(modSt!=chSt){
                modSt = 0;
            }

        }
        updateItemInl(modItem, modSt);
        if(firstMod) {
            allSt = modSt;
            firstMod = false;
        } else if(allSt != modSt)
            allSt = 0;
    }

    updateItemInl(rootTreeItem, allSt);
    ui->treeWidgetChCfg->blockSignals(false);
}

void ViewerMainWindow::updateTtvxsMarks()
{
    ui->treeWidgetChCfg->blockSignals(true);
    int allSt = 0;
    bool firstMod=true;
    for (DeviceIndex index: modTreeItem.keys()) {
        int modSt = 0;
        QTreeWidgetItem * const modItem = modTreeItem[index];
        int nch = modItem->childCount();
        auto devHasTtvxs = !config.tdc_ttvxs_map.value(index).isNull();
        for(int chIndex=0; chIndex<nch; ++chIndex) {
            auto chItem = modItem->child(chIndex);
            bool chHasTtvxs = devHasTtvxs;
            int chSt = chHasTtvxs ? 1 : -1;
            updateItemTtvxs(chItem, chSt);
            if(chIndex==0) {
                modSt = chSt;
            }
            if(modSt!=chSt){
                modSt = 0;
            }
        }
        updateItemTtvxs(modItem, modSt);
        if(firstMod) {
            allSt = modSt;
            firstMod = false;
        } else if(allSt != modSt)
            allSt = 0;
    }

    updateItemTtvxs(rootTreeItem, allSt);
    ui->treeWidgetChCfg->blockSignals(false);
}

void ViewerMainWindow::updateSourceLabel()
{
    QString msg, colorStr;
    switch (config.source) {
    case AfiViewerConfig::SOURCE_DATA::SOURCE_IDLE:
        colorStr = "gray";
        msg = "Idle";
        break;
    case AfiViewerConfig::SOURCE_DATA::SOURCE_FILE:
        colorStr = "black";
        msg = QString("file:%1").arg(config.file_reader_config.lastFile);
        break;
    case AfiViewerConfig::SOURCE_DATA::SOURCE_STREAM:
        colorStr = tcpClientStatus.connected ? "darkGreen" : "red";
        msg = QString("stream:%1").arg(selectedProgram.toString());
        msg.append(selectedInterface.isEmpty() ? " (no interface)":
            QString(" (%1 #%2)").arg(selectedInterface.type).arg(selectedInterface.id));
        break;
    }
    dataSourceLabel->setText(msg);
    setSourceColor(colorStr);
    ui->actionDisconnect_from_program->setEnabled(!selectedProgram.isEmpty());
    dataRateLabel->setVisible(config.source == AfiViewerConfig::SOURCE_DATA::SOURCE_STREAM);
}

void ViewerMainWindow::changeSourceState(AfiViewerConfig::SOURCE_DATA st)
{
    if(st == config.source)
        return;

    auto prevSource = config.source;
    config.source = st;

    onSourceStateChange(prevSource, config.source);
    emit configChanged(config);
}

void ViewerMainWindow::onSourceStateChange(AfiViewerConfig::SOURCE_DATA prev,
                                         AfiViewerConfig::SOURCE_DATA next)
{
    switch (prev) {
    case AfiViewerConfig::SOURCE_DATA::SOURCE_FILE:
        fileReaderDialog->hide();
        emit closeFile();
        break;
    case AfiViewerConfig::SOURCE_DATA::SOURCE_STREAM:
        disconnectFromProg();
        break;
    default:
        break;
    }

    emit set_can_skip(next == AfiViewerConfig::SOURCE_DATA::SOURCE_STREAM);

    updateSourceLabel();
}

QVector<ProgramDescription> ViewerMainWindow::getFreeProgs() const
{
    QVector<ProgramDescription> freeProgs;
    for(auto it=pnpCache.cbegin(); it!=pnpCache.cend(); ++it) {
        if(isFreeProg(*it))
            freeProgs.push_back(*it);
    }
    return freeProgs;
}

bool ViewerMainWindow::hasFreeProgs() const
{
    for(auto it=pnpCache.cbegin(); it!=pnpCache.cend(); ++it) {
        if(isFreeProg(*it))
            return true;
    }
    return false;
}

void ViewerMainWindow::updateTcpInterface(const ProgramDescription &descr)
{
    if(descr.getClientIndex() != selectedProgram)
        return;

    auto list = getFreeProgramInterfaceList(descr);
    for(auto i : list) {
        if(!selectedInterface.isEmpty()) {
            if(i.type != selectedInterface.type || i.id != selectedInterface.id)
                continue;
            if(tcpClientStatus.connected)
                return;
        }
        if(i.host != selectedInterface.host
            || i.port != selectedInterface.port) {
            qDebug().noquote()
                << QString("[MW] change interface from %1:%2 to %3:%4")
                       .arg(selectedInterface.host.toString()).arg(selectedInterface.port)
                       .arg(i.host.toString()).arg(i.port);
        }

        selectedInterface = i;
        emit tcpAddressChanged(i.host, i.port);
        updateSourceLabel();
        break;
    }
}

void ViewerMainWindow::updateRawDataControl()
{
    ui->pushButtonRequestRaw->setEnabled(!config.rawDataEveryEv && rawDataReqDone);
    ui->spinBoxRawEvents->setEnabled(!config.rawDataEveryEv);
    ui->checkBoxRawPeriodical->setEnabled(!config.rawDataEveryEv);
    ui->spinBoxRawTimer->setEnabled(!config.rawDataEveryEv && config.rawDataPeriodic);
}

void ViewerMainWindow::disconnectFromProg()
{
    qDebug().noquote()
        << QString("[MW] disconnect from %1 (%2:%3)")
           .arg(selectedProgram.toString())
          .arg(selectedInterface.host.toString()).arg(selectedInterface.port);

    selectedProgram = {};
    selectedInterface = {};
    config.source_program = {};
    config.sourceIntType.clear();
    config.sourceIntId = 0;
    emit configChanged(config);

    emit tcpAddressChanged(QHostAddress(), 0);
    emit resetAll();
}

void ViewerMainWindow::setSourceColor(QString col)
{
    setLabelColor(dataSourceLabel, col);
    setLabelColor(dataRateLabel, col);
}

void ViewerMainWindow::reloadCalCalibration()
{
    calConfig.reloadCalibration(config.useSeparateInl ? config.separateInlFolder : QString());
}

void ViewerMainWindow::updateGuiToConfig()
{
    setWindowTitle(afiViewerConfigDb->getWindowTitle());

    updateRawDataControl();
    ui->actionSet_INL_separate_location->blockSignals(true);
    ui->actionSet_INL_separate_location->setChecked(config.useSeparateInl);
    ui->actionSet_INL_separate_location->blockSignals(false);
    (config.rawDataEveryEv ? ui->radioButtonRawEvery : ui->radioButtonRawRequest)
            ->setChecked(true);
    ui->spinBoxRawEvents->setValue(config.rawDataLimit);
    ui->checkBoxRawPeriodical->setChecked(config.rawDataPeriodic);
    ui->spinBoxRawTimer->setValue(config.rawDataPeriod);
    rawDataPeriodicTimer->setInterval(config.rawDataPeriod*1000);
    if(config.rawDataPeriodic)
        rawDataPeriodicTimer->start();
    else
        rawDataPeriodicTimer->stop();

    fileReaderDialog->setPackSizeLimit(PACK_SIZE_LIMIT);

    selectedProgram = config.source_program;
    selectedInterface.type = config.sourceIntType;
    selectedInterface.id = config.sourceIntId;

    emit invertUpdated(config.invertMap);
    emit tdcTtvxMapUpdated(config.tdc_ttvxs_map);
    emit inlConfigUpdated(calConfig.allCal);
    emit set_can_skip(config.source==AfiViewerConfig::SOURCE_DATA::SOURCE_STREAM);
    updateInlMarks();
    updateTtvxsMarks();
}

void ViewerMainWindow::on_pushButtonSaveToFile_clicked()
{
    QString dir = QFileDialog::getExistingDirectory(
        this, "Choose storage directory...",
        storageDir, QFileDialog::ShowDirsOnly
    );

    if(!dir.isEmpty()){
        QFile file(storageDir + "/afi-viewer-data.txt");
        file.open(QIODevice::WriteOnly | QIODevice::Text);
        QTextStream out(&file);
        out << plaint_text.join('\n');
        file.close();
        if(storageDir != dir){
            storageDir = dir;
            QSettings settings;
            settings.beginGroup("ViewerMainWindow");
            settings.setValue("storageDir", storageDir);
            settings.endGroup();
        }
    }
}

