#include "CalConfig.h"

#include <QCoreApplication>
#include <QDateTime>
#include <QDebug>
#include <QSettings>

#include "CalResult.h"
#include "base/DeviceId.h"
#include "base/DeviceIndex.h"

namespace {
const auto BASE_CAL_PREFIX = "base_cal";

inline QString getSectionStr(const DeviceIndex &devIndex, const QString &sectionType)
{
    return QString("%1-%2-%3")
        .arg(getDeviceTypeName(devIndex.getDevId()))
        .arg(devIndex.getSerial(), 8, 16, QChar('0'))
        .arg(sectionType);
}

inline QString getKeyName(CalConfig::Type type, int ch)
{
    static const QMap<CalConfig::Type, QString> map {
        {CalConfig::Type::Baseline, "x1"},
        {CalConfig::Type::Rms, "rms_x1"},
        {CalConfig::Type::Baseline4, "x4"},
        {CalConfig::Type::Rms4, "rms_x4"},
        {CalConfig::Type::Dac, "thrOffset"},
    };
    return QString("%1_%2").arg(map[type]).arg(ch);
}

template<typename T>
bool readSetting(const QSettings &s, const QString &key, T *result)
{
    if (!s.contains(key))
        return false;
    auto var = s.value(key);
    if (var.isNull() || !var.canConvert<T>())
        return false;
    if (result)
        *result = var.value<T>();
    return true;
}

QString getCalSectionStr(const DeviceIndex &devIndex)
{
    return getSectionStr(devIndex, BASE_CAL_PREFIX);
}

DeviceIndex getDevIndexByGroup(QString groupStr, QString &type)
{
    DeviceIndex index;
    if (groupStr.endsWith(BASE_CAL_PREFIX)) {
        type = BASE_CAL_PREFIX;
    } else {
        return index;
    }
    QString devIndexStr = groupStr.left(groupStr.length() - type.length() - 1); // "TQDC16VS_E-046F0C75"
    QString devType = devIndexStr.left(devIndexStr.length() - 9);               // "-046F0C75"
    QString devSerial = devIndexStr.right(8);                                   // "046F0C75"
    bool ok;
    quint32 serial = devSerial.toUInt(&ok, 16);
    if (ok) {
        quint8 id = 1;
        for (; id != 0; ++id) {
            if (devType == getDeviceTypeName(id)) {
                break;
            }
        }
        if (id)
            return DeviceIndex(static_cast<DeviceId>(id), serial);
    }
    return DeviceIndex();
}

} // namespace

std::optional<double> CalConfig::getValue(Type type, const DeviceIndex &index, int channel) const
{
    auto foundDev = find(index);
    if (foundDev == end())
        return {};
    switch (type) {
    case Type::Baseline: {
        auto foundCh = foundDev->baseline_x1.find(channel);
        if (foundCh == foundDev->baseline_x1.end())
            return {};
        return foundCh->mean;
    }
    case Type::Rms: {
        auto foundCh = foundDev->baseline_x1.find(channel);
        if (foundCh == foundDev->baseline_x1.end())
            return {};
        return foundCh->stddev;
    }
    case Type::Baseline4: {
        auto foundCh = foundDev->baseline_x4.find(channel);
        if (foundCh == foundDev->baseline_x4.end())
            return {};
        return foundCh->mean;
    }
    case Type::Rms4: {
        auto foundCh = foundDev->baseline_x4.find(channel);
        if (foundCh == foundDev->baseline_x4.end())
            return {};
        return foundCh->stddev;
    }
    case Type::Dac: {
        auto foundCh = foundDev->dacCal_mV.find(channel);
        if (foundCh == foundDev->dacCal_mV.end())
            return {};
        return *foundCh;
    }
    }
    return {};
}

void CalConfig::reloadCalibration()
{
    {
        QSettings s(QSettings::IniFormat,
                    QSettings::UserScope,
                    QCoreApplication::organizationName(),
                    "tqdc2_cal");
        reloadCalibration(s);
    }
    {
        QSettings s;
        reloadCalibration(s);
    }
}

void CalConfig::writeAll()
{
    for (auto it = begin(); it != end(); ++it)
        updateBaseCal(it.value());
}

void CalConfig::updateBaseCal(const CalResult &calRes)
{
    const auto &index = calRes.index;
    operator[](index) = calRes;

    const auto baseline = calRes.baseline_x1;
    const auto baseline4 = calRes.baseline_x4;
    const auto thr = calRes.dacCal_mV;

    QSettings s;
    s.remove(getCalSectionStr(index));
    s.beginGroup(getCalSectionStr(index));
    s.sync();
    s.beginWriteArray("temp");
    int i = 0;
    for (auto it = calRes.temps.begin(); it != calRes.temps.end(); ++it) {
        s.setArrayIndex(i++); // s.setArrayIndex(it.key());
        s.setValue("v", *it);
    }
    s.endArray();
    s.setValue("data", QDateTime::currentDateTime().toString());
    int maxNCh = -1;
    for (auto it = calRes.baseline_x1.begin(); it != calRes.baseline_x1.end(); ++it) {
        s.setValue(getKeyName(CalConfig::Type::Baseline, it.key()), it.value().mean);
        s.setValue(getKeyName(CalConfig::Type::Rms, it.key()), it.value().stddev);
        maxNCh = qMax(maxNCh, it.key());
    }
    for (auto it = calRes.baseline_x4.begin(); it != calRes.baseline_x4.end(); ++it) {
        s.setValue(getKeyName(CalConfig::Type::Baseline4, it.key()), it.value().mean);
        s.setValue(getKeyName(CalConfig::Type::Rms4, it.key()), it.value().stddev);
        maxNCh = qMax(maxNCh, it.key());
    }
    for (auto it = calRes.dacCal_mV.begin(); it != calRes.dacCal_mV.end(); ++it) {
        s.setValue(getKeyName(CalConfig::Type::Dac, it.key()), it.value());
        maxNCh = qMax(maxNCh, it.key());
    }
    s.setValue("nch", calRes.nch);
    s.endGroup();
}

void CalConfig::reloadCalibration(QSettings &s)
{
    const auto &groups = s.childGroups();
    for (const QString &group : groups) {
        QString groupType;
        const DeviceIndex &devIndex = getDevIndexByGroup(group, groupType);
        if (devIndex.isNull())
            continue;

        auto &devCal = operator[](devIndex);
        devCal.index = devIndex;
        s.beginGroup(group);

        if (group.endsWith(BASE_CAL_PREFIX)) {
            bool ok;
            devCal.nch = s.value("nch", -1).toInt(&ok);
            if (!ok)
                continue;
            for (int ch = 0; ch < devCal.nch; ++ch) {
                readSetting(s, getKeyName(CalConfig::Type::Baseline, ch), &devCal.baseline_x1[ch].mean);
                readSetting(s, getKeyName(CalConfig::Type::Rms, ch), &devCal.baseline_x1[ch].stddev);
                readSetting(s, getKeyName(CalConfig::Type::Baseline4, ch), &devCal.baseline_x4[ch].mean);
                readSetting(s, getKeyName(CalConfig::Type::Rms4, ch), &devCal.baseline_x4[ch].stddev);
                readSetting(s, getKeyName(CalConfig::Type::Dac, ch), &devCal.dacCal_mV[ch]);
            }
        }

        s.endGroup();
    }
}
