/*
**    Copyright 2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "DeviceIdentityProxyModel.h"

#ifdef QT_WIDGETS_LIB
#include <QApplication>
#include <QPixmap>
#include <QStyle>
#endif
#include <QVariant>

#include "BaseDeviceTableModel.h"
#include "base/DnsLookup.h"
#include "IpPortPair.h"

DeviceIdentityProxyModel::DeviceIdentityProxyModel(QObject *parent)
    : QIdentityProxyModel(parent),
      dnsLookup(new DnsLookup(this))
{
    connect(dnsLookup, &DnsLookup::reverseMapUpdated,
            this, &DeviceIdentityProxyModel::reverseMapUpdated);
}

Qt::ItemFlags DeviceIdentityProxyModel::flags(const QModelIndex & index) const
{
    auto f = QIdentityProxyModel::flags(index);
    switch (index.column()) {
    case BaseDeviceTableModel::COL_ENABLE:
        return  f | Qt::ItemIsUserCheckable;
    default:
        return f;
    }
}

QVariant DeviceIdentityProxyModel::headerData(int section, Qt::Orientation orientation, int role) const
{
    if (orientation == Qt::Vertical && role == Qt::DisplayRole) {
        return section;
    }
    return QIdentityProxyModel::headerData(section, orientation, role);

}

QVariant DeviceIdentityProxyModel::data(const QModelIndex &index, int role) const
{
    switch (role) {
    case Qt::DisplayRole:
        return getDisplayData(index);
    case Qt::ToolTipRole:
        return getToolTipData(index);
    case Qt::DecorationRole:
        return getDecorationData(index);
    default:
        return QIdentityProxyModel::data(index, role);
    }
}

QVariant DeviceIdentityProxyModel::getDisplayData(const QModelIndex &index) const
{
    const QVariant &data = QIdentityProxyModel::data(index, Qt::DisplayRole);
    if (data.isNull())
        return data;

    switch(index.column()) {
    case BaseDeviceTableModel::COL_ENABLE:
    case BaseDeviceTableModel::COL_ONLINE:
        return QVariant();
    case BaseDeviceTableModel::COL_DYNAMIC: {
        auto flag = data.value<int>();
        if (flag == DeviceTableEntry::Static)
            return "static";
        if (flag == DeviceTableEntry::Dynamic)
            return "dynamic";
        return QVariant();
    }
    case BaseDeviceTableModel::COL_SLOT:
        return QString::number(data.value<int>());
    case BaseDeviceTableModel::COL_IP:
    case BaseDeviceTableModel::COL_MASTER:
    case BaseDeviceTableModel::COL_M_STREAM:
    {
        const auto &p = IpPortPair::parse(data.value<QString>());
        if (!resolved_hosts.contains(p.ip)) {
            dnsLookup->reverseLookup(QHostAddress(p.ip).toString());
            return p.toString();
        }
        return QString("%1:%2").arg(resolved_hosts.value(p.ip)).arg(p.port);
    }
    default:;
    }
    return data;
}

QVariant DeviceIdentityProxyModel::getToolTipData(const QModelIndex &index) const
{
    const QVariant &data = QIdentityProxyModel::data(index, Qt::DisplayRole); // not a ToolTipRole
    if (data.isNull())
        return data;

    switch(index.column()) {
    case BaseDeviceTableModel::COL_ONLINE:
        return data.toBool() ? "Device sends discover packets"
                             : "Device doesn't send discover packets";
//    case BaseDeviceTableModel::COL_DEV_TYPE_TXT:
//        return "0x" + QString("%1").arg(data.value<int>(), 2, 16, QChar('0')).toUpper();
    case BaseDeviceTableModel::COL_IP:
    case BaseDeviceTableModel::COL_MASTER:
    case BaseDeviceTableModel::COL_M_STREAM:
    {
        const auto &p = IpPortPair::parse(data.value<QString>());
        if (!resolved_hosts.contains(p.ip)) {
            dnsLookup->reverseLookup(QHostAddress(p.ip).toString());
            return QVariant();
        }
        return p.toString();
    }
    default:;
    }
    return QVariant();
}

QVariant DeviceIdentityProxyModel::getDecorationData(const QModelIndex &index) const
{
#ifdef QT_WIDGETS_LIB
    if(index.column() != BaseDeviceTableModel::COL_ONLINE)
        return QVariant();

    const QVariant &data = QIdentityProxyModel::data(index, Qt::UserRole);
    const auto &sp = data.toBool() ? QStyle::SP_DialogApplyButton
                                   : QStyle::SP_DialogCloseButton;
    const auto &p = QApplication::style()->standardPixmap(sp);
    return QIcon(p);
#else
    return QVariant();
#endif
}

QMap<quint32, QString> DeviceIdentityProxyModel::getHostMap() const
{
    return dnsLookup->getHostMap();
}

void DeviceIdentityProxyModel::reverseMapUpdated(const DnsReverseMap &map)
{
    if (map == resolved_hosts)
        return;
    beginResetModel();
    resolved_hosts = map;
    endResetModel();
}

//QString DeviceIdentityProxyModel::getIpStr(bool locked, quint32 ip, quint16 port) const
//{
//    QString masterStr;
//    if (locked) {
//        bool ipResolved = getHostMap().contains(ip);
//        auto ipStr = ipResolved ? getHostMap()[ip] : QHostAddress(ip).toString();
//        masterStr = QString("%1:%2").arg(ipStr).arg(port);
//    } else {
//        masterStr = "free";
//    }
//    return masterStr;
//}
