#include "AdcSpiAd9249Module.h"

#include <QDebug>
#include <QTime>
#include <QThread>

#include "base/DeviceId.h"
#include "regio/regop.h"
#include "sdb/AfiSdbId.h"

namespace  {
enum {
    REG_ID                  = 0x00,
    REG_ADC_CH_INFO         = 0x02,
    REG_ADC_BITS            = 0x03,

    REG_SPI_ADDR            = 0x8,
    REG_SPI_WR_DATA         = 0x9,
    REG_SPI_READ_DATA       = 0xA,
    REG_SPI_ACTIVE_SELECT   = 0xB
};

enum {
    REG_AD9249_POWER_MODES = 0x08,
    REG_AD9249_TEST_MODE = 0x0D,
    REG_AD9249_OUTPUT_MODE = 0x14,
    REG_AD9249_USER_PATT1_LSB = 0x19,
    REG_AD9249_USER_PATT1_MSB = 0x1A,
    REG_AD9249_USER_PATT2_LSB = 0x1B,
    REG_AD9249_USER_PATT2_MSB = 0x1C,
    REG_AD9249_SERIAL_OUTPUT_DATA_CTRL = 0x21
};
}

AdcSpiAd9249Module::AdcSpiAd9249Module(const MregBusInterface &bus)
    : BaseDeviceModule(bus)
{
    if (!isAvailable()) {
        if (getDeviceId() == DEVICE_ID_PHOS_FEC)
            setBaseAddr(0x4100, 0xFF);
    }
}

QList<AFI_SDB_ID> AdcSpiAd9249Module::getSupportedSdbIds()
{
    return {AFI_SDB_AD9249_SPI};
}

void AdcSpiAd9249Module::post_init()
{
    quint16 chInfo, bits;
    RegOpVector r;
    r.RegRead16(REG_ADC_CH_INFO, &chInfo);
    r.RegRead16(REG_ADC_BITS, &bits);
    if(!regOpExecRebased(r))
        return;
    adcChPerChip = chInfo >> 8;
    adcChips = chInfo & 0xFF;
    adcDataBits = bits;

    setupAdc();
}

bool AdcSpiAd9249Module::spi_test_impl(TestResult *result)
{
    TestResult testResult;

    std::vector<quint8> testData(adcChips);
    for (int i = 0; i < adcChips; i++) {
        testData[i] = std::rand() & 0xFF;
        testResult.map[i] = true;
    }

    for (int adcId = 0; adcId < adcChips; ++adcId) {
        quint16 active_adc_mask = 1 << adcId;

        auto rand_val = testData[adcId];
        if (!spi_write(REG_AD9249_USER_PATT1_LSB, rand_val, active_adc_mask))
            return {};
        quint16 readregval;
        if (!spi_read(REG_AD9249_USER_PATT1_LSB, &readregval, active_adc_mask))
            return {};
        if (rand_val != readregval) {
            testResult.map[adcId] = false;
        }
    }
    if (result)
        *result = testResult;
    return true;
}

bool AdcSpiAd9249Module::setupAdc()
{
    bool ok = true;
    qDebug() << "AD9249 setup";

    const quint16 &adc_mask=getAdcMask();

    ok &= spi_write(REG_AD9249_POWER_MODES, 0x3, adc_mask); //digital reset
    ok &= spi_write(REG_AD9249_POWER_MODES, 0x0, adc_mask); //digital reset

    switch(adcDataBits) {
    case 14:
        ok &= spi_write(REG_AD9249_SERIAL_OUTPUT_DATA_CTRL, 0xC1, adc_mask); //MSB serial first, 14bit
        break;
    default:
        ok &= spi_write(REG_AD9249_SERIAL_OUTPUT_DATA_CTRL, 0xC2, adc_mask); //MSB serial first, 12bit
    }
    if (true)
        ok &= spi_write(REG_AD9249_OUTPUT_MODE, 0x1, adc_mask); // two's complement (signed int)
    else
        ok &= spi_write(REG_AD9249_OUTPUT_MODE, 0x0, adc_mask); // offset binary (unsigned int)

    ok &= spi_test();

    //setCustomPatternMode(true);
    //setCustomPattern(0x40);
    return ok;
}

bool AdcSpiAd9249Module::setCustomPatternMode(bool mode)
{
    return spi_write(REG_AD9249_TEST_MODE, mode ? 0x08 : 0, 0xFF);
}

bool AdcSpiAd9249Module::setCustomPattern(quint16 data)
{
    data <<= 16-adcDataBits;
    return
        spi_write(REG_AD9249_USER_PATT1_LSB, data&0xFF, 0xFF) &&
        spi_write(REG_AD9249_USER_PATT1_MSB, (data>>8)&0xFF, 0xFF) &&
        spi_write(REG_AD9249_USER_PATT2_LSB, data&0xFF, 0xFF) &&
        spi_write(REG_AD9249_USER_PATT2_MSB, (data>>8)&0xFF, 0xFF);
}

bool AdcSpiAd9249Module::spi_write(quint16 addr, quint8 data, quint16 device_mask)
{
    RegOpVector r;
    r.RegWrite16(REG_SPI_ACTIVE_SELECT, device_mask);
    r.RegWrite16(REG_SPI_WR_DATA, data);
    r.RegWrite16(REG_SPI_ADDR, addr & 0x1FFF);
    return regOpExecRebased(r);
}

bool AdcSpiAd9249Module::spi_read(quint16 addr, quint16 *result, quint16 device_mask)
{
    bool ok = true;
    quint16 readval;

    RegOpVector r;
    r.RegWrite16(REG_SPI_ACTIVE_SELECT, device_mask);
    r.RegWrite16(REG_SPI_ADDR, ((addr&0x9FFF) | 0x8000));   //bits 13,14 = 0 , bit15 = 1 (read and W0=W1=0 - 1 byte operation)
    ok &= regOpExecRebased(r);

    ok &= regRead16(REG_SPI_READ_DATA, &readval);

    if (ok && result)
        *result = readval;
    return ok;
}
