/*
**    Copyright 2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "ModularDeviceStatus.h"

#include <QStringList>

static const char *hilight_error = "<span style=\"color:red;\">";
static const char *hilight_warn = "<span style=\"color:yellow;\">";
static const char *hilight_close = "</span>";

QString getStateTimeHr(quint32 time)
{
    const static quint32 DAY = 24*3600*1000;
    int days = time / DAY;
    return (days ? QString("%1d ").arg(days) : "") +
            QTime::fromMSecsSinceStartOfDay(time%DAY).toString("hh:mm:ss.zzz");
}

double TempsInfo::minTemp() const
{
    if (t.empty()) return {};
    const auto &tempList = t.values();
    return *std::min_element(tempList.begin(), tempList.end());
}

double TempsInfo::maxTemp() const
{
    if (t.empty()) return {};
    const auto &tempList = t.values();
    return *std::max_element(tempList.begin(), tempList.end());
}

QString TempsInfo::brief() const
{
    if (t.empty()) return {};
    QString s = QString("%1").arg(minTemp(), 0, 'f', 1);
    if (t.size() > 1)
        s += " - " + QString("%1").arg(maxTemp(), 0, 'f', 1);
    return s;
}

QString TempsInfo::full() const
{
    if (t.empty()) return {};
    QStringList fullInfo;
    for (auto it = t.begin(); it != t.end(); ++it) {
        fullInfo << QString("%1 - %2%3").arg(it.key()).arg(it.value(), 0, 'f', 1).arg(QChar(0x02da));
    }
    return fullInfo.join('\n');
}

QString PowerInfo::full() const
{
    if (t.empty()) return {};
    QStringList fullInfo;
    for (auto it = t.begin(); it != t.end(); ++it) {
        fullInfo << QString("%1 - %2").arg(it.key()).arg(it.value(), 0, 'f', 1);
    }
    return fullInfo.join('\n');
}

QString ModularDeviceStatus::formatHtmlErrorNumber(int number)
{
    auto s = QString::number(number);
    if (number > 0)
        s = hilight_error + s + hilight_close;
    return s;
}

QString ModularDeviceStatus::getDeviceName() const
{
    const auto &sdb = getSdb();
    if (!sdb.isValid())
        return deviceIndex.getDeviceIdName();
    QString name = sdb.ic().productName();
    if (name.isEmpty())
        return deviceIndex.getDeviceIdName();
    //    auto ver = sdb.productVersion();
    //    if (ver.isEmpty())
    return name;
    //    return name + "-v" + ver;
}

QString ModularDeviceStatus::getDeviceIdent() const
{
    return getDeviceName() + " " + getSerialStr();
}

bool ModularDeviceStatus::isSdbValid() const
{
    const auto &sdb = getSdb();
    bool sdbOk = sdb.isValid();
    if (sdb.crcPresent())
        sdbOk &= sdb.crcOk();
    return  sdbOk;
}

fw_version_t ModularDeviceStatus::getFwVer() const
{
    if (mlinkCsrStatus) {
        return mlinkCsrStatus->fw_version;
    } else {
        return {};
    }
}

QString ModularDeviceStatus::getFwStr() const
{
    const auto &sdb = getSdb();
    bool sdbOk = isSdbValid();
    const auto &commitId = sdb.syn().commit_id;
    if (sdbOk && !commitId.isEmpty()) {
        return commitId;
    }
    auto fw_ver = getFwVer();
    if (!fw_ver.isNull()) {
        return fw_ver.toString();
    } else {
        return {};
    }
}

QString ModularDeviceStatus::getFwText() const
{
    const auto &sdb = getSdb();
    bool sdbOk = isSdbValid();
    const auto &ic = sdb.ic();
    const auto &syn = sdb.syn();
    QStringList sl;
    if (sdbOk) {
        sl << "<b>" + ic.productName() + " " + ic.productVersion() + "</b>";
        sl << "Commit " + syn.commit_id + " on " + syn.date;
        sl << "Build by " + syn.user_name + " on " + ic.productDate();
    } else {
        sl << "<b>" + deviceIndex.getDeviceIdName() + " " + getFwStr() + "</b>";
    }
    return sl.join("\n").replace("\n", "<br>");
}

QString ModularDeviceStatus::getHwStr() const
{
    if (mlinkCsrStatus) {
        return mlinkCsrStatus->hw_version.toString();
    } else {
        return {};
    }
}

QString ModularDeviceStatus::getHwFwStr() const
{
    if (mlinkCsrStatus) {
        return getHwStr() + "/" + getFwStr();
    }
    return {};
}

QString ModularDeviceStatus::getDeviceDescrText() const
{
    QStringList info;
    if (mlinkCsrStatus) {
        QStringList sl;
        if (!mlinkCsrStatus->hw_version.isNull())
            sl << "PCB: " + mlinkCsrStatus->hw_version.toString();
        if (!mlinkCsrStatus->fw_version.isNull())
            sl << "Firmware: " + mlinkCsrStatus->fw_version.toString();
        DeviceIndex index(static_cast<DeviceId>(mlinkCsrStatus->device_id), mlinkCsrStatus->device_serial);
        sl << "Serial Number: " + index.getSerialStr();
        sl << "Device ID: " + QString("%1").arg(mlinkCsrStatus->device_id, 2, 16, QChar('0'))
                + " " + index.getDeviceIdName();
        if (! sl.isEmpty()) {
            sl.push_front("<b>CSR</b>");
            info << sl.join("<br>");
        }
    }
    const auto &sdb = getSdb();
    const SdbInterconnect &ic = sdb.ic();
    if (ic.isValid()) {
        QStringList sl;
        sl << "Product: " + ic.productName();
        sl << "Version: " + ic.productVersion();
        sl << "Date: " + ic.productDate();
        if (! sl.isEmpty()) {
            sl.push_front("<b>SDB</b>");
            info << sl.join("<br>");
        }
    }
    const auto &syn = sdb.syn();
    if (syn.isValid()) {
        QStringList sl;
        if (!syn.commit_id.isEmpty())
            sl << "Revision: " + syn.commit_id;
        if (!syn.user_name.isEmpty())
            sl << "User: " + syn.user_name;
        if (!syn.date.isEmpty())
            sl << "Date: " + syn.date;
        if (! sl.isEmpty()) {
            sl.push_front("<b>SDB build info</b>");
            info << sl.join("<br>");
        }
    }
    if (bmcStatus.has_value() && bmcStatus.value().isValid()) {
        QStringList sl;
        sl << "Version: " + getBmcVersionStr();
        if (! sl.isEmpty()) {
            sl.push_front("<b>BMC</b>");
            info << sl.join("<br>");
        }
    }
    return info.join("<br>");
}

QString ModularDeviceStatus::getBmcVersionStr() const
{
    if (!bmcStatus.has_value()) return {};
    return bmcStatus.value().getVerStr();
}

Severity ModularDeviceStatus::getBmcSeverity() const
{
    if (!bmcStatus.has_value()) return Severity::Unknown;
    auto sensorStatus = bmcStatus.value().get_overall_status();
    switch (sensorStatus) {
    case SENSOR_CRITICAL: return Severity::Critical;
    case SENSOR_WARNING: return Severity::Warning;
    case SENSOR_NORMAL: return Severity::Normal;
    case SENSOR_UNKNOWN: return Severity::Unknown;
    }
    return Severity::Unknown;
}

QString ModularDeviceStatus::getBmcText() const
{
    if (!bmcStatus.has_value()) return "[none]";
    const auto &bmc = bmcStatus.value();
    if (!bmc.isValid()) return "[invalid]";
    QStringList lines;
    lines << QString("BMC %1").arg(bmc.getVerStr());

    QStringList sensorText;
    auto subsystemStatus = bmc.getSubsystemStatus();
    for (auto it = subsystemStatus.begin(); it != subsystemStatus.end(); ++it) {
        sensorText << QString("%1 - %2").arg(it.key(), bmc.get_status_str(it.value()));
    }
    lines << sensorText.join("\n");

    lines << QString("PLL %1").arg(bmc.onboard_pll_ok() ? "locked" : "unlocked");
    auto pllUnlockCount = bmc.getPllUnlockCnt();
    if (pllUnlockCount > 0)
        lines << QString("PLL unlock count: %1").arg(pllUnlockCount);

    lines << "<b>Sensors</b>";
    lines << TempsInfo(bmc.getTemps()).full();
    lines << bmc.getPower().getSummary();

    return lines.join("<br>").replace("\n", "<br>");
}

QString ModularDeviceStatus::getSdbText() const
{
    if (!sdb.isValid())
        return "[none]";
    const auto &ic = sdb.ic();
    const auto &syn = sdb.syn();
    QStringList sl;
    sl << "<b>" + ic.productName() + " " + ic.productVersion() + "</b><br>";
    sl << "Commit " + syn.commit_id + " on " + syn.date;
    sl << "Build by " + syn.user_name + " on " + ic.productDate();
    static const auto sdb_crc_error_str = QString(hilight_error) + "SDB checksum error" + hilight_close;
    static const auto sdb_crc_ok_str = QString("SDB checksum correct");
    if (!sdb.crcPresent())
        sl << "No SDB checksum present";
    else
        sl << (sdb.crcOk() ? sdb_crc_ok_str : sdb_crc_error_str);
    QStringList dl;
    for (const SdbDevice &device : sdb.devices()) {
        std::optional<SdbModuleIndex> moduleIndex;
        for (const auto &mi : moduleIndexes) {
            bool id_match = (mi.id == device.component.product.device_id);
            bool base_match = (mi.baseAddr == (device.component.addr_first >> 1));
            if (id_match && base_match) {
                moduleIndex = mi; break;
            }
        }
        QString line =
            QString("%1-%2").arg(device.component.addr_first, 8, 16, QChar('0'))
                .arg(device.component.addr_last, 8, 16, QChar('0')).toUpper()
            + " | "
            + QString("%1-%2").arg(device.component.addr_first >> 1, 4, 16, QChar('0'))
                  .arg(device.component.addr_last >> 1, 4, 16, QChar('0')).toUpper()
            + " | "
            + QString("%1").arg(device.component.product.device_id, 8, 16, QChar('0')).toUpper()
            + " | "
            + QString("%1").arg(device.component.product.name, -19)
            + " "
            + QString("%1.%2").arg(device.abi_ver_major).arg(device.abi_ver_minor)
            + " | "
            + (moduleIndex ? "+" : " ");
        dl << line;
    }
    sl << "<pre>" + dl.join("<br>")+ "</pre>";
    return sl.join("<br>").replace("\n", "<br>");
}

QString ModularDeviceStatus::getNetworkPortsText() const
{
    QStringList sl;
    if (networkPortStatus.isEmpty())
        sl << "[none]";
    for (auto it = networkPortStatus.begin(); it != networkPortStatus.end(); ++it) {
        const auto &portStatus = it.value();
        const auto portIndex = portStatus.descr.id.value_or(it.key());
        sl << "<b>Port " + QString::number(portIndex) + "</b>";
        const auto &phy = portStatus.phy;
        const auto &pcspma = portStatus.pcspma;
        const auto &mac = portStatus.mac;
        const auto &hwip = portStatus.hwip;
        const auto &mstream = portStatus.mstream;
        sl << "<b>PHY</b>" << phy.toString();
        sl << "<b>PCS/PMA</b>" << pcspma.toString();
        sl << "<b>MAC</b>" << mac.toString();
        sl << "<b>HWIP</b>" << hwip.toString();
        sl << "<b>MStream</b>" << mstream.toString();
    }
    return sl.join("<br><br>").replace("\n", "<br>");
}

QString ModularDeviceStatus::getWrText() const
{
    QStringList sl;
    if (!wrStatus.has_value()) {
        return "[none]";
    }
    if (!wrStatus->acquiredAtMs) {
        return "[no data yet]";
    }
    bool wrTimeValid = wrStatus->timeValid();
    const auto &wrTime = QDateTime::fromMSecsSinceEpoch(wrStatus->mSecsSinceEpoch());
    int wrTimeCheck = wrStatus->wrTimeCheck();
    double wrTimeOffsetMs = wrStatus->mSecsSinceEpoch() - wrStatus->acquiredAtMs;

    QString timeStr = wrTime.toUTC().toString("yyyy-MM-dd hh:mm:ss"); // add ".zzz" for ms
    if (!wrTimeValid)
        timeStr = hilight_error + timeStr + hilight_close;

    QString strOffset;
    if (wrTimeOffsetMs/1000. > -60 && wrTimeOffsetMs/1000. < 60) {
        strOffset = QString(" %1").arg(wrTimeOffsetMs * 1e-3, 6, 'f', 3);
        if (wrTimeCheck == 1)
            strOffset = hilight_warn + strOffset + hilight_close;
        if (wrTimeCheck > 1)
            strOffset = hilight_error + strOffset + hilight_close;
    }
    if (!strOffset.isEmpty())
        strOffset = ", offset " + strOffset;
    sl << "WR Time: " + timeStr + strOffset;
    sl << "Link errors: " + formatHtmlErrorNumber(wrStatus->linkError);
    sl << "Sync fail: " + formatHtmlErrorNumber(wrStatus->syncFail);
    sl << "Code errors: " + formatHtmlErrorNumber(wrStatus->codeError);
    sl << "Uptime: " + QString::number(wrStatus->uptime);

    return sl.join("<br>").replace("\n", "<br>");
}

QString ModularDeviceStatus::getFeLinkText() const
{
    if (feLinkStatus.isEmpty())
        return "[none]";

    QStringList sl;
    for(auto it=feLinkStatus.cbegin(); it!=feLinkStatus.cend(); ++it) {
        const auto &st = *it;
        sl << "[port #"+ QString::number(it.key()) + "]";
        sl << "Remote index: " + it->remoteIndex.getIdent();

        auto statusStr = st.getInfoStr(false);
        if(!st.allOk())
            statusStr = hilight_error + statusStr + hilight_close;
        sl << "Status: " + statusStr;

        sl << "Bit slide: "+ QString::number(it->RX_Bitslide);
        sl << "Remote ver.: " + it->remoteVer.getInfoStr();

        auto timeStr = QDateTime::fromSecsSinceEpoch(st.timeCode.sec).toString();
        if(!st.timeCode.isValid())
            timeStr = hilight_error + timeStr + hilight_close;
        sl << "TAI: " + timeStr;

        sl << "Cur. link state time: " + getStateTimeHr(st.Current_link_state_time);
        sl << "RX_CODE_ERR: "+ formatHtmlErrorNumber(it->counter.RX_code_error);
        sl << "RX_CRC_ERR: "+ formatHtmlErrorNumber(it->counter.RX_CRC_error);
        sl << "TIME_FRAMES_SENT: "+ QString::number(it->counter.time_fr_sent);
        sl << "TRIG_FRAMES_SENT: "+ QString::number(it->counter.trigger_fr_sent);
        sl << "TIME_FRAMES_RECV: "+ QString::number(it->counter.time_fr_received);
        sl << "TRIG_FRAMES_RECV: "+ QString::number(it->counter.trigger_fr_received);
        sl << "UNKNOWN_FRAMES: "+ formatHtmlErrorNumber(it->counter.unknown_received);
    }

    return sl.join("<br>").replace("\n", "<br>");
}

QMap<QString, double> ModularDeviceStatus::getTemps() const
{
    QMap<QString, double> t;
    if (mlinkCsrStatus.has_value()) {
        t.insert(mlinkCsrStatus.value().getTemps());
    }
    if (bmcStatus.has_value())
        t.insert(bmcStatus.value().getTemps());
    return t;
}

std::optional<MstreamPortStatus> ModularDeviceStatus::getMstreamStatus() const
{
    for (auto it = networkPortStatus.begin(); it != networkPortStatus.end(); ++it) {
        const auto &portStatus = it.value();
        if (portStatus.mstream.isValid())
            return portStatus.mstream;
    }
    return {};
}

bool ModularDeviceStatus::isRunFinished() const
{
    bool running = false;
    if (runLogicStatus)
        running = runLogicStatus->isRunning();
    bool queue_empty = true;
    if (trigStatus)
        queue_empty = ((trigStatus->trigQueue & 0xF) == 0);
    return !running && queue_empty;
}

ModularDeviceStatus& ModularDeviceStatus::operator-=(const ModularDeviceStatus& r)
{
    if (r.hitCoreStatus.has_value()) {
        if (hitCoreStatus.has_value())
            *hitCoreStatus -= *r.hitCoreStatus;
    }
    if (r.runLogicStatus.has_value()) {
        if (runLogicStatus.has_value())
            *runLogicStatus -= *r.runLogicStatus;
    }
    if (r.runStatistic.has_value()) {
        if (runStatistic.has_value())
            *runStatistic -= *r.runStatistic;
    }
    return *this;
}
