/*
**    Copyright 2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef MODULARDEVICESTATUS_H
#define MODULARDEVICESTATUS_H

#include <optional>
#include <QMap>

#include "AdcDataTest/AdcDataTestStatus.h"
#include "BoardGpio/BoardGpioStatus.h"
#include "FirFilter/FirFilterStatus.h"
#include "HitCore/HitCoreStatus.h"
#include "MlinkCsr/MlinkCsrStatus.h"
#include "RunLogic/RunLogicStatus.h"
#include "RunStatistic/RunStatistic.h"
#include "SpillControl/SpillStatus.h"
#include "TrigCtrlStat/TrigCtrlStatStatus.h"
#include "TrigInfo/TrigInfoStatus.h"
#include "WaveBlc/WaveBlcStatus.h"
#include "WaveRec/WaveRecStatus.h"
#include "base/DeviceIndex.h"
#include "base/SdbModuleIndex.h"
#include "bmc/BmcStatus.h"
#include "device-modules/ClockControl/ClockStatus.h"
#include "device-modules/ClockDividerControl/ClockDividerStatus.h"
#include "device-modules/DeserializersControl/DeserializersControlStatus.h"
#include "device-modules/FeLinkStatus/FeLinkStatus.h"
#include "device-modules/FlashProg/FlashProgStatus.h"
#include "device-modules/NetworkPort/NetworkPortStatus.h"
#include "device-modules/TdcCore/TdcCoreStatus.h"

// #include "Tlu16SfpTrigger/Tlu16SfpTriggerBuildParams.h"
#include "Tlu40LvdsLuminosity/Tlu40LvdsLuminosityStatus.h"
#include "Tlu40LvdsTofTrigger/Tlu40LvdsTofTriggerStatus.h"
#include "TluDecDes/TluDecDesStatus.h"
// #include "TluTtlIoControl/TluTtlIoControlBuildParams.h"

#include "device-modules/WrStatus/WrStatus.h"
#include "sdb/SdbTypes.h"

struct TempsInfo
{
    explicit TempsInfo(const QMap<QString, double> &init = {}) : t(init) {}
    QMap<QString, double> t;
    double minTemp() const;
    double maxTemp() const;
    QString brief() const;
    QString full() const;
};

struct PowerInfo
{
    explicit PowerInfo(const QMap<QString, double> &init = {}) : t(init) {}
    QMap<QString, double> t;
    QString full() const;
};

class ModularDeviceStatus
{
public:
    explicit ModularDeviceStatus(const DeviceIndex &deviceIndex = {})
        : deviceIndex(deviceIndex) {}

    friend bool operator==(const ModularDeviceStatus &, const ModularDeviceStatus &) = default;
    ModularDeviceStatus& operator-=(const ModularDeviceStatus& r);
    friend ModularDeviceStatus operator-(ModularDeviceStatus l, const ModularDeviceStatus& r) { l -= r; return l; }

    bool isNull() const { return deviceIndex.isNull(); }
    DeviceIndex getDeviceIndex() const { return deviceIndex; }

    // convenience wrappers
    bool isSdbValid() const;
    QString getBmcVersionStr() const;
    QString getSerialStr() const { return deviceIndex.getSerialStr(); }
    fw_version_t getFwVer() const;
    QString getFwStr() const;
    QString getFwText() const;
    QString getHwStr() const;
    QString getHwFwStr() const;
    QString getDeviceName() const;
    QString getDeviceIdent() const;
    QString getDeviceDescrText() const;
    QString getBmcText() const;
    Severity getBmcSeverity() const;
    QString getSdbText() const;
    QString getNetworkPortsText() const;
    QString getWrText() const;
    QString getFeLinkText() const;
    static QString formatHtmlErrorNumber(int number);

    QMap<QString, double> getTemps() const;
    sdb_t getSdb() const { return sdb; }
    void setSdb(const sdb_t &r) { sdb = r; }
    std::optional<MstreamPortStatus> getMstreamStatus() const;
    bool isRunFinished() const;

    bool connected = false;

    std::optional<AdcDataTestStatus> adcDataTestStatus;
    std::optional<BmcStatus> bmcStatus;
    std::optional<BoardGpioStatus> boardGpio;
    std::optional<ClockStatus> clockStatus;
    std::optional<ClockDividerStatus> clockDividerStatus;
    std::optional<DeserializersControlStatus> desStatus;
    QMap<int, FeLinkStatus> feLinkStatus;
    std::optional<FlashProgStatus> flashProgStatus;
    std::optional<FirFilterStatus> firFilterStatus;
    std::optional<HitCoreStatus> hitCoreStatus;
    std::optional<MlinkCsrStatus> mlinkCsrStatus;
    QMap<int, NetworkPortStatus> networkPortStatus;
    std::optional<RunLogicStatus> runLogicStatus;
    std::optional<RunStatistic> runStatistic;
    std::optional<SpillStatus> spillStatus;
    std::optional<TdcCoreStatus> tdcCoreStatus;

    std::optional<TluDecDesStatus> tluDecDesStatus;
    std::optional<Tlu40LvdsLuminosityStatus> tlu40LvdsLuminosityStatus;
    std::optional<Tlu40LvdsTofTriggerStatus> tlu40LvdsTofTriggerStatus;

    std::optional<TrigInfoStatus> trigInfo;
    std::optional<TrigCtrlStatStatus> trigStatus;
    std::optional<WaveBlcStatus> waveBlc;
    std::optional<WaveRecStatus> waveRec;
    std::optional<WrStatus> wrStatus;

    sdb_t sdb;
    QList<SdbModuleIndex> moduleIndexes;

private:
    DeviceIndex deviceIndex;
};

Q_DECLARE_METATYPE(ModularDeviceStatus)

using BaseStatusMap = QMap<DeviceIndex, ModularDeviceStatus>;

template<class DerivedStatus>
inline BaseStatusMap statusMapToBase(const QMap<DeviceIndex, DerivedStatus> &m) {
    BaseStatusMap b;
    for (auto it = m.begin(); it != m.end(); ++it) {
        b.insert(it.key(), it.value());
    }
    return b;
}

#endif // MODULARDEVICESTATUS_H
