#include "MscHistModule.h"
#include "MscHistConfig.h"

#include "regio/RegOpStream.h"
#include "regio/regop.h"
#include "sdb/AfiSdbId.h"

namespace {
enum {
    //Histograms CSR
    REG_CTRL = 0x0,
    REG_NUMBER_OF_RAM_BITS_HIST_TIME = 0x1,
    REG_NUMBER_OF_RAM_BITS_HIST_PLOT = 0x2,
    REG_BASE_HIST_PLOT_START_ADDR = 0x100,
    REG_HIST_PLOT = 0x10,
    REG_HIST_PLOT_SLICE_TIME = 0x00, // 32-bit RW
    REG_HIST_PLOT_CUR_RAM_ADDR = 0x2,               // 32-bit RO
};

// Ver 2.0
enum {
    REG32_ID = 0x0,
    REG_V2_CTRL = 0x2,
    REG_V2_LOCK = 0x3,
    REG_V2_CLOCK_FREQ = 0x04,
    REG_V2_NUMBER_OF_CH = 0x06, // num of hist.
    // LH - Logarithmic histogram
    REG_V2_LH_RAM_BASE_ADDR = 0x08,
    REG_V2_LH_RAM_NUMBER_OF_RAM_BITS = 0x0A,
    REG_V2_LH_HIST_NUM = 0xB, // RO
    REG_V2_LH_REG_BASE = 0xC, // RO
    // PH - Plot histogram
    REG_V2_PH_RAM_BASE_ADDR = 0x10,
    REG_V2_PH_RAM_NUMBER_OF_RAM_BITS = 0x12,
    REG_V2_PH_HIST_NUM = 0x13, // RO
    REG_V2_PH_REG_BASE = 0x14, // RO
    REG_V2_PH_REG_MASK = 0x15, // RO
};

// Logarithmic histogram registers
enum {
    REG_V2_LH_TYPE_MASK = 0x04, // 32-bit RO
};

// Plot histogram registers array
enum {
    REG_V2_PH_SLICE_TIME = 0x00, // 32-bit RW
    REG_V2_PH_CUR_RAM_ADDR = 0x02, // 32-bit RO
    REG_V2_PH_LAST_ADDR_PREV_GATE = 0x04, // 32-bit RO
    REG_V2_PH_LAST_ADDR_PREV_CYCLE = 0x06, // 32-bit RO
    REG_V2_PH_GATE_NUM = 0x08, // 32-bit RO
};


enum {
    REG_CTRL_BIT_CNT_LOCK = 1 << 1,
    REG_CONTROL_LOCK_SOFTCLEAR = 1 << 0
};

static const int MAX_HIST_COUNT = 32;
}

QList<AFI_SDB_ID> MscHistModule::getSupportedSdbIds()
{
    return {AFI_SDB_MSC_HISTOGRAMS};
}

void MscHistModule::post_init()
{
    BaseDeviceModule::post_init();
    buildParams = readBuildParam();
}

mlink::RegOpVector MscHistModule::setConfigRegOp(const MscHistConfig &hist)
{
    RegOpVector r;
    if(majorVersion() == 1) {
        for (int hId : hist.plot.keys()) {
            auto chRegBase = REG_BASE_HIST_PLOT_START_ADDR + REG_HIST_PLOT * hId;
            r.RegWrite32(chRegBase + REG_HIST_PLOT_SLICE_TIME,
                         hist.plot[hId].plotSlice * 1e3);
        }
    } else if(majorVersion() == 2) {
        auto nHist = getPlotHistCount();
        if(!hist.log.isEmpty()) {
            quint32 logType = 0;
            for(int hId=0; hId<getLogHistCount(); ++hId) {
                if(hist.log.value(hId).logTypeIsLen)
                    logType |= 1<<hId;
            }
            r.RegWrite32(buildParams->regLogHistBase + REG_V2_LH_TYPE_MASK, logType);
        }

        for(int hId=0; hId<nHist; ++hId) {
            const auto& hCfg = hist.plot.value(hId);
            auto chRegBase = buildParams->regPlotHistBase +
                (buildParams->regPlotHistMask+1) * hId;
            r.RegWrite32(chRegBase + REG_V2_PH_SLICE_TIME,
                         hCfg.plotSlice * 1e3);
        }
    }
    return r;
}

bool MscHistModule::writeConfig(const MscHistConfig &hist)
{
    auto r = setConfigRegOp(hist);
    return regOpExecRebased(r);
}

int MscHistModule::getPlotHistCount() const
{
    if(majorVersion() == 1) {
        if (histNum.has_value()) {
            int n = histNum.value();
            std::clamp(n, 0, MAX_HIST_COUNT);
            return n;
        }
    } else if(majorVersion() == 2) {
        if (buildParams.has_value())
            return buildParams->plotHistNum;
    }
    return 0;
}

RegOpVector MscHistModule::getStatusRegOp(MscHistStatus *result) const
{
    RegOpVector r;
    if(!buildParams)
        return r;

    // update & lock counters
    if(majorVersion() == 1) {
        r.RegWrite16(REG_CTRL, 0);
        r.RegWrite16(REG_CTRL, REG_CTRL_BIT_CNT_LOCK);
    } else {
        r.RegWrite16(REG_V2_LOCK, 0);
        r.RegWrite16(REG_V2_LOCK, 1);
    }

    const auto histCount = getPlotHistCount();
    result->chSt.resize(histCount);
    for (uint16_t i = 0; i < histCount; ++i) {
        if(majorVersion() == 1) {
            auto chRegBase = REG_BASE_HIST_PLOT_START_ADDR + REG_HIST_PLOT * i;
            r.RegRead32(chRegBase + REG_HIST_PLOT_CUR_RAM_ADDR,
                        &result->chSt[i].plot_current_ram_address);
        } else if(majorVersion() == 2) {
            auto chRegBase = buildParams->regPlotHistBase +
                (buildParams->regPlotHistMask+1) * i;
            r.RegRead32(chRegBase + REG_V2_PH_CUR_RAM_ADDR,
                        &result->chSt[i].plot_current_ram_address);
            r.RegRead32(chRegBase + REG_V2_PH_LAST_ADDR_PREV_GATE,
                        &result->chSt[i].plot_last_ram_addr_gate);
            r.RegRead32(chRegBase + REG_V2_PH_LAST_ADDR_PREV_CYCLE,
                        &result->chSt[i].plot_last_ram_addr_cycle);
            r.RegRead32(chRegBase + REG_V2_PH_GATE_NUM,
                        &result->chSt[i].gate_num);
        }
    }

    // unlock counters
    if(majorVersion() == 1) {
        r.RegWrite16(REG_CTRL, 0);
    } else {
        r.RegWrite16(REG_V2_LOCK, 0);
    }

    return r;
}

std::optional<MscHistBuildParams> MscHistModule::readBuildParam()
{
    MscHistBuildParams bp;
    RegOpVector r;

    quint32 id = 0;
    bp.majVer = majorVersion();
    switch (majorVersion()) {
    case 1:
    {
        r.RegRead16(REG_NUMBER_OF_RAM_BITS_HIST_TIME, &bp.ramBitsNumLogHist);
        r.RegRead16(REG_NUMBER_OF_RAM_BITS_HIST_PLOT, &bp.ramBitsNumHistPlot);
        break;
    }
    case 2:
    {
        r.RegRead32(REG32_ID, &id);
        r.RegRead32(REG_V2_CLOCK_FREQ, &bp.clockFreq);
        r.RegRead16(REG_V2_NUMBER_OF_CH, &bp.chNum);

        r.RegRead32(REG_V2_LH_RAM_BASE_ADDR, &bp.ramBaseAddrLogHist);
        r.RegRead16(REG_V2_LH_RAM_NUMBER_OF_RAM_BITS, &bp.ramBitsNumLogHist);
        r.RegRead16(REG_V2_LH_REG_BASE, &bp.regLogHistBase);
        r.RegRead16(REG_V2_LH_HIST_NUM, &bp.logHistNum);

        r.RegRead32(REG_V2_PH_RAM_BASE_ADDR, &bp.ramBaseAddrPlotHist);
        r.RegRead16(REG_V2_PH_RAM_NUMBER_OF_RAM_BITS, &bp.ramBitsNumHistPlot);
        r.RegRead16(REG_V2_PH_REG_BASE, &bp.regPlotHistBase);
        r.RegRead16(REG_V2_PH_REG_MASK, &bp.regPlotHistMask);
        r.RegRead16(REG_V2_PH_HIST_NUM, &bp.plotHistNum);
        break;
    }
    default:
        return {};
    }

    if (!regOpExecRebased(r))
        return {};
    if(majorVersion()>=2 && !getSupportedSdbIds().contains(static_cast<AFI_SDB_ID>(id)))
        return {};
    return bp;
}

bool MscHistModule::readStatus(MscHistStatus *result)
{
    auto r = getStatusRegOp(result);
    return regOpExecRebased(r);
}

std::optional<MscHistStatus> MscHistModule::readStatus()
{
    MscHistStatus result;
    if (readStatus(&result))
        return result;
    return {};
}

int MscHistModule::getLogHistCount() const
{
    Q_ASSERT(majorVersion() != 1);
    if(majorVersion() == 2) {
        if (buildParams.has_value())
            return buildParams->logHistNum;
    }
    return 0;

}
