#include "MscInputModule.h"

#include "MscInputConfig.h"
#include "regio/regop.h"
#include "sdb/AfiSdbId.h"

enum {
    REG_CTRL = 0x0,
    REG_CURRENT_STATE_OF_EXTERNAL_CE = 0x1,
    REG_CURRENT_STATE_OF_INPUT_HITS = 0x2,

    REG_NUMBER_OF_INPUT_HIT_CHANNELS = 0xD,
    REG_NUMBER_OF_HISTOGRAMMERS = 0xE,
    REG_NUMBER_OF_EXTERNAL_CE_INPUTS = 0xF,

    REG_CE_EDGE_TIME_BASE = 0x80,

    REG_HIST_CH_SOURCE_BASE = 0xC0,

    REG_GATE_LUT_BASE = 0x100,
    REG_HIST_CE_LUT_BASE = 0x180,
};

enum {
    REG_CTRL_BIT_CNT_LOCK = 1 << 1,
};

QList<AFI_SDB_ID> MscInputModule::getSupportedSdbIds()
{
    return {AFI_SDB_MSC_INPUT};
}

void MscInputModule::post_init()
{
    BaseDeviceModule::post_init();

    readBuildParam();
}

mlink::RegOpVector MscInputModule::setConfigRegOp(const MscInputConfig &cfg) const
{
    RegOpVector r;
    if (!buildParams)
        return {};

    int lutSizeBit = 1 << buildParams->ceInputs;
    quint16 lutSizeWord16 = lutSizeBit / 16;
    lutSizeWord16 = std::max(lutSizeWord16, quint16(1));

    for (uint16_t hId = 0; hId < buildParams->hists; ++hId) {
        auto chId = cfg[hId].histCh;
        r.RegWrite16(REG_HIST_CH_SOURCE_BASE + hId, chId);

        quint16 in_addr = REG_GATE_LUT_BASE + hId * lutSizeWord16;
        r.RegWrite16(in_addr, cfg[hId].histLut);

        in_addr = REG_HIST_CE_LUT_BASE + hId * lutSizeWord16;
        r.RegWrite16(in_addr, cfg[hId].ceLut);
    }
    return r;
}

int MscInputModule::getCeInputCount() const
{
    if (!buildParams.has_value())
        return 0;
    return buildParams->ceInputs;
}

RegOpVector MscInputModule::getStatusRegOp(MscInputStatus *result) const
{
    auto ceInputCount = getCeInputCount();
    result->ceInputCount = ceInputCount;
    result->ceInputEdgeTimesRaw.resize(ceInputCount);
    RegOpVector r;
    r.RegWrite16(REG_CTRL, REG_CTRL_BIT_CNT_LOCK); // lock counters
    r.RegRead16(REG_CURRENT_STATE_OF_EXTERNAL_CE, &result->ceInputStateRaw);
    r.RegRead16(REG_CURRENT_STATE_OF_INPUT_HITS, &result->hitState);
    for (uint16_t hId = 0; hId < ceInputCount; ++hId) {
        uint16_t addr = REG_CE_EDGE_TIME_BASE + hId * 8;
        r.RegRead64(addr, &result->ceInputEdgeTimesRaw[hId].first);     // rising
        r.RegRead64(addr + 4, &result->ceInputEdgeTimesRaw[hId].second); // falling
    }
    r.RegWrite16(REG_CTRL, 0); // un lock counters
    return r;
}

bool MscInputModule::writeConfig(const MscInputConfig &cfg)
{
    auto r = setConfigRegOp(cfg);
    return regOpExecRebased(r);
}

bool MscInputModule::readStatus(MscInputStatus *result)
{
    auto r = getStatusRegOp(result);
    return regOpExecRebased(r);
}

std::optional<MscInputStatus> MscInputModule::readStatus()
{
    MscInputStatus status;
    if (!readStatus(&status))
        return {};
    return status;
}

QVector<quint64> MscInputModule::readCmpState(int num_samples)
{
    QVector<quint16> cmpState(num_samples);
    RegOpVector r;
    for (int i = 0; i < num_samples; ++i)
        r.RegRead16(REG_CURRENT_STATE_OF_INPUT_HITS, &cmpState[i]); // read multiple times
    if (!regOpExecRebased(r))
        return {};
    QVector<quint64> result;
    for (int i = 0; i < num_samples; ++i)
        result.push_back(cmpState[i]);
    return result;
}

void MscInputModule::readBuildParam()
{
    buildParams.reset();

    MscInputBuildParams st;
    RegOpVector r;
    r.RegRead16(REG_NUMBER_OF_INPUT_HIT_CHANNELS, &st.channels);
    r.RegRead16(REG_NUMBER_OF_HISTOGRAMMERS, &st.hists);
    r.RegRead16(REG_NUMBER_OF_EXTERNAL_CE_INPUTS, &st.ceInputs);

    if (!regOpExecRebased(r))
        return;

    buildParams = st;
}
