#ifndef NETWORKPORTSTATUS_H
#define NETWORKPORTSTATUS_H

#include <array>
#include <QtCore>
#include <optional>

typedef union {
    struct {
        uint phyCsr: 1;
        uint pcsPmaCsr: 1;
        uint macCsr: 1;
        uint hwipCsr: 1;
        uint mstreamCore: 1;
        uint unused_bits_5_15: 11;
    };
    uint16_t raw;
} descr_cap_reg_t;

struct NetworkPortDescription
{
    friend bool operator==(const NetworkPortDescription &, const NetworkPortDescription &) = default;
    bool isValid() const { return valid; }
    descr_cap_reg_t capability_reg() const { return {.raw = capability}; }

    bool valid = false;
    std::optional<uint16_t> id;
    uint16_t capability = 0;
    uint16_t phyRegBase = 0;
    uint16_t pcsPmaRegBase = 0;
    uint16_t macRegBase = 0;
    uint16_t hwipRegBase = 0;
    uint16_t mstreamRegBase = 0;
};

typedef union {
    struct {
        uint sfp_present: 1;
        uint sfp_osd: 1;
        uint sfp_tx_fault: 1;
        uint tx_disable: 1;
        uint speed: 2;
        uint unused_bits_6_7: 2;
        uint gtx_rx_error: 1;
        uint unused_bits_9_11: 3;
        uint loopback_enable: 1;
        uint gtx_phy_reset: 1;
        uint gtx_resetdone: 1;
        uint gtx_mmcm_locked: 1;
        uint unused_bits_16_31: 16;
    };
    uint32_t raw;
} phy_status_reg_t;

struct PhyPortStatus
{
    friend bool operator==(const PhyPortStatus &, const PhyPortStatus &) = default;
    bool isValid() const { return valid; }
    QString toString() const;
    phy_status_reg_t status_reg() const {
        phy_status_reg_t reg;
        reg.raw = status;
        return reg;
    }

    bool valid = false;
    uint32_t status = 0;
};

enum {
    PCSPMA_STATUS_BIT_LINK_UP = 0x1,
    PCSPMA_STATUS_BIT_RX_ERROR = 0x2
};

typedef union {
    struct {
        uint rx_sig_det: 1;
        uint pcs_rx_locked_ll: 1;
        uint pcs_rx_locked: 1;
        uint pcs_rx_link_status: 1;
        uint teng_pcs_rx_link_status: 1;
        uint pma_link_status: 1;
        uint unused_bits_6_7: 2;
        uint pcs_hiber: 1;
        uint pcs_rx_hiber_lh: 1;
        uint pcs_ber_count: 1;
        uint pcs_test_patt_err_count: 1;
        uint speed: 3;
        uint duplex: 1;
        uint unused_bits_16_31: 16;
    };
    uint32_t raw;
} pcspma_status_reg_t;

struct PcspmaPortStatus
{
    enum { counters_size = 4 };
    enum { timers_size = 4 };
    friend bool operator==(const PcspmaPortStatus &, const PcspmaPortStatus &) = default;
    bool isValid() const { return valid; }
    QString toString() const;
    pcspma_status_reg_t status_reg() const {
        pcspma_status_reg_t reg;
        reg.raw = status;
        return reg;
    }

    bool valid = false;
    uint32_t status = 0;
    std::array<uint32_t, counters_size> counters = {};
    std::array<uint32_t, timers_size> timers = {};
};

struct MacPortStatus
{
    enum { counters_size = 14 };
    friend bool operator==(const MacPortStatus &, const MacPortStatus &) = default;
    bool isValid() const { return valid; }
    QString toString() const;

    bool valid = false;
    uint32_t status = 0;
    uint32_t mtu = 0;
    std::array<uint32_t, counters_size> counters = {};
};

struct HwipPortStatus
{
    enum { counters_size = 3 };
    enum { timers_size = 7 };
    friend bool operator==(const HwipPortStatus &, const HwipPortStatus &) = default;
    bool isValid() const { return valid; }
    bool hasRegioError() const;
    QString getRegioErrorStr() const;
    QString toString() const;

    bool valid = false;
    uint32_t status = 0;
    uint32_t host_ip = 0;
    uint64_t host_mac = 0;
    uint32_t mlink_config = 0;
    uint32_t master_ip = 0;
    uint32_t master_port = 0;
    uint32_t mstream_ip = 0;
    uint32_t mstream_port = 0;
    std::array<uint32_t, counters_size> counters = {};
    std::array<uint32_t, timers_size> timers = {};
    uint32_t regio_error_addr = 0;
    uint32_t regio_error_code = 0;
};

struct MstreamPortStatus
{
    enum { counters_size = 16 };
    enum { timers_size = 8 };
    friend bool operator==(const MstreamPortStatus &, const MstreamPortStatus &) = default;
    bool isValid() const { return valid; }
    QString toString() const;
    int getAckSizeLimit() const;
    int getRetxBufSize() const;
    int getRetxBufCount() const;

    bool valid = false;
    uint32_t status = 0;
    uint16_t version() const { return status >> 16; }
    uint16_t mtu_min = 0;
    uint16_t mtu_max = 0;
    uint16_t mtu = 0;
    uint16_t max_packet_size = 0;
    uint32_t defrag_timeout = 0;
    std::array<uint32_t, counters_size> counters = {};
    std::array<uint32_t, timers_size> timers = {};
    uint32_t retx_ack_time = 0;
    uint32_t retx_est_rtt = 0;
    uint32_t retx_config = 0;
    uint32_t clk_khz = 0;
};

struct NetworkPortStatus
{
    friend bool operator==(const NetworkPortStatus &, const NetworkPortStatus &) = default;
    NetworkPortStatus & operator-=(const NetworkPortStatus& st) {
        return *this;
    }
    bool isValid() const;
    QString toString() const;

    NetworkPortDescription descr;
    PhyPortStatus phy;
    PcspmaPortStatus pcspma;
    MacPortStatus mac;
    HwipPortStatus hwip;
    MstreamPortStatus mstream;
};
Q_DECLARE_METATYPE(NetworkPortStatus);

QDebug operator <<(QDebug dbg, const PhyPortStatus &r);
QDebug operator <<(QDebug dbg, const PcspmaPortStatus &r);
QDebug operator <<(QDebug dbg, const MacPortStatus &r);
QDebug operator <<(QDebug dbg, const HwipPortStatus &r);
QDebug operator <<(QDebug dbg, const MstreamPortStatus &r);
QDebug operator <<(QDebug dbg, const NetworkPortStatus &r);

#endif // NETWORKPORTSTATUS_H
