/*
**    Copyright 2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef RUNLOGICFSM_H
#define RUNLOGICFSM_H

#include <QDebug>

class RunLogicFsm
{
public:
    enum State {Init, SentReset, Reset, SentRun, Run, SentPause, Pause};
    bool transitionAllowed(State toState) const {
        return transitionAllowed(state, toState);
    }
    static bool transitionAllowed(State fromState, State toState) {
        switch (fromState) {
        case Init: return (toState == SentReset || toState == SentRun);
        case SentReset: return (toState == Reset);
        case Reset: return (toState == Reset || toState == SentReset || toState == SentRun);
        case SentRun: return (toState == Run);
        case Run: return (toState == SentReset || toState == SentPause);
        case SentPause: return (toState == Pause);
        case Pause: return (toState == SentReset || toState == SentRun);
        default:
            return false;
        }
    }
    static bool allowedReset(State fromState) {
        return transitionAllowed(fromState, SentReset);
    }
    static bool allowedRunToggle(State fromState) {
        return transitionAllowed(fromState, SentPause)
                || transitionAllowed(fromState, SentRun);
    }
    bool setState(State toState) {
        if (!transitionAllowed(toState)) {
            qWarning() << "Ignored transition" << state << "to" << toState;
            return false;
        }
        qInfo() << "Transition" << state << "to" << toState;
        state = toState;
        return true;
    }
    State getState() const { return state; }
    static QString stateStr(State state)
    {
        static const QMap<State, QString> map{{Init, "Init"},
                                              {SentReset, "SentReset"},
                                              {Reset, "Reset"},
                                              {SentRun, "SentRun"},
                                              {Run, "Run"},
                                              {SentPause, "SentPause"},
                                              {Pause, "Pause"}};
        return map[state];
    }
    friend QDebug operator<<(QDebug dbg, State state) {
        dbg << stateStr(state);
        return dbg;
    }

    int seq = 0;
private:
    State state = Init;
};

#endif // RUNLOGICFSM_H
