/*
**    Copyright 2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "WaveTrigModule.h"

#include "WaveTrigConfig.h"
#include "regio/regop.h"
#include "sdb/AfiSdbId.h"

namespace {

union ch_reg_t {
    unsigned all;
    struct {
        unsigned enable: 1;
        unsigned polarity: 1;
        unsigned unused_14: 14;
    } bit;
};

enum {
    REG32_ID = 0x0,
    REG_CONTROL = 0x2,
    REG_LOCK = 0x3,
    REG_CHANNEL_COUNT = 0x4,
    REG_TRIG_LEVEL_BASE = 0xA,
    REG_SETTING_BASE = 0xB,
};

constexpr static const quint16 REG_CONTROL_BIT_RESET = 1;
} // namespace

WaveTrigModule::WaveTrigModule(const MregBusInterface &bus)
    : BaseDeviceModule(bus)
{
}

QList<AFI_SDB_ID> WaveTrigModule::getSupportedSdbIds()
{
    return {AFI_SDB_WAVE_TRIG};
}

void WaveTrigModule::post_init()
{
    buildParams.reset();
    WaveTrigBuildParams bp;
    RegOpVector r;
    r.RegRead32(REG32_ID, &bp.id);
    r.RegRead16(REG_CHANNEL_COUNT, &bp.nch);
    r.RegRead16(REG_SETTING_BASE, &bp.settOffset);
    r.RegRead16(REG_TRIG_LEVEL_BASE, &bp.trigLevelOffset);
    if (regOpExecRebased(r))
        buildParams = bp;
}

bool WaveTrigModule::writeConfig(const WaveTrigConfig &config)
{
    if (!buildParams)
        return false;

    RegOpVector r;
    r.RegWrite16(REG_CONTROL, REG_CONTROL_BIT_RESET);
    for (int i=0; i<buildParams->nch; i++) {
        const auto &chConf = config.getChannel(i);
        ch_reg_t ctrl = {};
        ctrl.bit.polarity = chConf.edge==WaveTrigChannelConfig::Edge::Rising ? 0 : 1;
        ctrl.bit.enable = chConf.en ? 1 : 0;
        r.RegWrite16(buildParams->settOffset + i, ctrl.all);
        r.RegWrite16(buildParams->trigLevelOffset + i, chConf.trigLevel);
    }
    r.RegWrite16(REG_CONTROL, 0);
    return regOpExecRebased(r);
}
