#include "FeLinkTreeModel.h"

#include <QFont>
#include <QColor>
#include <QTime>
#include <QTimer>

#include "base/DeviceId.h"
#include "base/DeviceIndexEnabled.h"
#include "base/Globals.h"

namespace {
bool FeLinkChConfig::*getConfigBoolPtr(FeLinkTreeColumns col) {
    switch(col) {
    case FeLinkTreeColumns::COL_RESET:     return &FeLinkChConfig::reset;
    case FeLinkTreeColumns::COL_RESET_SOFT_PLL:     return &FeLinkChConfig::resetSoftPll;
    case FeLinkTreeColumns::COL_RECV_TRIG: return &FeLinkChConfig::recvTrig;
    case FeLinkTreeColumns::COL_REMOTE_RESET: return &FeLinkChConfig::remoteReset;
    default: return nullptr;
    }
}


QVariant getRecursiveState(const QModelIndex &mIndex)
{
    auto model = mIndex.model();
    auto rows = model->rowCount(mIndex);
    if(rows == 0) {
        return model->data(mIndex, Qt::CheckStateRole);
    }
    if(rows) {
        QVariant allSt;
        for(int r=0; r<rows; ++r) {
            auto childIndex = model->index(r, mIndex.column(), mIndex);
            auto st = getRecursiveState(childIndex);
            if(st.isNull())
                 continue;
            if(allSt.isNull())
                allSt = st;
            else if(allSt != st)
                return Qt::PartiallyChecked;
        }
        return allSt;
    }
    return {};
}


QString getStateTimeHr(quint32 time)
{
    const static quint32 DAY = 24*3600*1000;
    int days = time / DAY;
    return (days ? QString("%1d ").arg(days) : "") +
            QTime::fromMSecsSinceStartOfDay(time%DAY).toString("hh:mm:ss.zzz");
}

static QMap<DeviceIndex, int> buildNchMap(const BaseDeviceAppConfig &config,
                                          const QMap<DeviceIndex, QMap<int, FeLinkStatus>> &dev_status)
{
    QMap<DeviceIndex, int> nch;
    for (auto devIt = config.knownConfigs.constBegin(); devIt != config.knownConfigs.cend(); ++devIt) {
        auto cfgNCh = devIt->feLink.chCfg.size();
        auto stNCh = dev_status.value(devIt.key()).size();
        nch.insert(devIt.key(), std::max(cfgNCh, stNCh));
    }
    return nch;
}

static bool needReset(const BaseDeviceAppConfig &config, const BaseDeviceAppConfig &newConfig)
{
    auto skipDisabledIt = [](QMap<DeviceIndex, ModularDeviceConfig>::const_iterator &it,
            const QMap<DeviceIndex, ModularDeviceConfig> &map) {
      while(it != map.cend()) {
          if(it->enabled)
              return;
          ++it;
      }
    };
    auto it1 = config.knownConfigs.constBegin();
    auto it2 = newConfig.knownConfigs.constBegin();
    while(true) {
        skipDisabledIt(it1, config.knownConfigs);
        skipDisabledIt(it2, newConfig.knownConfigs);
        if(it1 == config.knownConfigs.cend() ||
                  it2 == newConfig.knownConfigs.cend())
            break;

        if(it1->feLink.chCfg.size() != it2->feLink.chCfg.size())
            return true;
        ++it1; ++it2;
    }
    return it1 == config.knownConfigs.cend() &&
            it2 == newConfig.knownConfigs.cend();
}

static bool needReset(const QMap<DeviceIndex, QMap<int, FeLinkStatus>> &stMap,
                       const DeviceIndex &index,
                       const QMap<int, FeLinkStatus> &st)
{
    if (!stMap.contains(index))
        return true;
    return stMap[index].keys() != st.keys();
}

static bool hasChanges(const ModularDeviceConfig &c1, const ModularDeviceConfig &c2)
{
    if (c1.feLink.chCfg != c2.feLink.chCfg)
        return true;
    return false;
}

static bool hasChanges(const BaseDeviceAppConfig &c1, const BaseDeviceAppConfig &c2)
{
    if (hasChanges(c1.defaultSetup, c2.defaultSetup))
        return true;

    for (auto it1=c1.knownConfigs.cbegin(); it1!=c1.knownConfigs.cend(); ++it1) {
        auto it2 = c2.knownConfigs.constFind(it1.key());
        if (it2 == c2.knownConfigs.cend() || hasChanges(*it1, *it2))
            return true;
    }
    return false;
}

const quint32 FeLinkCounter::* getCounterFieldByCol(FeLinkTreeColumns col) {
    switch(col) {
    case FeLinkTreeColumns::COL_RX_CODE_ERR:
        return &FeLinkCounter::RX_code_error;
    case FeLinkTreeColumns::COL_RX_CRC_ERR:
        return &FeLinkCounter::RX_CRC_error;

    case FeLinkTreeColumns::COL_SYNC_FR_SENT:
        return &FeLinkCounter::sync_fr_sent;
    case FeLinkTreeColumns::COL_SYNC_FR_RECV:
        return &FeLinkCounter::sync_fr_received;
    case FeLinkTreeColumns::COL_TIME_FR_SENT:
        return &FeLinkCounter::time_fr_sent;
    case FeLinkTreeColumns::COL_TIME_FR_RECV:
        return &FeLinkCounter::time_fr_received;
    case FeLinkTreeColumns::COL_TRIG_FR_SENT:
        return &FeLinkCounter::trigger_fr_sent;
    case FeLinkTreeColumns::COL_TRIG_FR_RECV:
        return &FeLinkCounter::trigger_fr_received;
    case FeLinkTreeColumns::COL_XOFF_FR_SENT:
        return &FeLinkCounter::xoff_fr_sent;
    case FeLinkTreeColumns::COL_XOFF_FR_RECV:
        return &FeLinkCounter::xoff_fr_received;
    case FeLinkTreeColumns::COL_DATA_FR_SENT:
        return &FeLinkCounter::data_fr_sent;
    case FeLinkTreeColumns::COL_DATA_FR_RECV:
        return &FeLinkCounter::data_fr_received;

    case FeLinkTreeColumns::COL_UNKNOWN_FR:
        return &FeLinkCounter::unknown_received;
    case FeLinkTreeColumns::COL_RESET_PHY_CNT:
        return &FeLinkCounter::phy_reset;
    default:
        Q_UNREACHABLE();
    }
}
}

FeLinkTreeModel::FeLinkTreeModel(QObject *parent)
    : QAbstractItemModel(parent)
    , delayedSendConfigTimer(new QTimer(this))
{
    setObjectName(__PRETTY_FUNCTION__);
    delayedSendConfigTimer->setInterval(100);
    delayedSendConfigTimer->setSingleShot(true);
    connect(delayedSendConfigTimer, &QTimer::timeout,
            this, &FeLinkTreeModel::postCheckStateChange);
}

QVariant FeLinkTreeModel::data(const QModelIndex &index, int role) const
{
    if(!index.isValid())
        return QVariant();
    switch (role) {
    case Qt::DisplayRole:   return displayData(index);
    case Qt::FontRole:      return fontData(index);
    case Qt::ForegroundRole: return foregroundData(index);
    case Qt::TextAlignmentRole:   return textAlignmentData(index);
    case Qt::ToolTipRole:   return toolTipData(index);
    case Qt::CheckStateRole: return checkStateData(index);
    }
    return QVariant();
}

bool FeLinkTreeModel::setData(const QModelIndex &index, const QVariant &value, int role)
{
    if(!index.isValid())
        return false;

    switch (role) {
    case Qt::CheckStateRole:
    {
        auto col = getColumnByIndex(index);
        auto ptr = getConfigBoolPtr(col);
        if(!ptr)
            return false;

        const auto item = getItemByIndex(index);
        if(item->ch == -1) {
            auto rows = rowCount(index);
            bool ok = true;
            for(int r=0; r<rows; ++r){
                ok &= setData(this->index(r,index.column(),index), value, role);
            }
            return ok;
        } else {
            Q_ASSERT(!item->dev.isNull());

            auto devIt = config.knownConfigs.find(item->dev);
            Q_ASSERT(devIt != config.knownConfigs.end());
            if(devIt == config.knownConfigs.end())
                return false;
            Q_ASSERT(value!=Qt::PartiallyChecked);

            auto &chCfg = devIt->feLink.chCfg[item->ch];
            chCfg.*ptr = (value==Qt::Checked) ? true : false;
            enqueueConfigSend();
        }
        return true;
    }
    default:;
    }
    return false;
}

Qt::ItemFlags FeLinkTreeModel::flags(const QModelIndex &index) const
{
    if(!index.isValid())
        return Qt::NoItemFlags;
    auto f = QAbstractItemModel::flags(index);
    switch (getColumnByIndex(index)) {
    case FeLinkTreeColumns::COL_RESET:
    case FeLinkTreeColumns::COL_RESET_SOFT_PLL:
    case FeLinkTreeColumns::COL_REMOTE_RESET:
        f |= Qt::ItemIsUserCheckable | Qt::ItemIsEnabled;
        break;
    case FeLinkTreeColumns::COL_RECV_TRIG:
    {
        const auto item = getItemByIndex(index);
        auto stOpt = getStatus(index);
        if((item->ch == -1) || (stOpt && stOpt->baseStatus.link_ok))
            f |= Qt::ItemIsUserCheckable | Qt::ItemIsEnabled;
        break;
    }
    default:;
    }
    if(!enable)
        f = f & (~Qt::ItemIsEnabled);
    return f;
}

QVariant FeLinkTreeModel::headerData(int section, Qt::Orientation orientation, int role) const
{
    if(orientation != Qt::Horizontal)
        return QVariant();
    if(role != Qt::DisplayRole)
        return QVariant();
    switch (static_cast<FeLinkTreeColumns>(section)) {
    case FeLinkTreeColumns::COL_NAME: return "Name";
    case FeLinkTreeColumns::COL_RESET: return "Reset";
    case FeLinkTreeColumns::COL_RESET_SOFT_PLL: return "ResetSoftPll";
    case FeLinkTreeColumns::COL_RESET_PHY_BUT: return "PhyReset";
    case FeLinkTreeColumns::COL_REMOTE_RESET: return "RemoteReset";
    case FeLinkTreeColumns::COL_RECV_TRIG: return "RecvTrig";
    case FeLinkTreeColumns::COL_COMMUNICATION_ENV: return "Enviroment";
    case FeLinkTreeColumns::COL_IS_MASTER: return "Role";
    case FeLinkTreeColumns::COL_REMOTE_ID: return "Remote Id";
    case FeLinkTreeColumns::COL_REMOTE_SERIAL: return "Remote Serial";
    case FeLinkTreeColumns::COL_COMMUNICATION_REMOTE_ENV: return "Remote Env.";
    case FeLinkTreeColumns::COL_STATUS: return "Status";
    case FeLinkTreeColumns::COL_SFP_STATUS: return "SFP st.";
    case FeLinkTreeColumns::COL_PHASE_ADJ_EN: return "Phase adj.";
    case FeLinkTreeColumns::COL_BIT_SLIDE: return "BitSlide";
    case FeLinkTreeColumns::COL_REMOTE_VER: return "Remote ver.";
    case FeLinkTreeColumns::COL_TAI: return "TAI";
    case FeLinkTreeColumns::COL_CUR_LINK_STATE_TIME: return "Cur. link state time";
    case FeLinkTreeColumns::COL_RX_CODE_ERR: return "Rx code err";
    case FeLinkTreeColumns::COL_RX_CRC_ERR: return "Rx crc err";
    case FeLinkTreeColumns::COL_SYNC_FR_SENT: return "Sync fr. sent";
    case FeLinkTreeColumns::COL_SYNC_FR_RECV: return "Sync fr. recv";
    case FeLinkTreeColumns::COL_TRIG_FR_SENT: return "Trig fr. sent";
    case FeLinkTreeColumns::COL_TRIG_FR_RECV: return "Trig fr. recv";
    case FeLinkTreeColumns::COL_TIME_FR_SENT: return "Time fr. sent";
    case FeLinkTreeColumns::COL_TIME_FR_RECV: return "Time fr. recv";
    case FeLinkTreeColumns::COL_XOFF_FR_SENT: return "XOff fr. sent";
    case FeLinkTreeColumns::COL_XOFF_FR_RECV: return "XOff fr. recv";
    case FeLinkTreeColumns::COL_DATA_FR_SENT: return "Data fr. sent";
    case FeLinkTreeColumns::COL_DATA_FR_RECV: return "Data fr. recv";
    case FeLinkTreeColumns::COL_UNKNOWN_FR: return "Unknown fr.";
    case FeLinkTreeColumns::COL_RESET_PHY_CNT: return "Reset phy";
    default:
        return QVariant();
    }
}

QModelIndex FeLinkTreeModel::index(int row, int column, const QModelIndex &parent) const
{
    if(!hasIndex(row, column, parent))
        return QModelIndex();
    const FeLinkTreeModelItem *parentItem = parent.isValid() ?
        static_cast<FeLinkTreeModelItem*>(parent.internalPointer()) :
        &rootItem;
    FeLinkTreeModelItem *childItem = parentItem->child(row);
    if(childItem) {
        return createIndex(row, column, childItem);
    }
    return QModelIndex();
}

QModelIndex FeLinkTreeModel::parent(const QModelIndex &index) const
{
    if(!index.isValid())
        return QModelIndex();

    const FeLinkTreeModelItem *childItem = static_cast<FeLinkTreeModelItem*>(index.internalPointer());
    if(!childItem)
        return QModelIndex();
    FeLinkTreeModelItem *parentItem = childItem->parent;
    if(parentItem == &rootItem)
        return QModelIndex();

    return createIndex(parentItem->row(), 0, parentItem);
}

int FeLinkTreeModel::rowCount(const QModelIndex &parent) const
{
    const auto *parentItem = parent.isValid() ?
        static_cast<FeLinkTreeModelItem*>(parent.internalPointer()) :
        &rootItem;
    return parentItem->childCount();
}

void FeLinkTreeModel::setEnabled(bool en)
{
    if(enable == en)
        return;

    enable = en;
    auto top = index(0,0);
    auto lastDev = index(rowCount(top)-1, columnCount(top)-1, top);
    auto lastCh = index(rowCount(lastDev)-1, columnCount(lastDev)-1, lastDev);
    emit dataChanged(top, lastCh);
}

void FeLinkTreeModel::feLinkStatusUpdated(const DeviceIndex &index, const QMap<int, FeLinkStatus> &status)
{
    bool performReset = needReset(dev_status, index, status);
    bool hasAnyChanges = performReset ? true : (dev_status[index] != status);

    dev_status[index] = status;
    if(performReset) {
        doReset();
    } else {
        if(hasAnyChanges)
            sendDataChanged();
    }
}

void FeLinkTreeModel::clearStats()
{
    devStatusOffset = dev_status;
    sendDataChanged();
}

void FeLinkTreeModel::onDeviceConnected(const DeviceIndex &devIndex)
{
    devStatusOffset.remove(devIndex);
    sendDataChanged();
}

void FeLinkTreeModel::onFsmStateChange(const FsmState &state)
{
    bool inStateErrReady = state == FsmState::Ready || state == FsmState::Error;
    setEnabled(inStateErrReady || Globals::getDebugMode());
    if(state == FsmState::Starting)
        clearStats();
}

void FeLinkTreeModel::postCheckStateChange()
{
    sendDataChanged(QVector<int>{Qt::CheckStateRole});
    emit configChanged(this->config);
}

void FeLinkTreeModel::enqueueConfigSend()
{
    delayedSendConfigTimer->start();
}

void FeLinkTreeModel::doReset()
{
    beginResetModel();
    rootItem.children.clear();
    const auto &nchMap = buildNchMap(config, dev_status);
    if(!nchMap.isEmpty()) {
        FeLinkTreeModelItem *allItem = new FeLinkTreeModelItem(&rootItem);
        rootItem.addChild(allItem);

        for(auto it = nchMap.cbegin(); it != nchMap.cend(); ++it) {
            auto devItem = new FeLinkTreeModelItem(allItem);
            devItem->dev = it.key();
            auto nch = *it;
            for (int ch = 0; ch < nch; ch++) {
                auto chItem = new FeLinkTreeModelItem(devItem);
                chItem->ch = ch;
                devItem->addChild(chItem);
            }
            allItem->addChild(devItem);
        }
    }
    endResetModel();
}

void FeLinkTreeModel::sendDataChanged(QVector<int> roles)
{
    auto top = index(0,0);
    auto lastDev = index(rowCount(top)-1, columnCount(top)-1, top);
    auto lastCh = index(rowCount(lastDev)-1, columnCount(lastDev)-1, lastDev);
    emit dataChanged(top, lastCh, roles);
}

QVariant FeLinkTreeModel::displayData(const QModelIndex &index) const
{
    const auto item = getItemByIndex(index);
    const auto &col = getColumnByIndex(index);
    auto stOpt = getStatus(index);
    if(col == FeLinkTreeColumns::COL_NAME) {
        if(item->dev.isNull())
            return onFreeze ? "All (freezed)" : "All";
        if(item->ch == -1)
            return item->dev.getSerialStr();
        return QString::number(stOpt ? stOpt->txInfo.ch : item->ch);
    }

    if(!stOpt)
        return {};

    const auto &st = *stOpt;
    switch (col) {
    case FeLinkTreeColumns::COL_RESET:
    case FeLinkTreeColumns::COL_RESET_SOFT_PLL:
    case FeLinkTreeColumns::COL_RECV_TRIG:
    case FeLinkTreeColumns::COL_REMOTE_RESET:
        if(checkStateData(index).value<Qt::CheckState>() == Qt::Checked)
            return QChar(0x2713);
        return {};
    case FeLinkTreeColumns::COL_COMMUNICATION_ENV:
        return st.getCommunicationEnvStr(false);
    case FeLinkTreeColumns::COL_IS_MASTER:
        return st.txInfo.isMaster ? "Master" : "Slave";
    case FeLinkTreeColumns::COL_REMOTE_ID:
        return getDeviceTypeName(st.remoteIndex.getDevId());
    case FeLinkTreeColumns::COL_REMOTE_SERIAL:
        if(st.remoteIndex.getSerial()) return st.remoteIndex.getSerialStr(); break;
    case FeLinkTreeColumns::COL_COMMUNICATION_REMOTE_ENV:
        return st.getCommunicationEnvStr(true);
    case FeLinkTreeColumns::COL_STATUS:
        return st.getInfoStr(true);
    case FeLinkTreeColumns::COL_SFP_STATUS:
        if(st.communication_evn == FE_LINK_COM_ENV_QSFP
                || st.communication_evn == FE_LINK_COM_ENV_SFP)
            return st.getSfpStatusStr(false);
        break;
    case FeLinkTreeColumns::COL_PHASE_ADJ_EN:
        return st.phase_adjust_enabled ? "en" : "-";
    case FeLinkTreeColumns::COL_BIT_SLIDE:
        return st.RX_Bitslide;
    case FeLinkTreeColumns::COL_REMOTE_VER:
        return st.remoteVer.getInfoStr();
    case FeLinkTreeColumns::COL_TAI:
        return QDateTime::fromSecsSinceEpoch(st.timeCode.sec).toString("yyyy-MM-dd HH:mm:ss");
    case FeLinkTreeColumns::COL_CUR_LINK_STATE_TIME:
        return getStateTimeHr(st.Current_link_state_time);
    case FeLinkTreeColumns::COL_RX_CODE_ERR:
    case FeLinkTreeColumns::COL_RX_CRC_ERR:
    case FeLinkTreeColumns::COL_SYNC_FR_SENT:
    case FeLinkTreeColumns::COL_SYNC_FR_RECV:
    case FeLinkTreeColumns::COL_TIME_FR_SENT:
    case FeLinkTreeColumns::COL_TIME_FR_RECV:
    case FeLinkTreeColumns::COL_TRIG_FR_SENT:
    case FeLinkTreeColumns::COL_TRIG_FR_RECV:
    case FeLinkTreeColumns::COL_XOFF_FR_SENT:
    case FeLinkTreeColumns::COL_XOFF_FR_RECV:
    case FeLinkTreeColumns::COL_DATA_FR_SENT:
    case FeLinkTreeColumns::COL_DATA_FR_RECV:
    case FeLinkTreeColumns::COL_UNKNOWN_FR:
    case FeLinkTreeColumns::COL_RESET_PHY_CNT:
    {
        auto p = getCounterFieldByCol(col);
        auto val = st.counter.*p;
        if(auto stOffset = getStatusOffset(item->dev, item->ch)) {
            val -= (*stOffset).counter.*p;
        }
        return QString("%L1").arg(val);
    }
    default:;
    }
    return {};
}

QVariant FeLinkTreeModel::fontData(const QModelIndex &index) const
{
    const auto item = getItemByIndex(index);
    QFont font;
    font.setFamily("Monospace");
//    if(item->dev.isNull()) {
//        font.setItalic(true);
//        font.setBold(true);
//    } else
        if(item->ch == -1) {
        font.setItalic(true);
//    } else {
//            fg.setItalic(true);
    }
    return font;
}

QVariant FeLinkTreeModel::foregroundData(const QModelIndex &index) const
{
    static const auto &errCol = QColor(Qt::red);

    auto stOpt = getStatus(index);
    if(!stOpt)
        return {};

    const FeLinkStatus &st = *stOpt;
    auto col = getColumnByIndex(index);
    switch (col) {
    case FeLinkTreeColumns::COL_STATUS:
        if(!st.allOk())
            return errCol;
        if(st.baseStatus.remote_link_ok)
            return QColor(Qt::darkGreen);
        break;
    case FeLinkTreeColumns::COL_TAI:
        if(!st.timeCode.isValid())
            return errCol;
        break;
    case FeLinkTreeColumns::COL_RX_CODE_ERR:
    case FeLinkTreeColumns::COL_RX_CRC_ERR:
    case FeLinkTreeColumns::COL_UNKNOWN_FR:
    case FeLinkTreeColumns::COL_RESET_PHY_CNT:
    {
        auto p = getCounterFieldByCol(col);
        auto val = st.counter.*p;
        const auto &item = getItemByIndex(index);
        if(auto stOffset = getStatusOffset(item->dev, item->ch)) {
            val -= (*stOffset).counter.*p;
        }

        if(val > 0)
            return errCol;
        break;
    }
    default:;
    }
    return {};
}

QVariant FeLinkTreeModel::textAlignmentData(const QModelIndex &index) const
{
    const auto &col = getColumnByIndex(index);
    switch (col) {
    case FeLinkTreeColumns::COL_REMOTE_VER:
        return Qt::AlignHCenter;
    case FeLinkTreeColumns::COL_BIT_SLIDE:
    case FeLinkTreeColumns::COL_TAI:
    case FeLinkTreeColumns::COL_CUR_LINK_STATE_TIME:
    case FeLinkTreeColumns::COL_RX_CODE_ERR:
    case FeLinkTreeColumns::COL_RX_CRC_ERR:
    case FeLinkTreeColumns::COL_SYNC_FR_SENT:
    case FeLinkTreeColumns::COL_SYNC_FR_RECV:
    case FeLinkTreeColumns::COL_TIME_FR_SENT:
    case FeLinkTreeColumns::COL_TIME_FR_RECV:
    case FeLinkTreeColumns::COL_TRIG_FR_SENT:
    case FeLinkTreeColumns::COL_TRIG_FR_RECV:
    case FeLinkTreeColumns::COL_XOFF_FR_SENT:
    case FeLinkTreeColumns::COL_XOFF_FR_RECV:
    case FeLinkTreeColumns::COL_DATA_FR_SENT:
    case FeLinkTreeColumns::COL_DATA_FR_RECV:
    case FeLinkTreeColumns::COL_UNKNOWN_FR:
    case FeLinkTreeColumns::COL_RESET_PHY_CNT:
        return Qt::AlignRight;
    default:;
    }
    return {};
}

QVariant FeLinkTreeModel::toolTipData(const QModelIndex &index) const
{
    auto stOpt = getStatus(index);
    if(!stOpt)
        return {};
    const auto &st = *stOpt;

    const auto &col = getColumnByIndex(index);
    switch (col) {
    case FeLinkTreeColumns::COL_STATUS:
        return st.getInfoStr(false).replace(',', '\n');
    case FeLinkTreeColumns::COL_SFP_STATUS:
        return st.getSfpStatusStr(true);
    case FeLinkTreeColumns::COL_CUR_LINK_STATE_TIME:
        return QString("%L1").arg(st.Current_link_state_time);

    case FeLinkTreeColumns::COL_RX_CODE_ERR:
    case FeLinkTreeColumns::COL_RX_CRC_ERR:
    case FeLinkTreeColumns::COL_SYNC_FR_SENT:
    case FeLinkTreeColumns::COL_SYNC_FR_RECV:
    case FeLinkTreeColumns::COL_TIME_FR_SENT:
    case FeLinkTreeColumns::COL_TIME_FR_RECV:
    case FeLinkTreeColumns::COL_TRIG_FR_SENT:
    case FeLinkTreeColumns::COL_TRIG_FR_RECV:
    case FeLinkTreeColumns::COL_XOFF_FR_SENT:
    case FeLinkTreeColumns::COL_XOFF_FR_RECV:
    case FeLinkTreeColumns::COL_DATA_FR_SENT:
    case FeLinkTreeColumns::COL_DATA_FR_RECV:
    case FeLinkTreeColumns::COL_UNKNOWN_FR:
    case FeLinkTreeColumns::COL_RESET_PHY_CNT:
    {
        auto p = getCounterFieldByCol(col);
        auto tt = QString("%L1").arg(st.counter.*p);
        const auto item = getItemByIndex(index);
        if(auto stOffset = getStatusOffset(item->dev, item->ch)) {
            tt += QString(" - %L1").arg((*stOffset).counter.*p);
        }
        return tt;
    }

    default:;
    }
    return {};
}

QVariant FeLinkTreeModel::checkStateData(const QModelIndex &index) const
{
    const auto item = getItemByIndex(index);
    auto col = getColumnByIndex(index);
    auto ptr = getConfigBoolPtr(col);
    if(!ptr)
        return {};

    if(rowCount(index)) {
        return getRecursiveState(index);
    } else if(item->ch == -1) {
        return {};
    }
    auto v = config.knownConfigs.value(item->dev).feLink.chCfg;
    if(v.size() <= item->ch)
        return {};

    return v[item->ch].*ptr ? Qt::Checked : Qt::Unchecked;
}

FeLinkTreeColumns FeLinkTreeModel::getColumnByIndex(const QModelIndex &index) const
{
    return static_cast<FeLinkTreeColumns>(index.column());
}

FeLinkTreeModelItem *FeLinkTreeModel::getItemByIndex(const QModelIndex &index) const
{
    return static_cast<FeLinkTreeModelItem*>(index.internalPointer());
}

std::optional<FeLinkStatus> FeLinkTreeModel::getStatus(const QModelIndex &index) const
{
    const auto item = getItemByIndex(index);
    if(item->ch == -1 || item->dev.isNull())
        return {};

    auto devIt = dev_status.constFind(item->dev);
    if(devIt == dev_status.cend())
        return {};
    auto chIt = devIt->constFind(item->ch);
    if(chIt == devIt->cend())
        return {};
    return *chIt;
}

std::optional<FeLinkStatus> FeLinkTreeModel::getStatusOffset(const DeviceIndex &dev, int ch) const
{
    auto devIt = devStatusOffset.constFind(dev);
    if(devIt == devStatusOffset.cend())
        return {};
    auto chIt = devIt->constFind(ch);
    if(chIt == devIt->cend())
        return {};

    return *chIt;
}

int FeLinkTreeModel::getDevRow(const DeviceIndex &di) const
{
    FeLinkTreeModelItem * allItem = nullptr;
    if(!rootItem.children.isEmpty())
        allItem = rootItem.children.first();
    if(allItem) {
        for(int r=0; r<allItem->children.size(); ++r)
            if(allItem->children[r]->dev == di)
                return r;
    }

    return -1;
}

void FeLinkTreeModel::setConfig(const ConfigKey &config_key, const BaseDeviceAppConfig &newConfig)
{
    Q_UNUSED(config_key)
    if (config == newConfig)
        return;
    bool performReset = needReset(config, newConfig);
    bool hasAnyChanges = performReset ? true : hasChanges(config, newConfig);

    config = newConfig;
    if (performReset) {
        doReset();
    } else {
        if(hasAnyChanges)
            sendDataChanged();
    }
}

void FeLinkTreeModel::setFeLinkBp(const DeviceIndex &index, const FeLinkBuildParams &bp)
{
    auto it = devBp.find(index);
    if(it != devBp.end() && *it == bp)
        return;

    devBp[index] = bp;
    auto devIt = config.knownConfigs.find(index);
    if(devIt == config.knownConfigs.end())
        return;
    auto curSize = devIt->feLink.chCfg.size();
    if(curSize == bp.chNum)
        return;

    if(curSize < bp.chNum) {
        while(curSize < bp.chNum) {
            FeLinkChConfig ch{.recvTrig = !bp.chInfo.value(curSize).isMaster};
            devIt->feLink.chCfg.push_back(ch);
            ++curSize;
        }
    } else {
        devIt->feLink.chCfg.resize(bp.chNum);
    }

    postCheckStateChange();
}

void FeLinkTreeModel::setDeviceList(const DeviceIndexEnabledMap &map)
{
    setDeviceSet(map.getEnabled());
}

void FeLinkTreeModel::setDeviceSet(const DeviceIndexSet &set)
{
    if(deviceSet == set)
        return;
    deviceSet = set;

    for(auto i : dev_status.keys()) {
        if(!deviceSet.contains(i))
            dev_status.remove(i);
    }
    for(auto i : devStatusOffset.keys()) {
        if(!deviceSet.contains(i))
            devStatusOffset.remove(i);
    }

    doReset();
}
