#ifndef DEVICEDATA
#define DEVICEDATA

#include "base/DeviceIndex.h"
#include "base/TaiTime.h"

#include <QElapsedTimer>
#include <QMap>
#include <QVariant>
#include <QVector>

/*
// ======= Data types =======
TAI

BaseEvent
    TAI

// --- TTVXS ---

ttvxs_trig_src_t


TtvxsEvent : public BaseEvent
    ttvxs_trig_src_t

// --- TDC ---

TdcEdge

TdcChHit

// --- ADC ---

AdcChHit

TqdcChHits
    QVector<TdcChHit>
    QVector<AdcChHit>

TqdcEvent : public BaseEvent
    QMap<int, TqdcChHits>

// --- DevStat ---

DevStat

CompleteEvent
    QMap<DeviceIndex, TqdcEvent> // for: TDC72-VXS TDC64-VHLE TQDC16-VS ADC64-VE
    QMap<DeviceIndex, TtvxsEvent> // for: TTVXS-RC
    QMap<DeviceIndex, DevStat>

DecodedData
    QMap<quint32, CompleteEvent>

// ======= Statistic types =======

AnalyzerChStat

AnalyzerDevStat
    QMap<int, AnalyzerChStat>

AnalyzerStat
    QMap<DeviceIndex, AnalyzerDevStat>

struct WorkerStat
*/

// ======= Data types =======

struct __attribute__((packed)) TaiRawData
{
    TaiTime toTaiTime() const
    {
        return {.sec = static_cast<int32_t>(sec), .ns = nSec, .valid = flags == 0x2};
    }

    quint32 sec;
    quint32 flags : 2;
    quint32 nSec : 30;
};

struct BaseEvent
{
    virtual ~BaseEvent() = default;
    virtual void clear()
    {
        index = {};
        evNum = 0;
        tai = {};
    }

    DeviceIndex index;
    quint32 evNum = 0;
    TaiTime tai;
};

// --- TTVXS ---

typedef union {
    unsigned int all;
    struct
    {
        unsigned int pulser : 1;        // no connection
        unsigned int random_pulser : 1; // link data error
        unsigned int external_ttl : 1;  // pci bus signaled error
        unsigned int digital_sfp : 1;   // error propagation
        unsigned int reserved : 28;     // no clock from remote
    } bit;
} ttvxs_trig_src_t;

struct TtvxsEvent : public BaseEvent
{
    ttvxs_trig_src_t trig_src;
};

// --- TDC ---

struct TdcEdge
{
    TdcEdge(quint16 ts, quint32 t, bool l) : time(t), tdcTs(ts), rising(l) {}
    void clear()
    {
        time = 0;
        tdcTs = 0;
        rising = true;
    }

    quint32 time = 0;
    quint16 tdcTs = 0;
    bool rising = true;
};

struct TdcChHit
{
    void clear()
    {
        ch = 0;
        tdcTs = 0;
        leadingTime = 0;
        trailingTime = 0;
        hasLeadTimeInl = false;
        hasTrailTimeInl = false;
        hasWidthTimeInl = false;
    }

    double getLeadTimeNs(const QVector<double> &inlCorr);
    double getTrailTimeNs(const QVector<double> &inlCorr);
    double getHitTime(const QVector<double> &inlCorr);
    double getWidthNs(const QVector<double> &inlCorr);

    quint8 ch = 0;
    quint16 tdcTs = 0;
    quint32 leadingTime = 0;  // in HPTDC ticks
    quint32 trailingTime = 0; // in HPTDC ticks
    double leadTimeInl = 0;   // in ns
    bool hasLeadTimeInl = false;
    double trailTimeInl = 0; // in ns
    bool hasTrailTimeInl = false;
    double widthTimeInl = 0; // in ns
    bool hasWidthTimeInl = false;
};

// --- ADC ---

struct AdcChHit
{
    void clear()
    {
        ch = 0;
        adcTs = 0;
        taiTime = {};
        mode = 0;
        wf.clear();
    }

    quint8 ch = 0;
    quint16 adcTs = 0; // 8ns intervals from event TS
    TaiTime taiTime;
    uint mode = 0; // 1-cal; 2-samp; 3-sumQ
    std::vector<double> wf;
    //    quint32 sumQ = 0; // 18-bit, sum of adc samples
};

struct TqdcChHits
{
    QVector<TdcChHit> tdcHits;
    QVector<AdcChHit> adcHits;
};

struct TqdcEvent : public BaseEvent
{
    void clear()
    {
        BaseEvent::clear();
        hits.clear();
        temp.clear();
        fw_rev.clear();
        bmc_rev.clear();
        errors.clear();
        chMask = 0;
    }

    QMap<int, TqdcChHits> hits; // per channel
    QMap<int, double> temp;     // per tempId
    QVariant fw_rev;
    QVariant bmc_rev;
    QVector<quint32> errors; // full 32-bit hptdc words (0x<HptdId><HptdcErr>)
    quint64 chMask = 0;
};
Q_DECLARE_METATYPE(TqdcEvent)

struct TrcEvent : public BaseEvent
{
    void clear()
    {
        BaseEvent::clear();
        wf.clear();
        chStateBefore = 0;
        chStateAfter = 0;
    }

    quint32 chStateBefore = 0;
    quint32 chStateAfter = 0;
    QVector<quint32> wf;
};

// --- DevStat ---

struct DevStat
{
    void clear()
    {
        index = {};
        regIoTimeout = false;
        regIoError = false;
        stats.clear();
    }

    DeviceIndex index;
    bool regIoTimeout = false;
    bool regIoError = false;
    QMap<quint16, quint16> stats; // stat from regs: <regAdr, regVal>
};

struct MscSliceData
{
    TaiTime tai; // TAI for start of slice
    quint8 ver = 0;
    quint8 bits = 0;
    quint32 sliceNs = 0;

    std::vector<std::pair<quint8, quint32>> ch; // [ch] = ch_hit_count
    quint8 ceState = 0;
    quint32 number = 0; // Slice number in package
};

struct MscData
{
    void clear() { slices.clear(); }
    std::vector<MscSliceData> slices;
};

struct MscTimeSliceData
{
    quint32 sliceNumber = 0;
    quint32 ceState = 0; // at the beginning of the slice
    quint32 cnt = 0; // ch hit counter
};

struct MscChData
{
    TaiTime tai; // TAI for start of slice
    quint8 ver = 0;
    quint8 ch = 0;

    quint32 sliceNs = 0; // Slice interval
    quint32 missedCnt = 0; // Missing hits during suspend before current packet

    std::vector<MscTimeSliceData> data; // [ch] = ch_hit_count
};

struct MscData_v1
{
    void clear() { chMap.clear(); }
    QMap<quint8, std::vector<MscChData>> chMap;
};

struct CompleteEvent
{
    void clear()
    {
        evNum = 0;
        tqdcEv.clear();
        trcEv.clear();
        ttvxsEv.clear();
        devStats.clear();
    }

    quint32 evNum = 0;
    QMap<DeviceIndex, TqdcEvent> tqdcEv;   // for: TDC72-VXS TDC64-VHLE TQDC16-VS ADC64-VE
    QMap<DeviceIndex, TrcEvent> trcEv;     // for: UT24VE-TRC
    QMap<DeviceIndex, TtvxsEvent> ttvxsEv; // for: TTVXS-RC
    QMap<DeviceIndex, DevStat> devStats;
};
Q_DECLARE_METATYPE(CompleteEvent)

struct DecodedData
{
    void clear()
    {
        events.clear();
        mscData0.clear();
        mscData1.clear();
        event_cnt = 0;
    }

    QMap<quint32, CompleteEvent> events;
    QMap<DeviceIndex, MscData> mscData0;
    QMap<DeviceIndex, MscData_v1> mscData1;
    quint32 event_cnt = 0;
};
Q_DECLARE_METATYPE(DecodedData)

#endif // DEVICEDATA
