#include "TQDCAnalyzer.h"

#include <iostream>

#include <QByteArray>
#include <QDateTime>
#include <QDebug>
#include <QJsonObject>
#include <QJsonDocument>
#include <QSemaphore>
#include <QTimer>

#include "base/DeviceId.h"
#include "base/RcRunId.h"
#include "base/bmc/fpga_mcu_regs_addr.h"
#include "device-modules/OneWire/OneWire.h"
#include "mregdev/MregDeviceAdcm.h"

#include "DeviceHelper.h"
#include "HistParams.h"
#include "HptdcError.h"
#include "MpdDataDecoder.h"

namespace {
const int STATISTIC_TIMEOUT = 500;
const int CLEAR_ERR_STR_STACK_TIMEOUT = 1000;
}

TQDCAnalyzer::TQDCAnalyzer(QSemaphore *_analyzerSem,
                           QSemaphore *_histSem,
                           QSemaphore *_writerSem, QSemaphore *_zmqSem,
                           QObject *parent)
    : QObject(parent)
    , analyzerSem(_analyzerSem)
    , histSem(_histSem)
    , writerSem(_writerSem)
    , zmqSem(_zmqSem)
    , statTimer(new QTimer(this))
    , decoder(new MpdDataDecoder(this))
{
    connect(statTimer, &QTimer::timeout, this, &TQDCAnalyzer::sendStat);
    statTimer->start(STATISTIC_TIMEOUT);
    decoder->setFlushFunc(std::bind(&TQDCAnalyzer::flushData, this));

    connect(decoder, &MpdDataDecoder::rawDataUpdated,
            this, &TQDCAnalyzer::rawDataUpdated);
    connect(decoder, &MpdDataDecoder::gotStopEvent,
            this, &TQDCAnalyzer::gotStopEvent);
    connect(decoder, &MpdDataDecoder::runStarted,
            this, &TQDCAnalyzer::runStarted);
    connect(decoder, &MpdDataDecoder::runStoped,
            this, &TQDCAnalyzer::runStoped);
    connect(decoder, &MpdDataDecoder::rawDataUpdated,
            this, &TQDCAnalyzer::rawDataUpdated);
}

void TQDCAnalyzer::dataAccuired(const QByteArray &newData)
{
    decoder->dataAccuired(newData);
    analyzerSem->release();
}

void TQDCAnalyzer::reset()
{
    decoder->reset();
    lastErrStr.clear();
}

void TQDCAnalyzer::sendStat()
{
    if(CLEAR_ERR_STR_STACK_TIMEOUT < STATISTIC_TIMEOUT * (++clear_err_stack)==0) {
        clear_err_stack = 0;
        lastErrStr.clear();
    }
    printBadEvent = true;
    auto analyzerStats = decoder->getAnalyzerStats();
    analyzerStats.timer.start();
    emit deviceStatsUpdated(analyzerStats);
}

void TQDCAnalyzer::startWrite(QString, bool rawData)
{
    decoder->startWrite(!rawData);
}

void TQDCAnalyzer::stopWrite()
{
    decoder->stopWrite();
}

void TQDCAnalyzer::requestRawData(int events)
{
    decoder->requestRawData(events);
}

void TQDCAnalyzer::setRawDataEveryEv(bool everyEv)
{
    decoder->setRawDataEveryEv(everyEv);
}

void TQDCAnalyzer::invertUpdated(QMap<DeviceIndex, QVector<bool> > _invertMap)
{
    decoder->invertUpdated(_invertMap);
}

void TQDCAnalyzer::set_can_skip(bool skip)
{
    if(skip)
        decoder->setSkipLogic(std::bind(skip_work, analyzerSem));
    else
        decoder->setSkipLogic(nullptr);
}

void TQDCAnalyzer::prepareForStop(quint32 evNum)
{
    decoder->prepareForStop(evNum);
}

void TQDCAnalyzer::flushData()
{
    auto decodedData = decoder->getDecodedData();
    if(decodedData.events.empty() && decodedData.mscData0.isEmpty() && decodedData.mscData1.isEmpty())
        return;

    while (!histSem->tryAcquire(1, 100)) {
        if(!noHistSpace.isValid())
            qInfo()<<"There is no free space for histogram queue";
        noHistSpace.restart();
    }
    if(writeDecodedEnabled) {
        Q_ASSERT(writerSem);
        while (!writerSem->tryAcquire(1, 100)) {
            if(!noWriterSpace.isValid())
                qInfo()<<"There is no free space for writer queue";
            noWriterSpace.restart();
        }
    }
    while (!zmqSem->tryAcquire(1, 100)) {
        if(!noZmqSpace.isValid())
            qInfo()<<"There is no free space for zmq queue";
        noZmqSpace.restart();
    }
    if(noHistSpace.isValid()) {
        if(noHistSpace.elapsed()>5000) {
            qInfo()<<"Restore free space for histogram queue";
            noHistSpace.invalidate();
        }
    }
    if(noWriterSpace.isValid()) {
        if(noWriterSpace.elapsed()>5000) {
            qInfo()<<"Restore free space for writer queue";
            noWriterSpace.invalidate();
        }
    }
    if(noZmqSpace.isValid()) {
        if(noZmqSpace.elapsed()>5000) {
            qInfo()<<"Restore free space for zmq queue";
            noZmqSpace.invalidate();
        }
    }

    emit deviceDataUpdated(decodedData, writeDecodedEnabled);
}
