/*
**    Copyright 2021 Ilia Slepnev, Ivan Filippov
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "BaseConfigDispatcher.h"

#include <optional>

#include "BaseDeviceAppConfig.h"
#include "ModularMregDevice.h"
#include "base/ConfigIndexName.h"
#include "base/DeviceIndexEnabled.h"
#include "base/Globals.h"
#include "base/util.h"

class BaseConfigDispatcherPrivate
{
public:
    std::optional<ConfigIndexName> cacheCIN;
};

BaseConfigDispatcher::BaseConfigDispatcher(QObject *parent)
    : AbstractDispatcher(parent),
    d_ptr(new BaseConfigDispatcherPrivate)
{
    qRegisterMetaType<ConfigKey>();
    qRegisterMetaType<ConfigIndexName>();
    qRegisterMetaType<BaseDeviceAppConfig>();
}

BaseConfigDispatcher::~BaseConfigDispatcher()
{
    delete d_ptr;
}

void BaseConfigDispatcher::configSelectorChanged(const ConfigIndexName &cin)
{
    Q_D(BaseConfigDispatcher);

    if (d->cacheCIN.has_value() && d->cacheCIN.value() == cin)
        return;
    d->cacheCIN = cin;
    sendCache();
}

void BaseConfigDispatcher::sendCache()
{
    Q_D(const BaseConfigDispatcher);
    if (!d->cacheCIN.has_value())
        return;
    emit switch_to_config(d->cacheCIN.value());
    ClientIndex ci{Globals::getDaqModule(), d->cacheCIN.value().getIndex()};
    emit clientIndexChanged(ci);
    sendConfigKeyCache();
    emit deviceListChanged(getDeviceIndexEnabledMap());
    BaseDeviceAppConfig &cache = getAppConfigCache();
    emit networkPortConfigChanged(cache.defaultSetup.networkPort);
    emit sparseParamsChanged(cache.defaultSetup.waveRec.sparse);
    emit allowRemoteControlChanged(cache.allowRemoteControl);
    emit zmqConfigChanged(cache.devStatusZmqPublisher);
    emit errOnTrigOnXOffChanged(cache.errOnTrigOnXOff);
    emit errOnSerDesUnlockChanged(cache.errOnSerDesUnlock);
    emit errOnFeLinkErrChanged(cache.errOnFeLinkErr, cache.warnOnFeLinkErr);
    sendAdcSerDesCache();
}

void BaseConfigDispatcher::incrementConfigKey(bool sendFullCache)
{
    ++config_key;
    if(sendFullCache)
        sendCache();
    else
        sendConfigKeyCache();
}

void BaseConfigDispatcher::sendConfigKeyCache()
{
    const BaseDeviceAppConfig &cache = getAppConfigCache();
    emit configChanged(config_key, cache);
    emit configKeyChanged(config_key);
    for (const auto &cfg : cache.knownConfigs)
        emit deviceConfigChanged(config_key, cfg);
//    emit calConfigChanged(config_key, *cacheCalConfig);
//    for (const auto &cal : *cacheCalConfig)
//        emit calResultChanged(config_key, cal);

    emit dlnpCalibrationTriggerConfigChanged(cache.defaultSetup.dlnpCalibrationTrigger);
    emit firFilterConfigChanged(cache.defaultSetup.firFilter);
    emit adcLemoMuxConfigChanged(cache.defaultSetup.adcLemoMux);
    emit clockControlConfigChanged(cache.defaultSetup.clockControl);
    emit crosspointSwitchConfigChanged(cache.defaultSetup.crosspointSwitch);
    emit networkPortConfigChanged(cache.defaultSetup.networkPort);
    emit preamplifierConfigChanged(cache.defaultSetup.ad5622);
    emit runLogicConfigChanged(cache.defaultSetup.runLogic);
    emit sparseParamsChanged(cache.defaultSetup.waveRec.sparse);
    emit trigCtrlStatConfigChanged(cache.defaultSetup.trigCtrlStat);
    emit ttlCfgChanged(cache.defaultSetup.ttlIo);
    emit waveBlcConfigChanged(cache.defaultSetup.waveBlc);
    emit waveRecConfigChanged(cache.defaultSetup.waveRec);
}

BaseDeviceAppConfig &BaseConfigDispatcher::getAppConfigCache()
{
    static BaseDeviceAppConfig tmp;
//    Q_ASSERT(false);
    return tmp;
}

//BaseDeviceAppConfig BaseConfigDispatcher::getAppConfigCache() const
//{
//    Q_ASSERT(false);
//    return {};
//}

void BaseConfigDispatcher::forwardDeviceConfig(const ModularMregDevice *obj)
{
    connect(this, &BaseConfigDispatcher::deviceConfigChanged,
            obj, &ModularMregDevice::setDeviceConfig,
            Qt::UniqueConnection);
    logConnection(obj, "DeviceConfig", "->");
    sendCache();
}

void BaseConfigDispatcher::setErrOnTrigOnXOff(bool isErr)
{
    BaseDeviceAppConfig &cache = getAppConfigCache();
    if (isErr == cache.errOnTrigOnXOff)
        return;

    cache.errOnTrigOnXOff = isErr;
//    incrementConfigKey(); //todo: should we increment config_key?
    sendConfigKeyCache();
    emit errOnTrigOnXOffChanged(cache.errOnTrigOnXOff);
}

void BaseConfigDispatcher::setErrOnSerDesUnlock(bool isErr)
{
    BaseDeviceAppConfig &cache = getAppConfigCache();
    if (isErr == cache.errOnSerDesUnlock)
        return;

    cache.errOnSerDesUnlock = isErr;
    //    incrementConfigKey(); //todo: should we increment config_key?
    sendConfigKeyCache();
    emit errOnSerDesUnlockChanged(cache.errOnSerDesUnlock);
}

void BaseConfigDispatcher::setErrOnFeLinkErr(bool isErr, bool isWarn)
{
    BaseDeviceAppConfig &cache = getAppConfigCache();
    if (isErr == cache.errOnFeLinkErr && isWarn == cache.warnOnFeLinkErr)
        return;

    cache.errOnFeLinkErr = isErr;
    cache.warnOnFeLinkErr = isWarn;
    sendConfigKeyCache();
    emit errOnFeLinkErrChanged(cache.errOnFeLinkErr, cache.warnOnFeLinkErr);
}

void BaseConfigDispatcher::setDeviceList(const DeviceIndexEnabledMap &map)
{
    BaseDeviceAppConfig &cache = getAppConfigCache();
    if((map == cache.getDeviceIndexEnabledMap()))
        return;

    for (auto map_it = map.cbegin(); map_it != map.cend(); ++map_it) {
        const DeviceIndex &index = map_it.key();
        const DeviceIndexEnabled &index_enabled = map_it.value();
        auto it = cache.knownConfigs.find(index);
        if (it == cache.knownConfigs.end()) {
            it = cache.knownConfigs.insert(index, cache.defaultSetup);
        }
        it->enabled = index_enabled.enabled;
        it->index = index;
    }

    auto &devices = cache.knownConfigs;
    for (auto it = devices.begin(); it != devices.end();) {
        if (!map.contains(it.key()))
            it = devices.erase(it);
        else
            ++it;
    }

    incrementConfigKey();
    Q_ASSERT(map == cache.getDeviceIndexEnabledMap());
    emit deviceListChanged(map);
}

void BaseConfigDispatcher::sendAdcSerDesCache()
{
    const BaseDeviceAppConfig &cache = getAppConfigCache();
    for (const auto &cfg : cache.knownConfigs)
        emit devSerDesDelayChanged(cfg.index, cfg.serDesDelay);
}

void BaseConfigDispatcher::setConfig(const BaseDeviceAppConfig &newConfig)
{
    if (sender())
        qDebug() << "app config changed by" << qObjectAncestryStr(sender());
    BaseDeviceAppConfig &cache = getAppConfigCache();
    if (newConfig == cache)
        return;

    cache = newConfig;
    incrementConfigKey(true);
}

void BaseConfigDispatcher::setDevSerDesDelay(const DeviceIndex &index, const AdcSerDesDelayMap &delay)
{
    BaseDeviceAppConfig &cache = getAppConfigCache();
    auto it = cache.knownConfigs.find(index);
    if(it == cache.knownConfigs.end())
        return;
    if(it.value().serDesDelay == delay)
        return;
    it.value().serDesDelay = delay;
    incrementConfigKey();
    emit devSerDesDelayChanged(index, delay);
}

void BaseConfigDispatcher::setAllowRemoteControl(bool allow)
{
    BaseDeviceAppConfig &cache = getAppConfigCache();
    if (allow == cache.allowRemoteControl)
        return;

    cache.allowRemoteControl = allow;
    incrementConfigKey();
    emit allowRemoteControlChanged(allow);
}

void BaseConfigDispatcher::setZmqConfig(const ZmqConfig &zmqCfg)
{
    BaseDeviceAppConfig &cache = getAppConfigCache();
    if (cache.devStatusZmqPublisher == zmqCfg)
        return;

    cache.devStatusZmqPublisher = zmqCfg;
    incrementConfigKey();
    emit zmqConfigChanged(zmqCfg);
}

void BaseConfigDispatcher::setDlnpCalibrationTriggerConfig(const DlnpCalibrationTriggerConfig &cfg)
{
    BaseDeviceAppConfig &cache = getAppConfigCache();
    if (cache.defaultSetup.dlnpCalibrationTrigger == cfg)
        return;

    cache.defaultSetup.dlnpCalibrationTrigger = cfg;
    cache.unify_config();
    incrementConfigKey();
    emit dlnpCalibrationTriggerConfigChanged(cache.defaultSetup.dlnpCalibrationTrigger);
}

void BaseConfigDispatcher::setFirFilterConfig(const FirFilterConfig &config)
{
    BaseDeviceAppConfig &cache = getAppConfigCache();
    if (config == cache.defaultSetup.firFilter)
        return;

    cache.defaultSetup.firFilter = config;
    cache.unify_config();
    incrementConfigKey();
    emit deviceListChanged(getDeviceIndexEnabledMap());
}

void BaseConfigDispatcher::setAdcLemoMuxConfig(const AdcLemoMuxCtrlConfig &config)
{
    BaseDeviceAppConfig &cache = getAppConfigCache();
    if (cache.defaultSetup.adcLemoMux == config)
        return;

    cache.defaultSetup.adcLemoMux = config;
    cache.unify_config();
    incrementConfigKey();
    emit deviceListChanged(getDeviceIndexEnabledMap());
}

void BaseConfigDispatcher::setClockControlConfig(const ClockControlConfig &config)
{
    BaseDeviceAppConfig &cache = getAppConfigCache();
    if (cache.defaultSetup.clockControl == config)
        return;

    cache.defaultSetup.clockControl = config;
    cache.unify_config();
    incrementConfigKey();
}

void BaseConfigDispatcher::setCrosspointSwitchConfig(const CrosspointSwitchConfig &config)
{
    BaseDeviceAppConfig &cache = getAppConfigCache();
    if (config == cache.defaultSetup.crosspointSwitch)
        return;

    cache.defaultSetup.crosspointSwitch = config;
    cache.unify_config();
    incrementConfigKey();
}

void BaseConfigDispatcher::setNetworkPortConfig(const NetworkPortConfig &config)
{
    BaseDeviceAppConfig &cache = getAppConfigCache();
    if (config == cache.defaultSetup.networkPort)
        return;

    cache.defaultSetup.networkPort = config;
    cache.unify_config();
    incrementConfigKey();
    emit deviceListChanged(getDeviceIndexEnabledMap());
}

void BaseConfigDispatcher::setSparseParams(const SparseParams &config)
{
    BaseDeviceAppConfig &cache = getAppConfigCache();
    if (config == cache.defaultSetup.waveRec.sparse &&
            config == cache.defaultSetup.adcErc.sparse)
        return;

    cache.defaultSetup.waveRec.sparse = config;
    cache.defaultSetup.adcErc.sparse = config;
    cache.unify_config();
    incrementConfigKey();
    emit deviceListChanged(getDeviceIndexEnabledMap());
}

void BaseConfigDispatcher::setPreamplifierConfig(const Ad5622Config &config)
{
    BaseDeviceAppConfig &cache = getAppConfigCache();
    if (config == cache.defaultSetup.ad5622)
        return;

    cache.defaultSetup.ad5622 = config;
    cache.unify_config();
    incrementConfigKey();
    emit deviceListChanged(getDeviceIndexEnabledMap());
}

void BaseConfigDispatcher::setRunLogicConfig(const RunLogicConfig &config)
{
    BaseDeviceAppConfig &cache = getAppConfigCache();
    if (config == cache.defaultSetup.runLogic)
        return;

    cache.defaultSetup.runLogic = config;
    cache.unify_config();
    incrementConfigKey();
    emit deviceListChanged(getDeviceIndexEnabledMap());
}

void BaseConfigDispatcher::setTrigCtrlStatConfig(const TrigCtrlStatConfig &config)
{
    BaseDeviceAppConfig &cache = getAppConfigCache();
    if (config == cache.defaultSetup.trigCtrlStat)
        return;

    cache.defaultSetup.trigCtrlStat = config;
    cache.unify_config();
    incrementConfigKey();
    emit deviceListChanged(getDeviceIndexEnabledMap());
}

void BaseConfigDispatcher::setTtlCfg(const TtlIoConfig &ttlCfg)
{
    BaseDeviceAppConfig &cache = getAppConfigCache();
    if (ttlCfg == cache.defaultSetup.ttlIo)
        return;

    cache.defaultSetup.ttlIo = ttlCfg;
    cache.unify_config();
    incrementConfigKey();
    emit deviceListChanged(getDeviceIndexEnabledMap());
}

void BaseConfigDispatcher::setWaveBlcConfig(const WaveBlcConfig &config)
{
    BaseDeviceAppConfig &cache = getAppConfigCache();
    if (config == cache.defaultSetup.waveBlc)
        return;

    cache.defaultSetup.waveBlc = config;
    cache.unify_config();
    incrementConfigKey();
    emit deviceListChanged(getDeviceIndexEnabledMap());
}

void BaseConfigDispatcher::setWaveRecConfig(const WaveRecConfig &config)
{
    BaseDeviceAppConfig &cache = getAppConfigCache();
    if (config == cache.defaultSetup.waveRec)
        return;

    cache.defaultSetup.waveRec = config;
    cache.unify_config();
    incrementConfigKey();
    emit deviceListChanged(getDeviceIndexEnabledMap());
}
