/*
**    Copyright 2015-2022 AFI Electronics
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef MPDRAWTYPES
#define MPDRAWTYPES

#include <QByteArray>
#include <QMetaType>

#include <optional>

#include "base/TaiTime.h"

class DeviceIndex;
class QByteArray;

// === mstream-EventBuilder formats ===

enum MpdRawDataSync {
    MPD_SYNC_NORM = 0x2A502A50, // [deprecated] TLV mixed with evNUm
    MPD_SYNC_EOB  = 0x4A624A62, // [deprecated] TLV mixed with evNUm
    MPD_SYNC_RUN_START = 0x72617453,
    MPD_SYNC_RUN_STOP  = 0x706F7453,
    MPD_SYNC_FILE_BEGIN = 0x67654246, // FBeg
    MPD_SYNC_FILE_END   = 0x646E4546, // FEnd
    MPD_SYNC_NORM_TLV = 0x2A50D5AF, // type in format TLV
    MPD_SYNC_EOB_TLV  = 0x4A62B59D, // type in format TLV
    MPD_SYNC_JSON_TLV = 0x4E4F534A
};

enum MpdRawDataRunSync {
    MPD_SYNC_RUN_NUMBER = 0x236E7552,
    MPD_SYNC_RUN_INDEX = 0x78646E49
};

enum MpdRawDataOtherSync {
    MPD_SYNC_FILE_ID = 0x64496946, // FiId
    MPD_SYNC_EV_ORDER = 0x71655345 // ESeq
};

#pragma pack(push, 1)

struct __attribute__ ((packed)) MpdHdr
{
    MpdHdr()
        : sync(0), length(0) {}
    static bool knownSync(quint32 d);
    static QString getSyncName(const quint32 &s);
    QString getSyncName() const { return getSyncName(sync); }
    bool isValid() const;
    quint32 sync;
    quint32 length; // payload size in bytes
};

struct __attribute__ ((packed)) MpdRunNumberHdr
{
    MpdRunNumberHdr()
        : sync(), length(0) {}
    bool isValid() const;
    quint32 sync;
    quint32 length; // payload size in bytes
    quint32 runNumber;
};

struct __attribute__ ((packed)) MpdEventHdr
{
    MpdEventHdr()
        : sync(MPD_SYNC_NORM), length(0), evNum(0) {}
    bool isValid() const;
    quint32 sync;
    quint32 length; // payload size in bytes
    quint32 evNum;
};

struct __attribute__ ((packed)) MpdDeviceHdr
{
//    MpdDeviceHdr()
//        : deviceSerial(0), length(0), deviceId(0) {}
    unsigned int deviceSerial:32;
    unsigned int length:24; // payload size in bytes
    unsigned char  deviceId:8;
};

struct __attribute__ ((packed)) MpdMStreamHdr
{
    MpdMStreamHdr()
        : subtype(0), words32b(0), usrDefBits(0) {}
    quint8  subtype:2;
    quint32 words32b:22; // payload size in 32-bit words
    quint8  usrDefBits; // Adc64-channal, HRB-exactTrigPos
};


struct __attribute__ ((packed)) AdcDataType0 : public TaiTime
{
    AdcDataType0()
        : chLower(0), chUpper(0) {}
    quint32 chLower; // readout channels 31:0 bit mask
    quint32 chUpper; // readout channels 63:32 bit mask
};

struct __attribute__ ((packed)) AdcDataType1 : public TaiTime
{
};

// Use in TTVXS TDC72VXS TDC16-VSE TDC64-VHLE
struct __attribute__ ((packed)) DataBlockHeader
{
    DataBlockHeader()
        : len(0), specific({0, 0, 0}) {}
    inline quint8 type() const {return specific.adc.type; }

    quint16 len;
    union {
        struct {
            quint8 reserved:4;
            quint8 ch:8;
            quint8 type:4;
        } phose_adc;
        struct {
            quint8 reserved:8;
            quint8 ch:4;
            quint8 type:4;
        } adc;
        struct {
            quint8 fifio_overflow:1;
            quint16 reserved:11;
            quint8 type:4;
        } tdc;
        struct {
            quint8 regio_timeout:1;
            quint8 regio_error:1;
            quint16 reserved:10;
            quint8 type:4;
        } stat;
    } specific;
};

#pragma pack(pop)

struct MpdRawDataDescriptor
{
    MpdRawDataDescriptor() = default;
    int size() const { return sizeof(eventHdr) + sizeof(deviceHdr) + payloadSize(); }
    int payloadSize() const { return deviceHdr.length; }
    bool isValid() const;
    void setDeviceIndex(const DeviceIndex &i);
    void setData(const QByteArray &d);
    void setData(const char *d, int size);

    MpdEventHdr eventHdr = {};
    MpdDeviceHdr deviceHdr = {};
    QByteArray packet;
};
Q_DECLARE_METATYPE(MpdRawDataDescriptor)

struct MpdRawDataDevRecord
{
    MpdRawDataDevRecord() = default;
    int size() const { return data.size(); }
    std::optional<quint32> getSync() const;
//    bool isValid() const;

//    template <typename T>
//    void setHeader(const T &h);

    template <typename T>
    void addData(T d, int size)
    {
        data.append(static_cast<const char *>(static_cast<const void *>(d)),
                    size);
    }

    void clear();

    std::optional<MpdHdr> mpdHdr;
    std::optional<MpdEventHdr> eventHdr;
    std::optional<MpdDeviceHdr> deviceHdr;
    QByteArray data;
};
Q_DECLARE_METATYPE(MpdRawDataDevRecord)

#endif // MPDRAWTYPES
