#include "MstreamMregDevGlobal.h"

#include "DevStreamDecoder.h"

#include "ModularDeviceStatus.h"
#include "ModularMregDevice.h"
#include "DeviceStateController.h"
#include "base/ThreadController.h"
#include "lib-common/BaseConfigDispatcher.h"
#include "mldiscover/LldpDiscovery.h"
#include "mstream-lib/MStreamDump.h"
#include "histogramm/MpdDataDecoder.h"

MstreamMregDevGlobal::MstreamMregDevGlobal(ModularMregDevice * const _dev, ModularDeviceStatus &_status, QObject *parent)
    : QObject(parent)
    , index(_dev->getDeviceIndex())
    , dev(_dev)
    , status(_status)
    , deviceStateController(new DeviceStateController(dev, this))
{
    new ThreadController(dev, this);

    connect(dev, &QtMregDevice::deviceConnected,
            this, &MstreamMregDevGlobal::deviceConnected);
    connect(dev, &QtMregDevice::deviceDisconnected,
            this, &MstreamMregDevGlobal::deviceDisconnected);
    connect(deviceStateController, &DeviceStateController::deviceStateChanged,
            this, &MstreamMregDevGlobal::deviceStateChanged);
    connect(deviceStateController, &DeviceStateController::deviceStateChanged,
            this, &MstreamMregDevGlobal::onDeviceStateChange);

    connect(this, &MstreamMregDevGlobal::mstreamConnectionStateChanged,
            this, &MstreamMregDevGlobal::onMstreamConnectionStateChanged);

    connect(dev, &ModularMregDevice::deviceConfigWritten,
            this, &MstreamMregDevGlobal::configWritten);
}

MstreamMregDevGlobal::~MstreamMregDevGlobal()
{
    setDeleteFlag();
}

void MstreamMregDevGlobal::setDeleteFlag() const
{
    if (dev)
        dev->setDeleteFlag();
    if (mstream)
        mstream->setDeleteFlag();
}

bool MstreamMregDevGlobal::isDeviceConnected() const
{
    return deviceStateController->isConnected() && mstreamConnectionState;
}

void MstreamMregDevGlobal::setCreateDataDecoder(bool v)
{
    createDataDecoder = v;

    if(createDataDecoder) {
        addDataDecoder();
    } else {
        if(dataDecoder) {
            dataDecoder->deleteLater();
        }
        dataDecoder = nullptr;
    }
}

void MstreamMregDevGlobal::onMstreamConnectionStateChanged(const DeviceIndex &deviceIndex, bool state)
{
    mstreamConnectionState = state;
}

void MstreamMregDevGlobal::onDeviceStateChange(const DeviceIndex &deviceIndex, bool state)
{
    if (deviceIndex != index) // not my device
        return;
    if (state) {
        if (!mstream)
            addMStreamDump();
    } else {
        if (mstream){
            mstream->setDeleteFlag();
            mstream->deleteLater();
        }
        mstream = nullptr;
        if(dataDecoder) {
            dataDecoder->deleteLater();
        }
        dataDecoder = nullptr;
    }
}

void MstreamMregDevGlobal::setConfigDispatcher(BaseConfigDispatcher *cd)
{
    configDispatcher = cd;
}

void MstreamMregDevGlobal::resetMStream(int seqNum)
{
    emit msdump_reset(getDeviceIndex(), seqNum);
}

void MstreamMregDevGlobal::setDataDecodeEnabled(bool en) const
{
    if(!dataDecoder || !mstream)
        return;

    if(en) {
        connect(mstream, &MStreamDump::lastRecordUpdated,
                dataDecoder, &MpdDataDecoder::dataAccuired, Qt::UniqueConnection);
    } else {
        disconnect(mstream, &MStreamDump::lastRecordUpdated,
                dataDecoder, &MpdDataDecoder::dataAccuired);
    }
}

void MstreamMregDevGlobal::addMStreamDump()
{
    MStreamDumpParameters dump_parameters;
    dump_parameters.deviceIndex = index;
    dump_parameters.preRunInit();

    mstream = new MStreamDump(dump_parameters, false);

    connect(dev, &QtMregDevice::deviceConnected,
            mstream, &MStreamDump::connectToDevice);
    connect(dev, &QtMregDevice::deviceDisconnected,
            mstream, &MStreamDump::disconnectFromDevice);

    connect(mstream, &MStreamDump::closeProgram,
            this, &MstreamMregDevGlobal::closeProgram);
    connect(mstream, &MStreamDump::updateProgramDescription,
            this, &MstreamMregDevGlobal::updateProgramDescription);

    connect(mstream, &MStreamDump::connectionStateChanged,
            this, &MstreamMregDevGlobal::mstreamConnectionStateChanged);
    connect(mstream, &MStreamDump::stat_updated,
            this, &MstreamMregDevGlobal::stat_updated);
    connect(mstream, &MStreamDump::stat_from_receiver_updated,
            this, &MstreamMregDevGlobal::stat_from_receiver_updated);

    LldpDiscovery::subscribe(mstream, {index.getDevId()});

    connect(this, &MstreamMregDevGlobal::msdump_reset, mstream, &MStreamDump::msdump_reset);
    connect(this, &MstreamMregDevGlobal::preRunCheck, mstream, &MStreamDump::preRunCheck);
    connect(mstream, &MStreamDump::reseted, this, &MstreamMregDevGlobal::msdumpReseted);

    mstream->setObjectName("MSD_"+index.getSerialStr());
    new ThreadController(mstream, this);

    if(configDispatcher) {
        connect(configDispatcher, &BaseConfigDispatcher::switch_to_config,
                mstream, &MStreamDump::configSelectorChanged);
        configDispatcher->sendCache();
    }

    if(createDataDecoder)
        addDataDecoder();

    postAddMStreamDump(mstream);
}

void MstreamMregDevGlobal::addDataDecoder()
{
    if(!mstream)
        return;

    if(!dataDecoder) {
        dataDecoder = new DevStreamDecoder(this);
        new ThreadController(dataDecoder, this);
    }

    if(dataDecoder) {
        connect(mstream, &MStreamDump::lastRecordUpdated,
                dataDecoder, &MpdDataDecoder::dataAccuired, Qt::UniqueConnection);
        connect(dataDecoder, &DevStreamDecoder::deviceDataUpdated,
                this, &MstreamMregDevGlobal::deviceDataUpdated);
    }
}

