/*
**    Copyright 2021 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "ZmqKvSubscriber.h"

#include <QDebug>
#include <QEvent>
#include <QLoggingCategory>
#include <QThread>
#include <QTimer>

#include <zmq_addon.hpp>

#include "base/ServiceConfigUrl.h"
#include "ZmqContext.h"
// #include "ZmqMonitor.h"

namespace {
Q_DECLARE_LOGGING_CATEGORY(lc_zmq)
Q_LOGGING_CATEGORY(lc_zmq, "afi.zmq")
static const char *DEFAULT_URL = "tcp://0.0.0.0:5678";
static const bool DEFAULT_ENABLE = true;
static const int DEFERRED_INIT_TIMEOUT = 0;
}

ZmqKvSubscriber::ZmqKvSubscriber(const QString &serviceName, QObject *parent)
    : AbstractSubscriber(parent),
    context(ZmqContext::instance())
{
    setObjectName("ZmqKvSubscriber");
    ServiceConfigUrl u(serviceName, QUrl::fromUserInput(DEFAULT_URL), DEFAULT_ENABLE);

    enabled = u.isEnabled();
    url = u.getUrl();
    if (url.isEmpty())
        enabled = false;
    if (!enabled)
        return;

    // thread may change after ctor, use QueuedConnection
    auto t = new QTimer(this);
    t->setSingleShot(true);
    t->start(DEFERRED_INIT_TIMEOUT);
    connect(t, &QTimer::timeout, this, &ZmqKvSubscriber::deferredInit, Qt::QueuedConnection);
}

ZmqKvSubscriber::~ZmqKvSubscriber()
{
    deleteSocket();
}

void ZmqKvSubscriber::createSocket()
{
    deleteSocket();
    sock = nullptr;
    try {
        sock = new zmq::socket_t(context, ZMQ_SUB);
        // new ZmqMonitor(*sock, addr, this);
        QTimer::singleShot(0, this, &ZmqKvSubscriber::pollEvents);
    }  catch (std::exception &e) {
        qCritical(lc_zmq).nospace() << __func__ << ": " << e.what();
    }
}

void ZmqKvSubscriber::deleteSocket()
{
    if (!sock)
        return;
    delete sock;
    sock = nullptr;
}

bool ZmqKvSubscriber::event(QEvent* ev)
{
    if (ev->type() == QEvent::ThreadChange) {
        if (sock) {
            deleteSocket();
            QTimer::singleShot(0, this, &ZmqKvSubscriber::deferredInit);
        }
        return true;
    }
    return AbstractSubscriber::event(ev);
}

void ZmqKvSubscriber::deferredInit()
{
    createSocket();
    Q_ASSERT(sock);
    if (!sock)
        return;
    auto addr = url.toString().toLatin1();
    qDebug(lc_zmq).noquote().nospace() << "listen '" << addr << "'";
    try {
        sock->bind(addr);
        restoreSubscription();
    }  catch (std::exception &e) {
        qCritical(lc_zmq).nospace() << __func__ << ": " << e.what();
    }
}

void ZmqKvSubscriber::subscribe(const QString &topic)
{
    if (!sock) {
        qDebug(lc_zmq).nospace().noquote() << "deferred subscription to '" << topic << "'";
        topics.insert(topic);
        return;
    }
    try {
        sock->set(zmq::sockopt::subscribe, topic.toLatin1().data());
    }  catch (std::exception &e) {
        qCritical(lc_zmq).nospace() << __func__ << ": " <<  e.what();
    }
    qDebug(lc_zmq).nospace().noquote() << "subscribed to '" << topic << "'";
}

void ZmqKvSubscriber::restoreSubscription()
{
    for (const auto &s: qAsConst(topics))
        subscribe(s);
}

void ZmqKvSubscriber::decodeMultipart(const std::vector<zmq::message_t> &msgs)
{
    const size_t nparts = 2;
    if (msgs.size() == nparts) {
        auto topic = QString::fromLatin1(msgs[0].data<char>(), msgs[0].size());
        auto value = QByteArray(msgs[1].data<char>(), msgs[1].size());
//        Q_ASSERT(value.startsWith('{'));
//        Q_ASSERT(value.endsWith('}'));
        emit messageReceived(topic, value);
    } else {
        qCritical(lc_zmq).nospace() << __func__ << ": unknown message, expected " << nparts << " parts, got " << msgs.size();
    }
}

void ZmqKvSubscriber::pollEvents()
{
    if (!sock)
        return;
    zmq_pollitem_t items[1] =
        {{.socket = *sock, .fd = 0, .events = ZMQ_POLLIN, .revents = 0}};
    long timeout = 1; // TODO

    int n = zmq::detail::poll(items, sizeof(items)/sizeof(items[0]), timeout);
    if (n > 0) {
        std::vector<zmq::message_t> msgs;
        zmq::recv_multipart(*sock, std::back_inserter(msgs));
        decodeMultipart(msgs);
    }
    QTimer::singleShot(0, this, &ZmqKvSubscriber::pollEvents);
}
