/*
**    Copyright 2010-2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "mregdev/qtmregdevice.h"

#include <QLoggingCategory>
#include <QTimer>

#include "base/DeviceId.h"
#include "base/fw_version.h"
#include "mregdev/MregDevice.h"
#include "regio/mlinkdevice.h"
#include "regio/regiomlink.h"
#include "regio/regop.h"
#include "device-modules/MlinkCsr/MlinkCsrModule.h"
#include "sdb/SdbTypes.h"

using std::vector;

namespace {
Q_DECLARE_LOGGING_CATEGORY(lc_qtmregdevice)
Q_LOGGING_CATEGORY(lc_qtmregdevice, "afi.qtmregdevice")
}

static const int initMaxLinkRetryCount = 5;
static const int connectedMaxLinkRetryCount = 12;
const int detectRetryLimit = 3;
const int fsmPollTimeoutMs = 100; // milliseconds
const int onlineStatusPollTimeoutMs = 100; // milliseconds
const int offlineStatusPollTimeoutMs = 5000; // milliseconds

QtMregDevice::QtMregDevice(const DeviceIndex &index, QObject *parent) :
    QObject(parent),
    link(new mlink::RegIOMLink(this)),
    deviceIndex(index),
    mregDevice(new mlink::MregDevice(*link, 0xFEFE, this)),
    timerFsmPoll(new QTimer(this)),
    timerStatusPoll(new QTimer(this)),
    liveMagic(nextLiveMagic())
{
    setObjectName(index.getIdent());

    connect(mregDevice, &mlink::MlinkDevice::protocolMessage,
            this, &QtMregDevice::onProtocolMessage);
    connect(mregDevice, &mlink::MlinkDevice::protocolTimeout,
            this, &QtMregDevice::onProtocolTimeout);

    connect(timerFsmPoll, &QTimer::timeout, this, &QtMregDevice::fsmPoll);
    timerFsmPoll->start(fsmPollTimeoutMs);
    connect(&linkFSM, &MlinkStateFsm::stateChanged, this, &QtMregDevice::on_linkStateChanged);

    connect(timerStatusPoll, &QTimer::timeout, this, &QtMregDevice::statusPoll);
    timerStatusPoll->setSingleShot(true);
    timerStatusPoll->start(offlineStatusPollTimeoutMs);

    connect(this, &QtMregDevice::deviceConnected, this, &QtMregDevice::on_deviceConnected);
    connect(this, &QtMregDevice::deviceDisconnected, this, &QtMregDevice::on_deviceDisconnected);
}

QtMregDevice::~QtMregDevice()
{
    forceDisconnect();
    mlinkCsrModule.clear();
    delete mregDevice;
}

void QtMregDevice::setDeleteFlag()
{
    deleteFlag.store(true, std::memory_order_relaxed);
    if (link)
        link->setDeleteFlag();
}

bool QtMregDevice::hasDeleteFlag() const
{
    return deleteFlag.load(std::memory_order_relaxed);
}

sdb_t QtMregDevice::getSdb() const
{
    return mregDevice->getSdb();
}

DeviceId QtMregDevice::getDeviceId() const
{
    return deviceIndex.getDevId();
}

fw_version_t QtMregDevice::getDeviceFwVersion() const
{
    if (mlinkCsrModule)
        return mlinkCsrModule->getStatus().fw_version;
    return {};
}

fw_version_t QtMregDevice::getDeviceHwVersion() const
{
    if (mlinkCsrModule)
        return mlinkCsrModule->getStatus().hw_version;
    return {};
}

quint64 QtMregDevice::getDeviceSerialNumber() const
{
    return deviceIndex.getSerial();
}

QString QtMregDevice::getFirmwareVersionStr() const
{
    auto fw_ver = getDeviceFwVersion();
    if (fw_ver.isNull())
        return {};
    return fw_ver.toString();
}

QHostAddress QtMregDevice::getAddress() const
{
    return link->getHostAddress();
}

QString QtMregDevice::getAddressStr() const
{
    return getAddress().toString();
}

QString QtMregDevice::getIdent() const
{
    return getDeviceIndex().getIdent();
}

QMap<QString, double> QtMregDevice::getFpgaTemps() const
{
    if (!mlinkCsrModule)
        return {};
    if(mlinkCsrModule->getStatus().temps.isEmpty())
        mlinkCsrModule->readStatus(); // perorm read temp from reg.
    return mlinkCsrModule->getStatus().temps;
}

void QtMregDevice::setHostAddress(const DeviceIndex &index, const QHostAddress &address, int ttl)
{
    if (deviceIndex != index)
        return;
    bool addressChanged = address != link->getHostAddress();
    link->setHostAddress(address, ttl);
    if (isConnected() && addressChanged) {
        forceDisconnect();
    }
    fsmPoll();
}

QSharedPointer<MlinkCsrModule> QtMregDevice::createMlinkCsrModule()
{
    const QList<AFI_SDB_ID> &ids = MlinkCsrModule::getSupportedSdbIds();
    const QList<SdbDevice> &records = getSdb().filterDevices(ids);
    for (auto it = records.begin(); it != records.end(); ++it) {
        const auto &sdbDevice = *it;
        MregBusInterface bus(this, sdbDevice);
        auto module = QSharedPointer<MlinkCsrModule>(new MlinkCsrModule(bus));
        if (module)
            module->post_init();
        return module;
    }
    return {};
}

bool QtMregDevice::detect_mlink_csr()
{
    mlinkCsrModule.clear();
    auto found = mregDevice->detect();
    if (!found)
        return {};
    mlinkCsrModule = createMlinkCsrModule();
    if (!mlinkCsrModule)
        return {};

    if (!mlinkCsrModule->detect())
        return {};
    const auto &status = mlinkCsrModule->getStatus();
    const auto deviceId = static_cast<DeviceId>(status.device_id);
    const auto detectedIndex = DeviceIndex(deviceId, status.device_serial);
    if (detectedIndex != deviceIndex) {
        qWarning(lc_qtmregdevice).noquote() <<
            QString("[%1] MlinkCsr: unexpected device %2 at %3, disabling")
                .arg(getIdent())
                .arg(detectedIndex.getIdent())
                .arg(getAddressStr());
        setHostAddress(detectedIndex, {}, {});
        return {};
    } else {
        qDebug(lc_qtmregdevice).noquote()
            << QString("[%1] %2 MlinkCsr Ok").arg(getIdent()).arg(getAddressStr());
    }
    return true;
}

bool QtMregDevice::detect()
{
    Q_ASSERT(linkFSM.state() == MlinkStateFsm::Closed);

    if (link->getHostAddress().isNull()) return false;

    if (!detect_mlink_csr()) {
        return false;
    }
    return true;
};

quint16 QtMregDevice::nextLiveMagic() const
{
    quint16 tryMagic = liveMagic+1;
    if (tryMagic < 100)
        tryMagic = 100 + qrand() % 30000;
    return tryMagic;
}

bool QtMregDevice::writeLiveMagic()
{
    if (!enableLiveMagicCheck)
        return true;
    quint16 tryMagic = nextLiveMagic();
    quint16 wroteMagic = 0;
    bool ok = mlinkCsrModule->writeLiveMagic(tryMagic, &wroteMagic);
    ok &= (wroteMagic == tryMagic);
    if (!ok) {
        qWarning(lc_qtmregdevice).noquote()
            << QString("LiveMagic write error");
    }
    if (ok)
        liveMagic = wroteMagic;
    return ok;
}

bool QtMregDevice::checkLiveMagic()
{
    if (!enableLiveMagicCheck)
        return true;

    quint16 rd;
    if (!mlinkCsrModule->readLiveMagic(&rd))
        return false;
    bool match = (rd == liveMagic);
    if (!match) {
        qWarning(lc_qtmregdevice).noquote()
            << "liveMagic changed, disconnecting";
        emit liveMagicFailed(getDeviceIndex());
        forceDisconnect();
    }
    return match;
}

bool QtMregDevice::isConnected() const
{
    return linkFSM.state() == MlinkStateFsm::Connected;
}

bool QtMregDevice::isFailed() const
{
    return linkFSM.state() == MlinkStateFsm::Failed;
}

void QtMregDevice::statusPoll()
{
    if (!isConnected())
        return;
    if (!checkLiveMagic())
        return;

    readStatus();

    // restart timer
    if (isConnected())
        timerStatusPoll->start(onlineStatusPollTimeoutMs);
}

void QtMregDevice::onProtocolMessage(const QString &errorStr)
{
    if (linkFailureCount > 0)
        return; // mute repeated messaged while retrying
    qWarning(lc_qtmregdevice).noquote() <<
        QString("[%1] %2 %3").arg(getIdent()).arg(getAddressStr()).arg(errorStr);
}

void QtMregDevice::onProtocolTimeout()
{
    linkFSM.changeState(MlinkStateFsm::Failed);
}

void QtMregDevice::on_linkStateChanged(MlinkStateFsm::State oldState, MlinkStateFsm::State newState)
{
    emit deviceStateChanged(getDeviceIndex(), newState);
    //    qDebug(lc_qtmregdevice).noquote()
    //        << QString("[%1] link state changed %2 to %3")
    //               .arg(getIdent())
    //               .arg(MlinkStateFsm::stateStr(oldState))
    //               .arg(MlinkStateFsm::stateStr(newState));
    Q_ASSERT(oldState != newState);
    Q_ASSERT(linkFSM.state() == newState);

    switch (linkFSM.state()) {
    case MlinkStateFsm::Closed:
        detectRetryCount = 0;
        mregDevice->setMaxRetryCount(initMaxLinkRetryCount);
        break;
    case MlinkStateFsm::Detected:
        break;
    case MlinkStateFsm::Connected:
        break;
    case MlinkStateFsm::Failed:
        //        if (!linkFailureCount) {
        //            qWarning(lc_qtmregdevice).noquote() <<
        //                QString("[%1] %2 connection failed").arg(getIdent()).arg(getAddressStr());
        //        }
        linkFailureCount++;
        break;
    }

    if (newState == MlinkStateFsm::Connected) {
        qInfo(lc_qtmregdevice).noquote()
            << QString("[%1] %2 connected").arg(getIdent()).arg(getAddressStr());
        linkFailureCount = 0;
        mregDevice->setMaxRetryCount(connectedMaxLinkRetryCount);
        Q_EMIT deviceConnected(deviceIndex);
        timerStatusPoll->start(0);
    }

    if (oldState == MlinkStateFsm::Connected) {
        qInfo(lc_qtmregdevice).noquote()
            << QString("[%1] %2 disconnected").arg(getIdent()).arg(getAddressStr());
        timerStatusPoll->stop();
        Q_EMIT deviceDisconnected(deviceIndex);
    }
}

void QtMregDevice::fsmPoll()
{
    switch (linkFSM.state()) {
    case MlinkStateFsm::Closed:
        if (!link->getHostAddress().isNull()) {
            bool ok = detect(); // auto-connect
            if (ok) {
                linkFSM.changeState(MlinkStateFsm::Detected);
            } else {
                detectRetryCount++;
                if (detectRetryCount > detectRetryLimit)
                    linkFSM.changeState(MlinkStateFsm::Failed);
            }
        }
        break;
    case MlinkStateFsm::Detected:
        if (writeLiveMagic())
            linkFSM.changeState(MlinkStateFsm::Connected);
        else
            linkFSM.changeState(MlinkStateFsm::Failed);
        break;
    case MlinkStateFsm::Connected:
        break;
    case MlinkStateFsm::Failed:
        if (linkFSM.stateTimeMs() > failureRecoveryTimeoutMs())
            linkFSM.changeState(MlinkStateFsm::Closed);
        break;
    }
}

qint64 QtMregDevice::failureRecoveryTimeoutMs() const
{
    // progressive recovery: 0.1 s to 6.4s
    double baseTimeoutMs = 100;
    int factor = std::clamp(linkFailureCount, 1, 6);
    return baseTimeoutMs*pow(2, factor);
}

void QtMregDevice::forceDisconnect()
{
    if (linkFSM.state() != MlinkStateFsm::Connected)
        return;

    // TODO: release device, gracefully close connection
    linkFSM.changeState(MlinkStateFsm::Closed);
}

bool QtMregDevice::regOpExec(mlink::RegOpVector &data)
{
    if (isFailed())
        return {};
    return mregDevice->regOpExec(data);
}

bool QtMregDevice::memReadBlk(int startWord, size_t wordCount, std::vector<quint32> *result)
{
    if (isFailed())
        return {};
    return mregDevice->memReadBlk(startWord, wordCount, result);
}

bool QtMregDevice::memWriteBlk(int wordNumber, const std::vector<quint32> &data)
{
    if (isFailed())
        return {};
    return mregDevice->memWriteBlk(wordNumber, data);
}
