/*
**    Copyright 2010-2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/


#include "mlinkdevice.h"

#include <cassert>
#include <iomanip>
#include <sstream>
#include <sstream>
#include <stdexcept>

#include "mlink/frame_types.h"
#include "mlinkpacketprotocol.h"
#include "regio/regiobase.h"
#include "regiopacket.h"
#include "regop.h"
#include "RegIo16Codec.h"

namespace mlink {

typedef std::runtime_error exc_dev;

const int MlinkHostAddress = 1;

MlinkDevice::MlinkDevice(RegIOBase &p, int address, QObject *parent)
    : QObject(parent)
    , link(p)
    , myAddress(address)
{}

bool MlinkDevice::regOpExec(RegOpVector &v)
{
    RegIoPacket tx, rx;
    tx.src = MlinkHostAddress;
    tx.dst = myAddress;
    tx.type = FRAME_TYPE_CTRL_REQ;
    tx.data.clear();
    const int reserve = 4;

    auto chunkBegin = v.begin();
    for (auto it = v.begin(); it != v.end(); ++it) {
        const auto &op = *it;
        RegIo16Codec::encode_regio16_op(op, &tx.data);
        if (tx.data.size() + reserve >= (int)getMaxRegRequestSize()) {
            if (!ctrlExchangeSingle(tx, &rx))
                return {};
            auto chunkEnd = it; chunkEnd++;
            if (!RegIo16Codec::decode(rx.data, chunkBegin, chunkEnd))
                return {};
            tx.data.clear();
            chunkBegin = chunkEnd;
        }
    }
    if (!tx.data.empty()) {
        if (!ctrlExchangeSingle(tx, &rx))
            return {};
        auto chunkEnd = v.end();
        if (!RegIo16Codec::decode(rx.data, chunkBegin, chunkEnd))
            return {};
        tx.data.clear();
    }
    return true;
}

// read wordCount 32-bit words from RAM starting from startWord offset
bool MlinkDevice::memReadBlk(int startWord, size_t wordCount, std::vector<quint32> *result)
{
    std::vector<quint32> tmp;
    tmp.reserve(wordCount);
    RegIoPacket tx, rx;
    tx.src = MlinkHostAddress;
    tx.dst = myAddress;
    tx.type = FRAME_TYPE_RAM_REQ;
    //    const size_t txSz = 1;
    //    tx.data.resize(txSz);
    const size_t rxHdrSz = 1;

    size_t totalRead = 0;
    while(totalRead < wordCount) {
        size_t blockSize = std::min(wordCount - totalRead, getMaxRamRequestSize());
        tx.data.resize(1+blockSize); // FIXME: this is required for HWIP
        tx.data[0] = ( 0x80000000 |((blockSize&0x1FF)<<22)| ((startWord+totalRead)&0x3FFFFF) );
        const size_t rxSz = rxHdrSz+blockSize;

        bool ok = ctrlExchangeSingle(tx, &rx);
        if (!ok || rx.data.size() != (int)rxSz) {
            auto s = QString("Failed mem read[%1]: expected %2, received %3 words")
                         .arg(startWord, 6, 16, QChar('0')).arg(rx.data.size()).arg(rxSz);
            emit protocolMessage(s);
            return {};
        }
        for(size_t i=0; i<blockSize; i++)
            tmp.push_back(rx.data[rxHdrSz+i]);
        totalRead += blockSize;
    }
    if (result)
        *result = tmp;
    return true;
}

// write 32-bit words to RAM starting from startWord offset
bool MlinkDevice::memWriteBlk(int startWord, const std::vector<quint32> &data)
{
    RegIoPacket tx, rx;
    tx.src = MlinkHostAddress;
    tx.dst = myAddress;
    tx.type = FRAME_TYPE_RAM_REQ;
    const size_t txHdrSz = 1;
    const size_t wordCount = data.size();

    size_t totalWrite = 0;
    while(totalWrite < wordCount) {
        size_t blockSize = std::min(wordCount - totalWrite, getMaxRamRequestSize());
        const auto txSz = txHdrSz+blockSize;
        const auto rxSz = txSz;
        tx.data.resize(txSz);
        tx.data[0] = ( 0x00000000 |((blockSize&0x1FF)<<22)| ((startWord+totalWrite)&0x3FFFFF) );

        for(size_t i=0; i<blockSize; i++)
            tx.data[txHdrSz+i] = data[totalWrite+i];

        bool ok = ctrlExchangeSingle(tx, &rx);
        if (!ok || rx.data.size() != (int)rxSz) {
            auto s = QString("Failed mem write[%1]: expected %2, received %3 words")
                         .arg(startWord, 6, 16, QChar('0')).arg(rx.data.size()).arg(rxSz);
            emit protocolMessage(s);
            return {};
        }
        totalWrite += blockSize;
    }
    return true;
}


// maximum number of data words for single RAM read/write operation
size_t MlinkDevice::getMaxRamRequestSize() const
{
    // limit request size
    int eth_size = 1500; // MRU
    int hdr_size = 14+20+8+20; // eth+ip+udp+mlink(1)
    return std::min(link.getMaxDataSize()-1, (unsigned long)(eth_size-hdr_size)/4 /*0x1FFul*/); // 2^9 words maximum
}

// maximum number of 16-bit registers for single register read/write operation
size_t MlinkDevice::getMaxRegRequestSize() const
{
    return link.getMaxDataSize(); // one command word for one register
}

bool MlinkDevice::validateIoAck(const RegIoPacket &tx, const RegIoPacket &rx, const std::string &description)
{
    std::string preamble(description);
    if (preamble.empty())
        preamble = "I/O operation";
    int exp_rx_size = MlinkPacketProtocol::ackSize(tx);
    unsigned int exp_rx_type = MlinkPacketProtocol::ackType(tx);
    unsigned int exp_rx_src = tx.dst;
    unsigned int exp_rx_dst = tx.src;
    bool ok = true;
    std::ostringstream ost;
    ost << "Communication error" << std::endl;
    bool isEmpty = rx.data.empty() && !rx.dst && !rx.seq && !rx.src && !rx.type;
    if (isEmpty) {
        ost << preamble << ": " << "Empty (no reply)" << std::endl;
        ok = false;
    }
    if (rx.type != exp_rx_type) {
        ost << preamble << ": " << "Bad acknowledge type, received 0x" << std::hex << rx.type << ", expected 0x" << exp_rx_type << std::dec << std::endl;
        ok = false;
    }
    // exclude broadcast
    if (tx.dst != 0xFEFE && rx.src != exp_rx_src) {
        ost << preamble << ": " << "Bad acknowledge src addr, received 0x" << std::hex << rx.src << ", expected 0x" << exp_rx_src << std::dec << std::endl;
        ok = false;
    }
    if (rx.dst != exp_rx_dst) {
        ost << preamble << ": " << "Bad acknowledge dst addr, received 0x" << std::hex << rx.dst << ", expected 0x" << exp_rx_dst << std::dec << std::endl;
        ok = false;
    }
    // validate size for known frame types
    if ( (tx.type == FRAME_TYPE_LOOPTEST)
         || (tx.type == FRAME_TYPE_CTRL_REQ)
         || (tx.type == FRAME_TYPE_EEPROM_REQ)
         || (tx.type == FRAME_TYPE_RAM_REQ)
         ) {
        if (rx.data.size() != exp_rx_size) {
            ost << preamble << ": " << "Bad acknowledge size, received " << rx.data.size() << " words, expected " << exp_rx_size << std::endl;
            if (rx.data.size() < exp_rx_size)
                ok = false;
        }
        if (rx.seq != tx.seq) {
            ost << preamble << ": " << "Bad sequence number, received " << rx.seq << ", expected " << tx.seq << std::endl;
            ok = false;
        }
    }
    // validate CTRL request contents
    if (ok && (tx.type == FRAME_TYPE_CTRL_REQ)) {
        for(int i=0; i<tx.data.size(); i++) {
            const PacketRawDataType txData = tx.data[i];
            const PacketRawDataType rxData = rx.data[i];
            const bool isRead = 0x80000000u && txData;
            unsigned int mask = 0xFFFFFFFFu;
            if (isRead) mask &= ~0x0000FFFFu; // mask read data
            if ((rxData & mask) != (txData & mask)) {
                ost << preamble << ": " << "Bad acknowledge data, sent "
                    << std::hex << txData << " received " << rxData << std::endl;
                ok = false;
            }
        }
    }
    if (!ok) {
        //        tx.dump();
        //        rx.dump();
        ost << "TX: " << tx << std::endl;
        ost << "RX: " << rx << std::endl;
        ost << std::endl;
        auto s = QString("validateIoAck: %1").arg(QString::fromStdString(ost.str()));
        emit protocolMessage(s);
        return {};
    }
    return ok;
}

bool MlinkDevice::hasAddressExpired() const
{
    return link.hasAddressExpired();
}

bool MlinkDevice::ctrlExchangeSingle(const RegIoPacket &tx_in, RegIoPacket *rx_out)
{
    Q_ASSERT(rx_out);
    if (!rx_out)
        return false;
    int retry_count = 0;
    int seq = nextTxPacketSequenceNumber();
    protocolTimer.restart();
    while (!link.hasDeleteFlag()) {
        RegIoPacket tx(tx_in);
        tx.seq = seq;

        link.rawSend(tx);
        double random_delay = MlinkPacketProtocol::random_delay(retry_count);
        auto timeout = MlinkPacketProtocol::getTimeout() * random_delay;
        bool received = link.getCtrlPacketBySeq(rx_out, tx.seq, timeout);
        if (received && rx_out->seq == tx.seq) {
            bool valid = validateIoAck(tx, *rx_out);
            if (valid) {
                if (retry_count > 3) { // ignore first retry
                    auto s = QString("slow reply: %1 retries, %2 ms")
                                 .arg(retry_count)
                                 .arg(protocolTimer.elapsed());
                    emit protocolMessage(s);
                }
                return true;
            }
        }
        retry_count++;
        // detect offline device (address timed out)
        bool addressExpired = link.hasAddressExpired();
        auto retry_limit = addressExpired ? 3 : maxRetryCount;
        if (retry_count > retry_limit) {
            auto reason = hasAddressExpired() ? "device offline" : "communication timeout";
            emit protocolMessage(reason);
            emit protocolTimeout();
            return false;
        }
    }
    return false;
}

QString MlinkDevice::getIdent() const
{
    return link.getHostAddress().toString();
}

} // namespace mlink
