#include "MongoCredentialsFile.h"

#include "base/Globals.h"
#include "base/ServiceConfigUrl.h"
#include "mongo/MongoConsts.h"
#include "mongoqt/MongoHost.h"

#include <QSettings>
#include <QUrl>
#include <QUrlQuery>

namespace {
constexpr const char *DEFAULT_URL = "mongodb://localhost";
constexpr const bool DEFAULT_ENABLE = true;
constexpr const char *serviceName = "mongodb";
}

MongoHost MongoCredentialsFile::get_host(const QString &environment)
{
    MongoHost cred;
    const auto &settings = Globals::getSettings();
    const QString base_path = environment + "/" + serviceName;
    cred.url = ServiceConfigUrl::getUrl(environment, serviceName, QUrl(DEFAULT_URL));
    cred.enable = ServiceConfigUrl::isEnabled(serviceName, DEFAULT_ENABLE);

    // QString database = settings->value(base_path + "_database").toString();

    QString default_db = settings->value(base_path + "_default_db").toString();

    cred.environment = environment;

    if (cred.url.scheme() != "mongodb" && cred.url.scheme() != "mongodb+srv") { // mongodb or mongodb+srv
        qWarning().noquote().nospace()
                << "Invalid URL scheme " << cred.url.scheme() << ", assuming " << "mongodb";
        cred.url.setScheme("mongodb");
    }
    if (!cred.url.path().isEmpty())
        cred.url.setPath("/");

    if(!default_db.isEmpty())
        cred.db = default_db;

    QUrlQuery q(cred.url.query());
    // add query options from config, replace existing
    const QString optKeyPrefix = base_path + "_opt_";
    const auto optKeys = settings->allKeys().filter(optKeyPrefix);
    if (!optKeys.isEmpty()) {
        for (const auto &k : optKeys) {
            const QString value = settings->value(k).toString();
            QString key = k;
            key.remove(0, optKeyPrefix.size());
            q.removeQueryItem(key);
            q.addQueryItem(key, value);
        }
    }
    // add default query options
    const QList<QPair<QString, QString>> defaultOptions {
        {"connectTimeoutMS", QString::number(connectTimeoutMS)},
        {"serverSelectionTimeoutMS", QString::number(serverSelectionTimeoutMS)}
    };
    for (const auto &i : defaultOptions) {
        const auto &key = i.first;
        const auto &value = i.second;
        if (!q.hasQueryItem(key))
            q.addQueryItem(key, value);
    }

    cred.url.setQuery(q);
    if (cred.url.hasQuery() && cred.url.path().isEmpty())
        cred.url.setPath("/");
    return cred;
}

QList<MongoHost> MongoCredentialsFile::get_all_hosts()
{
    const auto &settings = Globals::getSettings();

    const auto &groups = settings->childGroups();
    const auto &allKeys = settings->allKeys();

    QList<MongoHost> creds_list;
    for(const auto &g : groups) {
        if(allKeys.filter(g + "/" + serviceName).isEmpty())
            continue;
        creds_list.append(get_host(g));
    }

    return creds_list;
}
