// QString s = "<!DOCTYPE pnp_message>"
//             "<discover_request/>"

// QString s = "<!DOCTYPE pnp_message>"
//             "<program name=\"pnp_search\">"
//             " <options>"
//             "  <option name=\"targets\" value=\"all\"/>"
//             " <options>"
//             "</program>";

//    QString s = "<!DOCTYPE pnp_message><program name=\"mstream\">"
//            "<options><option name=\"address\" value=\"10.93.120.171\"/><option name=\"port\" value=\"1222\"/></options>"
//            "<interfaces><interface type=\"data downflow\" host=\"0.0.0.0\" port=\"33511\"/></interfaces></program>";

#include "PNPServer.h"

#include <QCoreApplication>
#include <QDebug>
#include <QNetworkInterface>
#include <QTimer>
#include <QSettings>
#include <QUdpSocket>
#include <arpa/inet.h>
#ifdef QT_DEBUG
#include <QTime>
#endif

#include "base/Globals.h"
#include "multicast-listener/MultiCastListener.h"

#include "PNPServerUtil.h"

namespace {
constexpr const int SEARCH_INTERVAL_MS = 5000;
constexpr const quint16 PNP_DISCOVER_PORT = 33304;
constexpr const auto PNP_DISCOVER_IP_ADDR = "239.192.1.2";
constexpr const auto k_pnp = "pnp";
}

PNPServer::PNPServer(QObject *parent)
    : QObject(parent)
    , searchTimer(new QTimer(this))
    , multiCastListener(new MultiCastListener(QHostAddress(PNP_DISCOVER_IP_ADDR), getPnpPort(), this))
    , multiCastSender(new QUdpSocket(this))
    , deferredSearchResponceTimer(new QTimer(this))
{
    setObjectName("PnP Server");
    searchTimer->setInterval(SEARCH_INTERVAL_MS);
    connect(searchTimer, &QTimer::timeout, this, &PNPServer::sendSearchRequest);
    connect(multiCastListener, &MultiCastListener::received, this, &PNPServer::decode);
    connect(multiCastListener, &MultiCastListener::multicastInterfacesUpdated,
            this, &PNPServer::sendSearchRequest);
    qRegisterMetaType<ProgramDescription>();
    if(!multiCastSender->bind())
        qFatal("Can't bind UDP socket for PNP multicast");
    u_char ttl=15;
    if(setsockopt(static_cast<int>(multiCastSender->socketDescriptor()),
                  IPPROTO_IP, IP_MULTICAST_TTL, &ttl,sizeof(ttl)) != 0) {
        qCritical("Can't set TTL to %d for PNP multicast", ttl);
    }
    connect(qApp, &QCoreApplication::aboutToQuit,
            this, &QObject::deleteLater);
    connect(deferredSearchResponceTimer, &QTimer::timeout, this, &PNPServer::deferredSearchResponce);
    deferredSearchResponceTimer->setSingleShot(true);
    deferredSearchResponceTimer->setInterval(500);
}

PNPServer::~PNPServer()
{
    if(searchEnabled)
        stopSearch();
    while(!progDesc.isEmpty()) {
        closeProgram(progDesc.first());
    }
    multiCastSender->flush();
}

QString PNPServer::getSearchStr()
{
    PNPMessage mes;
    mes.type = PNPMessage_Type_Search;
    mes.searchTargets = searchTargets;
    return mes.getXml();
}

bool PNPServer::decode(const QByteArray &data, const QHostAddress &host)
{
    PNPMessage msg;
    bool ok = msg.decode(data, host);
    if(ok){
        switch (msg.type) {
        case PNPMessage_Type_Search:
        {
            bool instantSend = !deferredSearchResponceTimer->isActive() &&
                (!lastSearchTime.isValid() || lastSearchTime.elapsed()>500);
            bool wasSent = false;
            for (auto it=progDesc.cbegin(); it!=progDesc.cend(); ++it) {
                if(msg.searchTargets.isEmpty() || msg.searchTargets.contains(it->getDaqModule())){
                    if(instantSend) {
                        sendProgDescr(*it);
                        wasSent = true;
                    } else
                        searchRequestSet << it->uuid;
                }
            }
            if(wasSent)
                lastSearchTime.restart();
            if(!searchRequestSet.isEmpty() && !deferredSearchResponceTimer->isActive())
                deferredSearchResponceTimer->start();
            break;
        }
        case PNPMessage_Type_Discover:
        case PNPMessage_Type_Close:
            if(searchEnabled) {
                if(searchTargets.isEmpty() || searchTargets.contains(msg.progDesc.getDaqModule())){
                    bool isMine=false;
                    for (const auto &pd : qAsConst(progDesc)) {
                        if(msg.progDesc.uuid == pd.uuid){
                            isMine = true;
                            break;
                        }
                    }
                    if(!isMine){
//                        print("==========");
//                        print(msg.toString());
                        if(msg.type==PNPMessage_Type_Discover) {
                            emit programDiscovered(msg.progDesc, data.size());
                        } else
                            emit programClosed(msg.progDesc);
                    }
                }
            }
            break;
        default:
            qWarning()<<"Unknown msg type:"<<data;
            ok=false;
            break;
        }
    } else {
        static QSet<QUuid> blockedApps;
        const auto &uuid = msg.progDesc.uuid;
        if(!uuid.isNull()) {
            if(!blockedApps.contains(uuid)) {
                if(blockedApps.size()>1000)
                    blockedApps.clear();
                blockedApps << uuid;
                qWarning()<<"Failed decode msg:"<<data;
            }
        } else {
            static QMap<quint32, QElapsedTimer> hostWarnTimeout;
            auto ip = host.toIPv4Address();
            auto it = hostWarnTimeout.find(ip);
            if(it==hostWarnTimeout.end() || it->elapsed()>10e3) {
                hostWarnTimeout[ip].start();
                qWarning()<<"Failed decode msg:"<<data;
            }
        }
    }
    return ok;
}

void PNPServer::sendSearchRequest()
{
    if(!searchEnabled)
        return;
    const QString searchStr = getSearchStr();
    sendMulticastMsg(searchStr);
}

void PNPServer::deferredSearchResponce()
{
    auto cachedSet = searchRequestSet; // searchRequestSet will be modified by sendProgDescr(...)
    for(auto i : cachedSet) {
        for(auto pd : progDesc) {
            if(pd.uuid == i) {
                sendProgDescr(pd);
                break;
            }
        }
    }
    searchRequestSet.clear();
    lastSearchTime.restart();
}

void PNPServer::changeProgramDescription(const ProgramDescription &pd)
{
    progDesc.clear();
    if(pd.isEmpty())
        return;
    progDesc.append(pd);
    sendProgDescr(pd);
}

void PNPServer::updateProgramDescription(const ProgramDescription &pd)
{
    if(pd.isEmpty())
        return;

    bool found = false;
    for(int i=0; i<progDesc.size(); ++i) {
        ProgramDescription &oldPd = progDesc[i];
        if(oldPd.sameProgram(pd)){
            if(progDesc[i] == pd)
                return;
            progDesc[i] = pd;
            found = true;
            break;
        }
    }

    if(!found)
        progDesc.append(pd);

    sendProgDescr(pd);
}

void PNPServer::closeProgram(const ProgramDescription &pd)
{
    const QString closingStr = pd.getClosingStr();
    for(int i=0; i<progDesc.size();) {
        if(progDesc[i].sameProgram(pd)){
            progDesc.removeAt(i);
        } else
            ++i;
    }
//    for (auto p : progDesc) {
//        if(p.sameProgram(pd)){
//            progDesc.removeAt(progDesc.indexOf(p));
//        }
//    }
    sendMulticastMsg(closingStr);
}

void PNPServer::sendProgDescr(ProgramDescription pd)
{
    searchRequestSet.remove(pd.uuid);
    pd.seq = ++(seqMap[pd.uuid]);
    sendMulticastMsg(pd.getAnnounceStr());
}

void PNPServer::sendMulticastMsg(const QString &msg)
{
    const QByteArray data = msg.toLatin1();
    qint64 nwr = multiCastSender->writeDatagram(data, QHostAddress(PNP_DISCOVER_IP_ADDR), getPnpPort());
    if (nwr != data.size()) {
        qCritical() << "[PNPServer::sendMulticastMsg] "
                    << multiCastSender->errorString();
    }
//        qInfo() << "PNPServer Send "<< nwr << " bytes:" << msg;
}

void PNPServer::print(QString s) const
{
#if QT_VERSION < QT_VERSION_CHECK(5, 5, 0)
    qDebug()<<s;
#else
    qInfo()<<s;
#endif
}

quint16 PNPServer::getPnpPort()
{
    const auto &settings = Globals::getSettings();
    QString key = Globals::getEnvironment() + "/" + k_pnp + "_port";
    quint16 port = 0;
    if (settings->contains(key)) {
        port = settings->value(key).toInt();
    }

    return port ? port : PNP_DISCOVER_PORT;
}

void PNPServer::startSearch(DaqModuleSet targets, bool single)
{
    searchTargets = targets;
    searchEnabled = true;
    sendSearchRequest();
    if(single) {
//        print(QString("Pnp_search single request for targets:%1")
//              .arg(getSearchString(searchTargets)));
    } else {
        searchTimer->start();
        print(QString("Pnp_search started with %1ms interval, targets:%2")
                  .arg(SEARCH_INTERVAL_MS).arg(getSearchString(searchTargets))
              );
    }
}

void PNPServer::stopSearch()
{
    searchEnabled = false;
    searchTimer->stop();
    print("Pnp_search stopped");
}

QVector<QHostAddress> PNPServer::getLocalHostAddresses()
{
    QVector<QHostAddress> addresses;
    const auto &allInterfaces = QNetworkInterface::allInterfaces();
    for (const auto &interface : allInterfaces) {
        if ( !(interface.flags() & QNetworkInterface::IsUp) ) continue;
        if ( !(interface.flags() & QNetworkInterface::IsRunning) ) continue;
        if (  (interface.flags() & QNetworkInterface::IsLoopBack) ) continue;
        const auto &allAddresses = interface.allAddresses();
        for (const auto &addr : allAddresses) {
            if(addr.isNull() || addr.protocol() != QAbstractSocket::IPv4Protocol)
                continue;
            if(addr.isLoopback())
                continue;
            addresses.push_back(addr);
        }
    }
    return addresses;
}

void PNPServer::enableSearch()
{
    searchEnabled = true;
    sendSearchRequest();
    searchTimer->start();
    print(QString("Pnp_search started with %1ms interval, targets:%2")
              .arg(SEARCH_INTERVAL_MS).arg(getSearchString(searchTargets))
          );
}
